// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
  Ejemplo de programa servidor. Atiende peticiones aritmeticas, el protocolo de transporte
  sera el comm::Transport y el contenido del mensaje sera el resutaldo de un comm::Codec con los
  (x, y, op) -> El resultado sera estos tres componente mas result.
  
  Ejemplo de uso del sistema de receiveres, que son capaces de tratar N peticiones de forma
  totalmetne simultanea en caso de estar en un entorno MT.

  Para poder probar el sistema de congestion se puede indicar un numero de milisegundos de 
  retardo aplicados a cada contestacion.

  Los clientes pueden ser: client.p o kclient.p
*/
#include <iostream>

#include <anna/core/core.hpp>
#include <anna/comm/comm.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/app/functions.hpp>

#include <anna/test/Request.hpp>
#include <anna/test/Response.hpp>
#include <anna/test/Communicator.hpp>

using namespace std;
using namespace test;

class MyCommunicator : public test::Communicator {
public:
   MyCommunicator () :  test::Communicator () {;}
};

class MyReceiver : public Receiver {
public:
   static const char* className () throw () { return "MyReceiver"; }

private:
   Request a_request;
   Response a_response;
   MyCommunicator* a_communicator;
   
   MyReceiver () : Receiver ("MyReceiver") { ; }
   void initialize () throw (RuntimeException);
   void apply (comm::ClientSocket &, const Message&) throw (RuntimeException);
   
   friend class Allocator <MyReceiver>;      
};

class ArithmeticServer : public comm::Application {
public:
   ArithmeticServer ();

   comm::DatagramSocket* getOutput () throw () { return a_output; }
      
private:
   MyCommunicator* a_communicator;
   ReceiverFactoryImpl <MyReceiver> a_receiverFactory;
   comm::DatagramSocket* a_input;
   comm::DatagramSocket* a_output;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
   xml::Node* asXML (xml::Node* app) const throw ();
};

using namespace std;
using namespace anna::comm;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   ArithmeticServer app;

   srand (time (NULL));
   
   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("arithmeticServer", new TraceWriter ("file.trace",4096000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

ArithmeticServer::ArithmeticServer () : 
   Application ("arithmeticServer", "Servidor de operaciones (iRS)", "1.0"),
   a_communicator (NULL)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   
   commandLine.add ("as", CommandLine::Argument::Optional, "Direccion broadcast en el que servidor atiende peticiones.");
   commandLine.add ("ps", CommandLine::Argument::Mandatory, "Puerto en el que el servidor atiende las peticiones.");
   commandLine.add ("ac", CommandLine::Argument::Optional, "Direccion broadcast en el que cliente atiende respuestas.");
   commandLine.add ("pc", CommandLine::Argument::Mandatory, "Puerto al que enviar las respuestas");
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
}

//-----------------------------------------------------------------------------------------
// Inicializa el servidor de sockets.
//-----------------------------------------------------------------------------------------
void ArithmeticServer::initialize () 
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   const comm::Device* device = Network::instantiate ().find (Device::asAddress (cl.getValue ("as")));
   int port = cl.getIntegerValue ("ps");

   a_communicator = new MyCommunicator ();

   INetAddress localAddress (device, port);
   a_input = new DatagramSocket (DatagramSocket::ReadOnly, localAddress);
   a_input->setReceiverFactory (a_receiverFactory);
   a_communicator->attach (a_input);

   device = Network::instantiate ().find (Device::asAddress (cl.getValue ("ac")));
   port = cl.getIntegerValue ("pc");

   INetAddress remoteAddress (device, port);
   a_output = new DatagramSocket (DatagramSocket::WriteOnly, remoteAddress);
   a_output->connect ();
}

//-----------------------------------------------------------------------------------------
// Atiende las peticiones.
// Cuando hay un nuevo mensaje invocar�a Communicator::eventReceiveMessage
//-----------------------------------------------------------------------------------------
void ArithmeticServer::run ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   a_communicator->accept ();
}

xml::Node* ArithmeticServer::asXML (xml::Node* app) const 
   throw ()
{
   xml::Node* node = app::Application::asXML (app);
   
   node->createAttribute ("MaxMessage", a_communicator->getMaxMessage ());
   node->createAttribute ("Message", a_communicator->getMessage ());
   
   return node;
}

void MyReceiver::initialize ()
   throw (RuntimeException)
{
   a_communicator = app::functions::component <MyCommunicator> (ANNA_FILE_LOCATION);
}

void MyReceiver::apply (ClientSocket&, const Message& message) 
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyReceiver", "apply", ANNA_FILE_LOCATION));

   a_request.decode (message.getBody ());

   a_communicator->delay ();

   a_response.x = a_request.x;
   a_response.y = a_request.y;
   a_response.initTime = a_request.initTime;

   switch (a_response.op = a_request.op) {
      case '+':
         a_response.result = a_request.x + a_request.y;
         break;
      case '-':
         a_response.result = a_request.x - a_request.y;
         break;
      case '*':
         a_response.result = a_request.x * a_request.y;
         break;
      case '/':
         a_response.result = (a_request.y != 0) ? (a_request.x / a_request.y): 0;
         break;
   }

   LOGINFORMATION (
      string msg = anna::functions::asString ("%d %c %d = %d", a_request.x, a_request.op, a_request.y, a_response.result);
      msg += anna::functions::asText (" | InitTime: ",  a_response.initTime);
      Logger::information (msg, ANNA_FILE_LOCATION);
   )

   try {
      static_cast <ArithmeticServer&> (app::functions::getApp ()).getOutput ()->send (a_response);
   }
   catch (Exception& ex) {
      ex.trace ();
   }
}
