// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <iostream>

#include <anna/core/mt/ThreadManager.hpp>
#include <anna/core/util/CommandLine.hpp>
#include <anna/core/mt/Mutex.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/mt/Runnable.hpp>
#include <anna/core/mt/Thread.hpp>
#include <anna/core/tracing/TraceWriter.hpp>
#include <anna/core/util/SortedVector.hpp>

#include <anna/io/functions.hpp>

using namespace std;
using namespace anna;

class Client : public Runnable {
public:
   Client (const int id, const int value) : Runnable (), a_id (id), a_value (value) {;}

private:
   const int a_id;
   const int a_value;

   void run () throw (RuntimeException);

   anna_complete_runnable (Client);
};

class SOAP {
public:
   static int theMethod (const int id, const int value) throw (RuntimeException);
};

class Receiver;

class Waiter {
public:
   int getId () const throw () { return a_id; }
   int getValue () const throw () { return a_value; }

   void setId (const int id) throw () { a_id = id; }
   void setValue (const int value) throw () { a_value = value; }

   void sendRequest () throw (RuntimeException);
   int waitResponse () throw (RuntimeException);
   void notifyResponse (const int result) throw (RuntimeException);   

   void initialize () throw (RuntimeException);
   
private:
   static const int ReadChannel = 0;
   static const int WriteChannel = 1;
   
   static ThreadManager a_threadManager;

   int a_pipe [2];
   int a_id;
   int a_value;
   int a_result;
   Receiver* a_receiver;
};

class Context : public SafeRecycler <Waiter> {
public:  
   Waiter* createWaiter (const int id, const int value) throw (RuntimeException);
   void destroyWaiter (Waiter* waiter) throw (RuntimeException);
   Waiter* findWaiter (const int id) throw () { return a_waiters.find (id); }

private:
   struct SortByID {
      static int value (const Waiter* waiter) throw () { return waiter->getId (); }
   };

   typedef SortedVector <Waiter, SortByID, int> waiter_container;

   waiter_container a_waiters;
};

class Receiver : public Runnable {
public:
   Receiver (const int id, const int value) : a_id (id), a_value (value) {;}

private:
   int a_id;
   int a_value;

   void run () throw (RuntimeException);

   anna_complete_runnable (Waiter);
};

Context* g_context = NULL;
ThreadManager Waiter::a_threadManager ("WaiterThreadManager");

int main (const int argc, const char** argv)
{
   CommandLine& ccll = CommandLine::instantiate ();
   ThreadManager* threadManager;
   
   g_context = new Context;

   try {
      ccll.add ("maxthr", CommandLine::Argument::Mandatory, "Maximum number of threads");
      ccll.add ("mode", CommandLine::Argument::Mandatory, "Manager instance mode");
      ccll.add ("nthr", CommandLine::Argument::Mandatory, "Number of threads to create");
      ccll.add ("trace", CommandLine::Argument::Optional, "Trace level");

      ccll.initialize (argv, argc);
      ccll.verify ();

      Logger::initialize ("thread", new TraceWriter ("file.trace", 2048000));

      if (ccll.exists ("trace"))
         Logger::setLevel (Logger::asLevel (ccll.getValue ("trace")));
         
      ThreadManager::Mode::_v mode = ThreadManager::Mode::asEnum (ccll.getValue ("mode"));

      if (mode == ThreadManager::Mode::None) {
         string msg (ccll.getValue ("mode"));
         msg += " mode not valid (";
         msg +=  ThreadManager::Mode::asList ();
         msg += ")";
         throw RuntimeException (msg, ANNA_FILE_LOCATION);
      }

      const int maxThread = ccll.getIntegerValue ("maxthr");
      int nthr = max (2, ccll.getIntegerValue ("nthr"));
     
      threadManager = new ThreadManager ("MyManager", mode, maxThread, Thread::Flag::Joinable);
      Thread* thread = NULL;

      // Primero genera los clientes que generan la primera peticin de cloulo. Todos
      // terminan llamando al mtodo SOAP::theMethod, que ser el que monte la  espera
      // y "visualize la respuesta"
      for (int ii = 0, min = 0, max = 10; ii < nthr; ii ++) {
          
         Client* client = new Client (ii, min + rand () % max);
         // Crear los threads Tx 
         thread = threadManager->createThread ();
         thread->start (*client);

         min = max;
         max += 10;
      }
      
      threadManager->join ();
   }
   catch (RuntimeException& ex) {
      cout << ex.asString () << endl << endl;
      ex.trace ();
   }
}

// Esto simula el proceso cliente que enva la peticin al proceso gSOAP
/* Se ejecuta desde el thread Tx */
void Client::run () 
   throw (RuntimeException)
{
   int result = SOAP::theMethod (a_id, a_value);

   LOGDEBUG (
      string msg (asString ());
      msg += functions::asText (" | Result: ", result);
      Logger::debug (msg, ANNA_FILE_LOCATION);
   );
}

// La implementacin del mtodo SOAP debe devolver un valor, pero no puede 
// calcular por s misma la respuesta, sino que tiene que apoyarse en un tercero
// que ser el que le d la respuesta "real" en algn momento.
/* Tx */
int SOAP::theMethod (const int id, const int value)
   throw (RuntimeException)
{
   int result = -1;

   Waiter* waiter = NULL;

   try {
      waiter = g_context->createWaiter (id, value);
      waiter->initialize ();
      waiter->sendRequest (); 
      result = waiter->waitResponse ();
      
      cout << "Id: " << id << " | Value: " << value << " | Result: " << result << endl << endl;      
   }
   catch (RuntimeException& ex) {
      cout << "Id: " << id << " | Value: " << value << " | " << ex.asString () << endl << endl;
      ex.trace ();
   }  

   g_context->destroyWaiter (waiter);  // Slo lo libera cuando ya no necesita usar sus datos.
   
   return result;
}

/* Tx */
void Waiter::initialize () 
   throw (RuntimeException)
{
   if (pipe (a_pipe) == -1) {
      string msg (functions::asText ("Waiter::initialize | Id: ", a_id));
      throw RuntimeException (msg, errno, ANNA_FILE_LOCATION); 
   }     
}

/* Tx */
void Waiter::sendRequest () 
   throw (RuntimeException)
{
   try {
      /* Ojo que esto generar memory-leaks, pero no es el objeto de ste ejemplo la buena gestin de la memoria */
      Receiver* receiver = new Receiver (a_id, a_value);   
      a_threadManager.createThread ()->start (*receiver);         
   }
   catch (RuntimeException& ex) {
      string msg ("Waiter::sendRequest | ");
      msg += ex.asString ();
      throw RuntimeException (msg, ANNA_FILE_LOCATION);
   }
}

/* Tx */
int Waiter::waitResponse () 
   throw (RuntimeException)
{
   bool receive = io::functions::waitInput (a_pipe [ReadChannel], (Millisecond)500);
   
   if (receive == false) {      
      string msg (anna::functions::asText ("Id: ", a_id));
      msg += anna::functions::asText (" | Value: ", a_value);
      msg += " | Timeout";            
      throw RuntimeException (msg, ANNA_FILE_LOCATION);     
   }
   
   // Esta variable se ha establecido en el notifyResponse (Ty)
   return a_result;
}

/* Ty */
void Waiter::notifyResponse (const int result) 
   throw (RuntimeException)
{
   int byte = 0;
   
   a_result = result;
   
   write (a_pipe [WriteChannel], &byte, sizeof (byte));
}

// Modela al thread que recibe la respuesta correspondiente a la peticin por la 
// que debera haber un Waiter esperando.
/* Ty */
void Receiver::run () 
   throw (RuntimeException)
{
   anna::functions::sleep ((Millisecond)(rand () % 1000));

   int result = a_value + a_value;

   // Protege el contexto para evitar condiciones de carrera, como que en el "mismo" instante
   // en que el receptor decide responder, el waiter en Tx decida que ha terminado de esperar
   // la respuesta.
   Guard guard (g_context, "Receiver::run");
   
   Waiter* waiter = g_context->findWaiter (a_id);

   // Si el Waiter se cans de esperar la respuesta ...
   if (waiter == NULL)
      return;

   waiter->notifyResponse (result);
}

Waiter* Context::createWaiter (const int id, const int value) 
   throw (RuntimeException)
{
   Guard guard (*this, "Context::createWaiter");
   
   Waiter* result = a_waiters.find (id);
   
   if (result != NULL) 
      throw RuntimeException ("Waiter already used", ANNA_FILE_LOCATION);
   
   result = Recycler <Waiter>::create ();
   
   result->setId (id);
   result->setValue (value);
   
   a_waiters.add (result);
   
   return result;   
}

void Context::destroyWaiter (Waiter* waiter) 
   throw (RuntimeException)
{
   Guard guard (this, "Context::destroyWaiter");

   // Lo elimina de la lista ordenada
   a_waiters.erase (waiter);   
   
   // Lo marca como disponible para usar
   Recycler <Waiter>::release (waiter);   
}

