// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Project
#include <anna/core/core.hpp>
#include <anna/diameter/functions.hpp>
#include <anna/time/functions.hpp>
#include <anna/diameter/codec/Engine.hpp>
#include <anna/diameter.comm/Response.hpp>
#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/OriginHost.hpp>

// Process
#include <MyLocalServer.hpp>
#include <MyDiameterEngine.hpp>
#include <MyDiameterEntity.hpp>
#include <Launcher.hpp>
#include <anna/testing/TestManager.hpp>


void MyLocalServer::eventRequest(anna::diameter::comm::ServerSession *serverSession, const anna::DataBlock &message)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tm("launcher::MyLocalServer", "eventRequest", ANNA_FILE_LOCATION));
  // Performance stats:
  Launcher& my_app = static_cast <Launcher&>(anna::app::functions::getApp());
  anna::diameter::comm::OriginHost *my_node = my_app.getOriginHost(getEngine()->getOriginHostName());
  anna::diameter::codec::Engine *codecEngine = my_node->getCodecEngine();

  // CommandId:
  anna::diameter::CommandId cid = anna::diameter::codec::functions::getCommandId(message);
  LOGDEBUG
  (
    std::string msg = "Request received: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    msg += " | DiameterServer: ";
    msg += anna::functions::socketLiteralAsString(serverSession->getAddress(), serverSession->getPort());
    msg += " | EventTime: ";
    msg += anna::time::functions::currentTimeAsString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  // Write reception
  if(my_node->logEnabled()) my_node->writeLogFile(message, "recvfc", serverSession->asString());

  // If no answer is programmed and entity is configured, the failed request would be forwarded even being wrong (delegates at the end point)
  int code = cid.first;
  anna::diameter::codec::Message *programmed_answer = a_reactingAnswers.getMessage(code);
  bool programmed = (programmed_answer != NULL);

  anna::diameter::comm::Entity *entity = my_node->getEntity();
  if(!programmed && entity) {  // forward condition (no programmed answer + entity available)
    anna::diameter::comm::Message *msg = my_node->createCommMessage();
    msg->forwardEndToEnd(); // end-to-end will be kept
    msg->setBody(message);
    msg->setRequestServerSessionKey(serverSession->getKey());
    bool success = entity->send(msg);

    // Detailed log:
    if(my_node->logEnabled()) {
      anna::diameter::comm::Server *usedServer = entity->getLastUsedResource();
      anna::diameter::comm::ClientSession *usedClientSession = usedServer ? usedServer->getLastUsedResource() : NULL;
      std::string detail = usedClientSession ? usedClientSession->asString() : "<null client session>"; // esto no deberia ocurrir
      my_node->writeLogFile(message, (success ? "fwd2e" : "fwd2eError"), detail); // forwarded
    }


    // Testing:
    anna::testing::TestManager::instantiate().receiveMessage(message, serverSession);

    return;
  }

  // Error analisys:
  bool analysisOK = true; // by default
  anna::diameter::codec::Message *answer_message = NULL;
  anna::diameter::codec::Message codecMsg;
  anna::diameter::codec::Message codecAnsMsg;

  CommandLine& cl(anna::CommandLine::instantiate());
  if(!cl.exists("ignoreErrors")) {  // Error analysis
    answer_message = (anna::diameter::codec::Message *)&codecAnsMsg;
    answer_message->clear();

    // Decode
    try { codecMsg.decode(message, answer_message); } catch(anna::RuntimeException &ex) { ex.trace(); }

    answer_message->setStandardToAnswer(codecMsg, my_node->getCommEngine()->getOriginHostName(), my_node->getCommEngine()->getOriginRealmName());
    analysisOK = (answer_message->getResultCode() == anna::diameter::helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS);
  }

  // Programmed answer only when all is ok
  if(analysisOK) {
    if(programmed) {
      answer_message = programmed_answer;
      // Prepare answer:
      my_app.getCommunicator()->prepareAnswer(answer_message, message);
    } else {

      // Testing:
      anna::testing::TestManager::instantiate().receiveMessage(message, serverSession);

      return; // nothing done
    }
  }

  anna::diameter::codec::Engine::ValidationMode::_v backupVM = codecEngine->getValidationMode();

  if(!analysisOK)
    codecEngine->setValidationMode(anna::diameter::codec::Engine::ValidationMode::Never);

  anna::diameter::comm::Message *msg;
  try {
    msg = my_node->createCommMessage();
    msg->setBody(answer_message->code());
    /* response = NULL =*/serverSession->send(msg);

    if(my_node->logEnabled()) my_node->writeLogFile(*answer_message, "sent2c", serverSession->asString());
  } catch(anna::RuntimeException &ex) {
    ex.trace();
    if(my_node->logEnabled()) my_node->writeLogFile(*answer_message, "send2cError", serverSession->asString());
  }

  // release msg
  my_node->releaseCommMessage(msg);

  // Restore validation mode
  codecEngine->setValidationMode(backupVM);

  // Pop front the reacting answer:
  if(analysisOK && programmed) a_reactingAnswers.nextMessage(code);

  // Testing:
  anna::testing::TestManager::instantiate().receiveMessage(message, serverSession);
}

void MyLocalServer::eventResponse(const anna::diameter::comm::Response &response)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tm("launcher::MyLocalServer", "eventResponse", ANNA_FILE_LOCATION));
  Launcher& my_app = static_cast <Launcher&>(anna::app::functions::getApp());
  anna::diameter::comm::OriginHost * my_node = my_app.getOriginHost(getEngine()->getOriginHostName());
  anna::diameter::comm::ClassCode::_v code = response.getClassCode();
  anna::diameter::comm::Response::ResultCode::_v result = response.getResultCode();
  anna::diameter::comm::Message* request = const_cast<anna::diameter::comm::Message*>(response.getRequest());
  const anna::DataBlock* message = response.getMessage();
  const anna::diameter::comm::ServerSession *serverSession = static_cast<const anna::diameter::comm::ServerSession *>(response.getSession());
  bool isBindResponse = (code == anna::diameter::comm::ClassCode::Bind);
  bool isApplicationMessage = (code == anna::diameter::comm::ClassCode::ApplicationMessage);
  bool contextExpired = (result == anna::diameter::comm::Response::ResultCode::Timeout);
  bool isUnavailable = (result == anna::diameter::comm::Response::ResultCode::DiameterUnavailable);
  bool isOK = (result == anna::diameter::comm::Response::ResultCode::Success);

  // CommandId:
  anna::diameter::CommandId request_cid = request->getCommandId();
  LOGDEBUG
  (
    std::string msg = "Response received for original diameter request: ";
    msg += anna::diameter::functions::commandIdAsPairString(request_cid);
    msg += " | Response: ";
    msg += response.asString();
    msg += " | LocalServer: ";
    msg += anna::functions::socketLiteralAsString(serverSession->getAddress(), serverSession->getPort());
    msg += " | EventTime: ";
    msg += anna::time::functions::currentTimeAsString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(isUnavailable) {
    //if (isApplicationMessage)
    LOGWARNING(anna::Logger::warning("Diameter client unavailable for Diameter Request", ANNA_FILE_LOCATION));
  }

  if(contextExpired) {
    //if (isApplicationMessage)
    LOGWARNING(anna::Logger::warning("Context Expired for Diameter Request which was sent to the client", ANNA_FILE_LOCATION));

    if(request_cid != anna::diameter::helpers::base::COMMANDID__Capabilities_Exchange_Request) {  // don't trace CEA
      if(my_node->logEnabled()) my_node->writeLogFile(*request, "req2c-expired", serverSession->asString());
    }
  }

  if(isOK) {
    LOGDEBUG(
      std::string msg = "Received response for diameter message:  ";
      msg += anna::diameter::functions::commandIdAsPairString(request_cid);
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );

    // Write reception
    if(my_node->logEnabled()) my_node->writeLogFile(*message, "recvfc", serverSession->asString());

    // This is not very usual, but answers could arrive from clients:
    anna::diameter::comm::Entity *entity = my_node->getEntity();

    if(entity) {
      anna::diameter::comm::ClientSession *usedClientSession = my_node->getCommEngine()->findClientSession(request->getRequestClientSessionKey());
      std::string detail;

      if(my_node->logEnabled()) detail = usedClientSession ? usedClientSession->asString() : "<null client session>";  // this should not happen
      anna::diameter::comm::Message *msg;

      try {
        msg = my_node->createCommMessage();
        msg->forwardEndToEnd(); // end-to-end will be kept
        msg->setBody(*message);

        // Metodo 1:
        if(usedClientSession) /* response = NULL =*/usedClientSession->send(msg);

        // Metodo 2:
        //msg->setRequestClientSessionKey(request->getRequestClientSessionKey());
        //bool success = entity->send(msg);

        if(my_node->logEnabled()) my_node->writeLogFile(*message, "fwd2e", detail);  // forwarded
      } catch(anna::RuntimeException &ex) {
        ex.trace();
        if(my_node->logEnabled()) my_node->writeLogFile(*message, "fwd2eError", detail);  // forwarded
      }

      // release msgs
      my_node->releaseCommMessage(msg);
      my_node->releaseCommMessage(request);
    }
  }

  // Testing:
  if(isOK) anna::testing::TestManager::instantiate().receiveMessage(*message, serverSession);
}

void MyLocalServer::eventUnknownResponse(anna::diameter::comm::ServerSession *serverSession, const anna::DataBlock &message)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tm("launcher::MyLocalServer", "eventUnknownResponse", ANNA_FILE_LOCATION));
  // Performance stats:
  Launcher& my_app = static_cast <Launcher&>(anna::app::functions::getApp());
  anna::diameter::comm::OriginHost *my_node = my_app.getOriginHost(getEngine()->getOriginHostName());
  // CommandId:
  anna::diameter::CommandId cid = anna::diameter::codec::functions::getCommandId(message);
  LOGDEBUG
  (
    std::string msg = "Out-of-context response received from client: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    msg += " | DiameterServer: ";
    msg += anna::functions::socketLiteralAsString(serverSession->getAddress(), serverSession->getPort());
    msg += " | EventTime: ";
    msg += anna::time::functions::currentTimeAsString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(my_node->logEnabled()) my_node->writeLogFile(message, "recvfc-ans-unknown", serverSession->asString());
}

void MyLocalServer::eventDPA(anna::diameter::comm::ServerSession *serverSession, const anna::DataBlock &message)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tm("launcher::MyLocalServer", "eventDPA", ANNA_FILE_LOCATION));
  // Performance stats:
  Launcher& my_app = static_cast <Launcher&>(anna::app::functions::getApp());
  anna::diameter::comm::OriginHost *my_node = my_app.getOriginHost(getEngine()->getOriginHostName());
  // CommandId:
  anna::diameter::CommandId cid = anna::diameter::codec::functions::getCommandId(message);
  LOGDEBUG
  (
    std::string msg = "Disconnect-Peer-Answer response received from client: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    msg += " | DiameterServer: ";
    msg += anna::functions::socketLiteralAsString(serverSession->getAddress(), serverSession->getPort());
    msg += " | EventTime: ";
    msg += anna::time::functions::currentTimeAsString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  // Write reception
  if(my_node->logEnabled()) my_node->writeLogFile(message, "recvfc", serverSession->asString());

  // Testing:
  anna::testing::TestManager::instantiate().receiveMessage(message, serverSession);
}
