// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
  Ejemplo de programa servidor. Atiende peticiones aritmeticas, el protocolo de transporte sera HTTP
  ver http::Transport y el contenido del mensaje sera el resutaldo de un comm::Codec con los
  (x, y, op) -> El resultado sera estos tres componente mas result.

  Para poder probar el sistema de congestion se puede indicar un numero de milisegundos de
  retardo aplicados a cada contestacion.

  Los clientes pueden ser: http_kclient.p http_client.p
*/
#include <iostream>

#include <anna/core/core.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/comm.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Handler.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/functions.hpp>

#include <anna/test/Request.hpp>
#include <anna/test/Response.hpp>
#include <anna/test/Communicator.hpp>

using namespace std;

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("http_server::MyHandler") {    
      allocateResponse ()->createHeader (http::Header::Type::Date); 
   }

private:
   test::Request a_testRequest;
   test::Response a_testResponse;

   void evRequest (ClientSocket&, const http::Request& request) throw (RuntimeException);
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException) {;}
};

class MyCommunicator : public test::Communicator {
public:
   MyCommunicator () : 
      a_contexts ("Contexts")
   {;}

private:
   ThreadData <MyHandler> a_contexts;
   
   void eventReceiveMessage (comm::ClientSocket&, const Message&) throw (RuntimeException);
};

class HTTPArithmeticServer : public comm::Application {
public:
   HTTPArithmeticServer ();
      
private:
   MyCommunicator a_communicator;
   comm::ServerSocket* a_serverSocket;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
   xml::Node* asXML (xml::Node* app) const throw ();
};

using namespace std;
using namespace anna::comm;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   HTTPArithmeticServer app;

   http::functions::initialize ();

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("http_server", new anna::TraceWriter ("file.trace", 4048000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

HTTPArithmeticServer::HTTPArithmeticServer () : 
   Application ("http_server", "Servidor HTTP de operaciones aritmeticas", "1.0") 
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que atender peticiones");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccin IP en la que atender");
   commandLine.add ("d", CommandLine::Argument::Mandatory, "Retardo aplicado a la contestacion");
   commandLine.add ("r", CommandLine::Argument::Optional, "Indicador de reuso de direccin", false);
   commandLine.add ("limit", CommandLine::Argument::Mandatory, "% de ocupacion que permitimos");
   commandLine.add ("n", CommandLine::Argument::Optional, "Numero de mensajes a servir", true);
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
   commandLine.add ("timeout", CommandLine::Argument::Optional, "Milisegundos transcurridos sin actividad para cerrar el socket");
}

void HTTPArithmeticServer::initialize () 
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   int port = cl.getIntegerValue ("p");
   const comm::Device* device = Network::instantiate ().find (Device::asAddress (cl.getValue ("a")));

   a_serverSocket = new ServerSocket (INetAddress (device, port), cl.exists ("r"), &http::Transport::getFactory ());
}

void HTTPArithmeticServer::run ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   a_communicator.attach (a_serverSocket);
   a_communicator.setDelay ((Millisecond)cl.getIntegerValue ("d"));

   if (cl.exists ("n") == true)
      a_communicator.setMaxMessage (cl.getIntegerValue ("n"));

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   CongestionController::instantiate ().setLimit (cl.getIntegerValue ("limit"));

   if (cl.exists ("timeout"))
      a_communicator.setTimeout ((Millisecond)cl.getIntegerValue ("timeout"));

   a_communicator.accept ();
}

xml::Node* HTTPArithmeticServer::asXML (xml::Node* app) const 
   throw ()
{
   xml::Node* node = app::Application::asXML (app);
   
   node->createAttribute ("MaxMessage", a_communicator.getMaxMessage ());
   node->createAttribute ("Message", a_communicator.getMessage ());
   
   return node;
}

void MyCommunicator::eventReceiveMessage (ClientSocket& clientSocket, const Message& message)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   if (clientSocket.support (http::Transport::className ()) == false)
      return;

   if (canContinue (clientSocket) == false)
      return;   

   delay ();   

   MyHandler& httpHandler = a_contexts.get ();

   httpHandler.apply (clientSocket, message);
}

void MyHandler::evRequest (ClientSocket& clientSocket, const http::Request& request)
   throw (RuntimeException)
{
   const DataBlock& body = request.getBody ();

   if (body.getSize () == 0)
      throw RuntimeException ("La peticion no incorpora los parametros de operacion", ANNA_FILE_LOCATION);

   LOGINFORMATION (
      string msg ("Body recibido: ");
      msg += anna::functions::asString (body);
      Logger::information (msg, ANNA_FILE_LOCATION);
   );

   if (Codec::getType (body) != test::Request::Id) 
      throw RuntimeException ("El mensaje recibido no es una peticion aritmetica", ANNA_FILE_LOCATION);

   a_testRequest.decode (body);

   a_testResponse.x = a_testRequest.x;
   a_testResponse.y = a_testRequest.y;
   a_testResponse.initTime = a_testRequest.initTime;
   
   http::Response* response = allocateResponse ();

   response->setStatusCode (200);

   switch (a_testResponse.op = a_testRequest.op) {
      case '+':
         a_testResponse.result = a_testRequest.x + a_testRequest.y;
         break;
      case '-':
         a_testResponse.result = a_testRequest.x - a_testRequest.y;
         break;
      case '*':
         a_testResponse.result = a_testRequest.x * a_testRequest.y;
         break;
      case '/':
         if (a_testRequest.y == 0) {
            response->setStatusCode (400);
            response->setReasonPhrase ("Division por cero");
            a_testResponse.result = 0;
         }
         else
            a_testResponse.result = a_testRequest.x / a_testRequest.y;
         break;
   }

   response->setBody (a_testResponse.code ());      

   response->find (http::Header::Type::Date)->setValue ("Mon, 30 Jan 2006 14:36:18 GMT");
   
   http::Header* keepAlive = response->find ("Keep-Alive");
   
   if (keepAlive == NULL)
      keepAlive = response->createHeader ("Keep-Alive");   
   
   keepAlive->setValue ("Verificacion del cambio 1.0.7");
      

   LOGINFORMATION (
      string msg = anna::functions::asString ("%d %c %d = %d", a_testRequest.x, a_testRequest.op, a_testRequest.y, a_testResponse.result);
      Logger::information (msg, ANNA_FILE_LOCATION);
   );

   try {
      clientSocket.send (*response);
   }
   catch (Exception& ex) {
      ex.trace ();
   }
}

