// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_app_Application_hpp
#define anna_app_Application_hpp

#include <vector>

#include <anna/core/RuntimeException.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace app {

class Component;
struct functions;

/**
   Application abstraction.

   Gather all the information of resources (version, title, command line,
   threads, etc ...) used by our applications.

   Only one single instance could exists for the application, accessed by mean anna::functions::getApp.
*/
class Application {
public:
  typedef std::vector <Component*>::iterator iterator;
  typedef std::vector <Component*>::const_iterator const_iterator;
  typedef std::vector <pid_t> pid_container;
  typedef pid_container::iterator pid_iterator;
  typedef pid_container::const_iterator const_pid_iterator;

  /**
     Constructor.

     @param shortName Instance logical name.
     @param title Application title (appears when the application starts).
     @param version Program version (recommended X.YRZZn with X = main version, Y = secondary version, ZZ = delivery number
     \param date Component build date. Normally the macro __DATE__.
     \param time Component build time. Normally the macro __TIME__.
  */
  Application(const char* shortName, const char* title, const char* version, const char* date = NULL, const char* time = NULL);

  /**
     Destructor.
  */
  virtual ~Application() { a_components.clear(); }

  // Getters
  /**
     Short name given in constructor.
     \return Short name given in constructor.
  */
  const char* getShortName() const throw() { return a_shortName; }

  /**
     Version given in constructor.
     \return Version given in constructor.
  */
  const std::string& getVersion() const throw() { return a_version; }

  /**
     Devuelve el titulo indicado en el contructor de esta aplicacion.
     \return el titulo indicado en el contructor de esta aplicacion.
  */
  const std::string& getTitle() const throw() { return a_title; }

  /**
     Devuelve el pid asignado por el sistema operativo a la aplicacion que estamos ejecutando.
     @return El pid asignado por el sistema operativo a la aplicacion que estamos ejecutando.
  */
  pid_t getPid() const throw() { return a_pid; }

  /**
     Activa la salida por pantalla del mensaje referente a la licencia GPL 3.0.
     \warning Para que tenga efecto debe ser invocado antes de ejecutar el metodo Application::start.
  */
  void activateGeneralPublicLicense() throw() { a_enableGPL = true; }

  /**
   * Crea un nuevo proceso a partir de este usando el metodo \em fork.
   *
   * Estrictamente este metodo retonara dos veces, una en el ambito del proceso que lo invoco y
   * otra en el ambito del nuevo proceso, llamado proceso hijo.
   *
   * Si la duplicacion se realiza de forma correcta se invoca a los metodos #do_cloneChild y #do_cloneParent de nuestra
   * aplicacion y Component::do_cloneChild y Component::do_cloneParent de cada uno de los componentes registrados.
   *
   * Para saber en cual de los procesos
   * \return La instancia de la nueva aplicacion.
   */
  Application& clone() throw(RuntimeException);

  /**
     Devuelve la instancia del componente que corresponde con el nombre recibido.
     \return La instancia del componente que corresponde con el nombre recibido. Puede ser NULL si no
     hay ningun componente asociado a la �ta aplicacion que corresponda con el nombre.
  */
  Component* find(const char* className) throw();

  /**
     Inicializa los elementos propios de nuestra aplicacio, invoca al metodo #initialize
     e invoca al metodo #run.
  */
  void start() throw(RuntimeException);

  /**
     Devuelve un iterador al primer componente definido en la aplicacion.
     \return un iterador al primer componente definido en la aplicacion.
  */
  const_iterator begin() const throw() { return a_components.begin(); }

  /**
     Devuelve un iterador al primer componente definido en la aplicacion.
     \return un iterador al primer componente definido en la aplicacion.
  */
  iterator begin() throw() { return a_components.begin(); }

  /**
     Devuelve un iterador al ultimo componente definido en la aplicacion.
     \return un iterador al ultimo componente definido en la aplicacion.
  */
  const_iterator end() const throw() { return a_components.end(); }

  /**
     Devuelve un iterador al ultimo componente definido en la aplicacion.
     \return un iterador al ultimo componente definido en la aplicacion.
  */
  iterator end() throw() { return a_components.end(); }

  /**
     Vuelva un documento XML con el contexto de la aplicacion.
     \param file Ruta completa del fichero en el que grabar el contexto de la aplicacion.
  */
  void writeContext(const std::string& file) throw(RuntimeException);

  /**
     metodo que puede ser reescrito en las clases heredadas de Application y que nos
     da la posibilidad de inicializar los elementos particulares de nuestra aplicacion.
     Por defecto no realiza ninguna operacion.
     \warning En el momento de invocar a esta funcin los componentes (\see Component) pueden no
     estar disponibles.
  */
  virtual void initialize() throw(RuntimeException) {;}

  /**
     Devuelve si esta aplicacion tiene soporto para comunicaciones entre procesos.
     \return \em true si la aplicacion tiene soporta para comunicacion entre proceso o \em false
     en otro caso.
  */
  virtual bool supportCommunication() const throw() { return false; }

  /**
   * Este m�todo se invocar� cuando alguna capa superior a �sta detecte un problema tan grave
   * como para parar la aplicaci�n de forma s�bita.
   * \param className Nombre de la clase que genera el evento.
   *
   * \warning En el momento de invocar a este m�todo la aplicaci�n puede estar en un estado
   * no muy estable por lo que se deber�an minizar las operaciones a realizar para evitar
   * bloquear la aplicaci�n.
   */
  virtual void eventAbnormalTermination(const char* className) throw() { ; }

  /**
     Devuelve una cadena con la informacion referente a esta instancia.
     \param parent Nodo XML del que dependende la informacion.
     @return Una cadena con la informacion referente a esta instancia.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

  /**
     Devuelve el objeto sobre el que esta posicionado el iterator recibido como parametro.
     \param ii Iterator que deberia estar comprendido entre #begin y #end.
     \return El objeto sobre el que esta posicionado el iterator recibido como parametro.
  */
  static const Component* component(const_iterator& ii) throw() { return *ii; }

  /**
     Devuelve el objeto sobre el que esta posicionado el iterator recibido como parametro.
     \param ii Iterator que deberia estar comprendido entre #begin y #end.
     \return El objeto sobre el que esta posicionado el iterator recibido como parametro.
  */
  static Component* component(iterator& ii) throw() { return *ii; }

protected:
  static Application* st_application;

  /**
     metodo que debemos implementar para ejecutar nuestra aplicacion.
  */
  virtual void run() throw(RuntimeException) = 0;

  /**
     Handler for SIGUSR1. Application context written by default.
  */
  virtual void signalUSR1() throw(RuntimeException);

  /**
     Handler for SIGUSR2. Nothing done by default.
  */
  virtual void signalUSR2() throw(RuntimeException);

  /**
     Metodo manejador que podemos re-escribir para tratar la recepcion de la senhal SIGTERM.
  */
  virtual void signalTerminate() throw(RuntimeException) {;}

  /**
   * Metodo que debemos implementar en la clase heredada para tratar el proceso de clonado de
   * esta aplicacion desde el punto de vista del proceso original.
   * \warning Si se detecta algun error terminara abortara la ejecucion del proceso hijo.
   */
  virtual void do_cloneParent() throw(RuntimeException) {;}

  /**
   * Metodo que debemos implementar en la clase heredada para tratar el proceso de clonado de
   * esta aplicacion desde el punto de vista del proceso hijo.
   * \warning Si se detecta algun error terminara abortara la ejecucion de este proceso.
   */
  virtual void do_cloneChild() throw(RuntimeException) {;}

  /**
     Devuelve un iterador al primer proceso hijo creado por la aplicacion (ver #clone),
     \return un iterador al primer proceso hijo definido en la aplicacion.
  */
  pid_iterator pid_begin() throw() { return a_pids.begin(); }

  /**
     Devuelve un iterador al final de lista lista de procesos hijos creados por la aplicacion (ver #clone).
     \return un iterador al primer proceso hijo definido en la aplicacion.
  */
  pid_iterator pid_end() throw() { return a_pids.end(); }

  /**
   * Devuelve el numero de procesos hijos.
   * \return el numero de procesos hijos.
   */
  int pid_size() const throw() { return a_pids.size(); }

  /**
     Devuelve un iterador al primer proceso hijo creado por la aplicacion (ver #clone),
     \return un iterador al primer proceso hijo definido en la aplicacion.
  */
  const_pid_iterator pid_begin() const throw() { return a_pids.begin(); }

  /**
     Devuelve un iterador al final de lista lista de procesos hijos creados por la aplicacion (ver #clone).
     \return un iterador al primer proceso hijo definido en la aplicacion.
  */
  const_pid_iterator pid_end() const throw() { return a_pids.end(); }

  /**
     Devuelve el PID sobre el que esta posicionado el iterator recibido como parametro.
     \param ii Iterator que deberia estar comprendido entre #pid_begin y #pid_end.
     \return El objeto sobre el que esta posicionado el iterator recibido como parametro.
  */
  static pid_t pid(pid_iterator& ii) throw() { return *ii; }

  /**
     Devuelve el PID sobre el que esta posicionado el iterator recibido como parametro.
     \param ii Iterator que deberia estar comprendido entre #pid_begin y #pid_end.
     \return El objeto sobre el que esta posicionado el iterator recibido como parametro.
  */
  static pid_t pid(const_pid_iterator& ii) throw() { return *ii; }

private:
  bool a_running;
  std::string a_version;
  const std::string a_title;
  std::vector <Component*> a_components;
  pid_t a_pid;
  const char* a_shortName;
  bool a_enableGPL;
  pid_container a_pids;

  void attach(Component*) throw(RuntimeException);
  void detach(Component*) throw(RuntimeException);

  void startComponents() throw(RuntimeException);
  void stopComponents() throw(RuntimeException);
  void sendSignalToChilds(const int signal) throw();

  void signalUSR(int) throw(RuntimeException);
  static void handlerSignalUSR(int) throw();
  static void handlerSignalTerminate(int) throw();
  static void handlerChildTerminate(int sig) throw();

  friend class Component;  // Para poder acceder a los metodo attach y detach
  friend struct functions;
};

}
}

#endif

