// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Server_hpp
#define anna_comm_Server_hpp

#include <vector>
#include <algorithm>

#include <anna/config/defines.hpp>

#include <anna/comm/Resource.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Host;
class TransportFactory;
class Service;
class Message;
class INetAddress;
class ClientSocket;
class ServerAllocator;
class ReceiverFactory;

/**
   Clase que modela los procesos servidores. Cada maquina (ver Host) contiene un numero indeterminado de
   procesos servidores a los que puede enviar peticiones, bien directamente, o bien a traves del
   sistema de reparto de carga.

   Para facilitar el diseo de soluciones HA cada servidor puede tener un numero indeterminado por donde
   recibe/envia peticiones, es por esto que cada servidor puede tener asociado un numero indeterminado
   de instancias de la clase RemoteConnection.

   La instanciacion de procesos servidores se hara mediante la invocacion al metodo Host::createServer.

   \warning Este metodo no hace que nuestro proceso se convierta en un servidor, sino que conecta
   nuestra aplicacion a un servidor remoto.

   @see Host::createServer
   @see Service
*/
class Server : public Resource {
public:
  typedef std::vector <Service*> Services;
  typedef Services::iterator iterator;
  typedef Services::const_iterator const_iterator;

  /**
     Destructor.
  */
  virtual ~Server();

  /**
     Devuelve la instancia del Host indicada en el constructor.
     \return la instancia del Host indicada en el constructor.
  */
  const Host* getHost() const throw() { return &a_host; }

  /**
     Devuelve la instancia de ClientSocket asociada a este servidor. Puede ser NULL.
     \return la instancia de ClientSocket asociada a este servidor.
  */
  const ClientSocket* getClientSocket() const throw() { return a_clientSocket; }

  /**
     Devuelve el puerto remoto donde establece las conexiones este proceso servidor.
     \return El puerto remoto donde establece las conexiones este proceso servidor.
  */
  int getRemotePort() const throw() { return a_remotePort; }

  /**
     Devuelve el estado del indicador de recuperacin automatica. En caso de perder la conexin, por defecto,
     siempre se intentara reconectar con el servidor.
     \return El estado del indicador de recuperacin automatica.
  */
  bool autoRecovery() const throw() { return a_autoRecovery; }

  /**
     Configura el estado del indicador de recuperacin automatica. En caso de perder la conexin, por defecto,
     siempre se intentara reconectar con el servidor.
     \param autoRecovery Indicador de recuperacin automatica.
  */
  void setAutoRecovery(bool autoRecovery = true) throw();

  /**
     Devuelve el estado activo o inactivo de este proceso servidor. Un proceso servidor estara
     activo si ha conseguido establecer el socket con el proceso remoto que representa por alguno
     de los sockets cliente establecidos.

     @return \em true si el proceso servidor esta preparado para enviar peticiones o \em false
     en otro caso.
  */
  bool isAvailable() const throw(RuntimeException);

  /**
     Devuelve la factoria de transporte indicada en el constructor.
     \return la factoria de transporte indicada en el constructor.
  */
  TransportFactory* getTransportFactory() throw() { return a_transportFactory; }

  /**
     Devuelve el numero maximo de milisegundos esperados para obtener conexion al invocar
     al metodo #connect.
     \return el numero maximo de milisegundos esperados para obtener conexion al invocar
     al metodo #connect.
  */
  const Millisecond &getMaxConnectionDelay() const throw() { return a_msMaxConnectionDelay; }

  /**
     Devuelve el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.
     \return Devuelve el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.
  */
  const Millisecond &getMaxWriteDelay() const throw() { return a_msMaxWriteDelay; }

  /**
     Devuelve la factoria de receptores usada por este servidor.
     \return la factoria de receptores usada por este servidor.
  */
  ReceiverFactory* getReceiverFactory() throw() { return a_receiverFactory; }

  /**
     Establece el numero maximo de milisegundos esperados para obtener la conexion al
     invocar al metodo #connect.
     \param msMaxConnectionDelay Numero de milisegundos esperados para obtener conexion.
  */
  void setMaxConnectionDelay(const Millisecond &msMaxConnectionDelay)
  throw() {
    a_msMaxConnectionDelay = msMaxConnectionDelay;
  }

  /**
     Establece el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.

     \param msMaxWriteDelay Numero de milisegundos esperados en caso de que el buffer del socket se llene.
  */
  void setMaxWriteDelay(const Millisecond &msMaxWriteDelay) throw() { a_msMaxWriteDelay = msMaxWriteDelay; }

  /**
     Establece la factoria de receptores usada por este socket.
     \param receiverFactory Factoria de receptores desde la que obtener el receptor asociado al
     comm::ClientSocket usado por este servidor.
  */
  void setReceiverFactory(ReceiverFactory& receiverFactory) throw();

  /**
   * Devuelve \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   * \return \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   */
  bool getIgnoreIncomingMessages() const throw() { return a_ignoreIncomingMessages; }

  /**
   * Establece el indicador que provoca ignorar los mensajes entrantes.
   * \param ignoreIncomingMessages \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   */
  void setIgnoreIncomingMessages(const bool ignoreIncomingMessages) throw() { a_ignoreIncomingMessages = ignoreIncomingMessages; }

  /**
     Asocia este servidor con un servicio de reparto.
     \param service Servicio de reparto al que vamos a relacionar este servicio.
     \warning se invoca automatica desde
  */
  void attach(Service* service)
  throw(RuntimeException) {
    if(std::find(begin(), end(), service) == end())
      a_services.push_back(service);
  }

  /**
     Crea una conexion al servidor mediante algunas de las conexiones que deberian estar
     disponibles en la maquina asociada a este servidor.
  */
  void connect() throw(RuntimeException);

  /**
     Envia el mensaje recibido como parametro. El bloque de datos recibido se codifica segun las
     reglas establecidas por el protocolo asociado en el contructor.

     \param message Mensaje vamos codificar para enviar a la capa de transporte.

     \return La instancia del ClientSocket usada para enviar el mensaje.
  */
  ClientSocket* send(Message& message) throw(RuntimeException);

  /**
     Envia el mensaje recibido como parametro. El bloque de datos recibido se codifica segun las
     reglas establecidas por el protocolo asociado en el contructor.

     \param message Mensaje vamos codificar para enviar a la capa de transporte.

     \return La instancia del ClientSocket usada para enviar el mensaje.
  */
  ClientSocket* send(Message* message) throw(RuntimeException);

  /**
     Libera el RemoteConnection asociado a esta instancia. Se invoca automaticamente
     cuando el extremo remoto cierra la conexion.
  */
  void reset() throw(RuntimeException);

  /**
     Devuelve un iterador al comienzo de la lista de RemoteConnections asociados a este proceso servidor.
     \return Un iterador al comienzo de la lista de RemoteConnections asociados a este proceso servidor.
  */
  const_iterator begin() const throw() { return a_services.begin(); }

  /**
     Devuelve un iterador al final de la lista de RemoteConnections asociados a este proceso servidor.
     \return Un iterador al final de la lista de RemoteConnections asociados a este proceso servidor.
  */
  const_iterator end() const throw() { return a_services.end(); }

  /**
     Devuelve un iterador al comienzo de la lista de RemoteConnections asociados a este proceso servidor.
     \return Un iterador al comienzo de la lista de RemoteConnections asociados a este proceso servidor.
  */
  iterator begin() throw() { return a_services.begin(); }

  /**
     Devuelve un iterador al final de la lista de RemoteConnections asociados a este proceso servidor.
     \return Un iterador al final de la lista de RemoteConnections asociados a este proceso servidor.
  */
  iterator end() throw() { return a_services.end(); }

  /**
     Devuelve una cadena con la informacion referente a este proceso servidor.
     @return Una cadena con la informacion referente a este proceso servidor.
  */
  std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacion referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacion.
     \return Un nodo XML con la informacion referente a este objeto.
  */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

  /**
     Devuelve la instancia del RemoteConnection sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia del RemoteConnection sobre el que esta posicionado el iterador recibido
  */
  static Service* service(iterator& ii) throw() { return *ii; }

  /**
     Devuelve la instancia del RemoteConnection sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia del RemoteConnection sobre el que esta posicionado el iterador recibido
  */
  static const Service* service(const_iterator& ii) throw() { return *ii; }

  /**
     Devuelve el nombre logico de esta clase.
     \return el nombre logico de esta clase.
  */
  static const char* className() throw() { return "anna::comm::Server"; }

protected:
  /**
     Constructor.

     \param name Nombre logico del servidor.
     \param host Instancia de la maquina sobre la que esta atendiento peticiones.
     \param remotePort Puerto sobre el que atiende peticiones.
     \param autoRecovery Indica si en caso de caida se debe intenrar la recuperacion
     automatica de la conexion.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
  */
  Server(const std::string& name, const Host& host, const int remotePort, const bool autoRecovery, TransportFactory* transportFactory);

private:
  const Host& a_host;
  const int a_remotePort;
  const bool a_autoRecovery;
  Services a_services;
  ClientSocket* a_clientSocket;
  TransportFactory* a_transportFactory;
  Millisecond a_msMaxConnectionDelay;
  Millisecond a_msMaxWriteDelay;
  ReceiverFactory* a_receiverFactory;
  bool a_ignoreIncomingMessages;
  int a_sequence;

  virtual ClientSocket* allocateClientSocket(const INetAddress&, TransportFactory*) const throw();

  friend class Host;
  friend class ServerAllocator;
};

}
}

#endif
