// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Socket_hpp
#define anna_comm_Socket_hpp

#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <anna/config/defines.hpp>
#include <anna/core/mt/Mutex.hpp>

#include <anna/comm/INetAddress.hpp>
#include <anna/comm/AccessPoint.hpp>

namespace anna {

class DataBlock;

namespace xml {
class Node;
}

namespace comm {

class TransportFactory;
class ReceiverFactory;

/**
   Esta clase es la superclase de la que heredan todos los socket definidos en este paquete. Es usada tanto para
   crear socket de la parte cliente y de la parte servidora.
*/
class Socket : public Mutex  {
public:
  /**
     Tipos de dominios soportados.
  */
  struct Domain { enum _v {  Unix = PF_UNIX, Inet = PF_INET } ;  };

  /**
     Tipos de sockets soportados.
  */
  struct Type { enum _v {  Stream = SOCK_STREAM, Datagram = SOCK_DGRAM, Raw = SOCK_RAW } ;  };

  /**
     Tipos de notificaciones que nos puede indicar el Socket.
     \see Socket
  */
  struct Notify {
    enum _v {
      None,          /**< No hay actividad en el socket */
      ReceiveData,   /**< Datos recibidos */
      Close,         /**< El extremo remoto ha cerrado el socket */
      Corrupt        /**< El mensaje recibido no ha sido reconocido por la capa de transporte */
    };
  };

  /**
     Destructor.
  */
  virtual ~Socket();

  /**
     Devuelve el descriptor de fichero asociado a este socket.
     @return El descriptor de fichero asociado a este socke. Si el socket no ha sido creado devolver�-1.
  */
  int getfd() const throw() { return a_fd; }

  /**
     Devuelve el tipo de socket.
     \return El tipo de socket.
  */
  Type::_v getType() const throw() { return a_type; }

  /**
     Devuelve el dominio de este socket.
     \return El dominio de este socket.
  */
  Domain::_v getDomain() const throw() { return a_domain; }

  /**
     Devuelve la categoria asociada a este socket.
     \return La categoria asociada a este socket.
  */
  int getCategory() const throw() { return a_category; }

  /**
     Informa sobre si el socket es capaz de procesar un determinado protocolo de transporte.
     \param transportClassName Normalmente se pasara el resultado del metodo \em className de
     alguno de los protocolos definidos, por ejemplo, Transport::className o
     LiteTransport::className
     \return \em true si soporta el nombre de protocolo recibido como parametro o \em false
     en otro caso.
  */
  bool support(const char* transportClassName) const throw();

  /**
     Devuelve el estado de la conexion de este socket.

     @return \em true si el servidor de socket ha sido conectado o \em false en otro caso.
  */
  bool isBound() const throw() { return a_isBound; }

  /**
     Devuelve el estado del socket.
     @return \em true si el socket este abierto o \em false en otro caso.
  */
  bool isOpened() const throw() { return a_fd != -1; }

  /**
   * Devuelve \em false si el socket usa un protocolo de comunicaciones seguro o \em false
   * en otro caso.
   * \return \em false si el socket usa un protocolo de comunicaciones seguro o \em false
   * en otro caso.
   */
  virtual bool isSecure() const throw() { return false; }

  /**
     Devuelve la direccion local del socket.
     \return La direccion local del socket.
  */
  const AccessPoint& getLocalAccessPoint() const throw() { return a_localAccessPoint; }

  /**
     Devuelve la factoria de la capa de transporte usada en este socket.
     \return la factoria de la capa de transporte usada en este socket.
  */
  TransportFactory* getTransportFactory() const throw() { return a_transportFactory; }

  /**
     Devuelve la factoria de receptores usada en este socket.
     \return la factoria de receptores usada en este socket.
  */
  ReceiverFactory* getReceiverFactory() throw() { return a_receiverFactory; }

  /**
     Activa o desactiva el modo de bloqueo.
     \param blockingMode \em true si queremos activar el bloqueo o \em false en otro caso.
     @return El modo de bloqueo establecido antes de invocar este masodo.
  */
  bool setBlockingMode(const bool blockingMode) throw(RuntimeException);

  /**
     Activa o desactiva el modo de reuso de la direccion.
     \param reuseMode \em true si queremos activar el reuso o \em false en otro caso.
     \warning Solo servira para acelerar el uso del socket una vez que el proceso que lo tenia
     halla dejado de funcionar.
     \return El modo de reuso establecido antes de invocar a este metodo.
  */
  bool setReuseMode(const bool reuseMode) throw(RuntimeException);

  /**
     Establece la capa de transporte usada en este socket.
     \warning Exclusivamente uso interno.
  */
  void setTransportFactory(TransportFactory* transportFactory)  throw() { a_transportFactory = transportFactory; }

  /**
     Establece la factoria de receptores usada por este socket.
     \param receiverFactory Factoria de receptores desde la que obtener el receptor asociado a este Socket.
  */
  void setReceiverFactory(ReceiverFactory& receiverFactory)  throw() { a_receiverFactory = &receiverFactory; }

  /**
     Establece la categoria de este socket.
     La categoria es una concepto del ambito de la aplicacion que el nucleo de anna.comm
     no usa para nada. Unicamente hay que tener en cuenta que todos los anna::comm::ClientSocket creados
     a partir de un anna::comm::ServerSocket comparten su misma categoria.
     \param category Categoria asociada a este socket.
  */
  void setCategory(const int category) throw() { a_category = category; }

  /**
     Cierra este socket. Si el socket no ha sido creado no tendra ningn efecto.
  */
  void close() throw();

  /**
     Intenta la asociar este socket con los parametros indicados en el constructor.
  */
  virtual void bind() throw(RuntimeException);

  /**
     Devuelve una cadena con la informacion referente a este socket.
     @return Una cadena con la informacion referente a este socket.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacion referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacion.
     \return Un nodo XML con la informacion referente a este objeto.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

protected:
  const Domain::_v  a_domain;
  const Type::_v a_type;
  int a_fd;
  AccessPoint a_localAccessPoint;
  bool a_isBound;
  TransportFactory* a_transportFactory;
  ReceiverFactory* a_receiverFactory;
  int a_category;

  /**
     Crea un servidor de socket liberado.
     \param domain Dominio del socket.
     \param type Tipo de socket.
     \param transportFactory Factoria de protocolos de transporte a usar por este sockets. Si se indica
     NULL ser usara el protocolo devuelto por anna::comm::Application::getDefaultTransportFactory.
     \warning La factoria de protocolos debe estar disponible mientras el Socket este activo.
  */
  Socket(const Domain::_v domain, const Type::_v type, TransportFactory* transportFactory = NULL);

  /**
     Crea un socket INET que sera asociado a la direccion y puerto locales indicados.

     \param localAddress Puede ser usado para limitar la direccion por la que atendiende peticiones un servidor de socket
     instalado en una maquina con mas de una direccion.
     \param type Tipo de socket.
     \param transportFactory Factoria de protocolos de transporte a usar por este sockets. Si se indica
     NULL ser usara el protocolo devuelto por anna::comm::Application::getDefaultTransportFactory.
     \warning La factoria de protocolos debe estar disponible mientras el Socket este activo.
  */
  Socket(const INetAddress& localAddress,  const Type::_v type, TransportFactory* transportFactory = NULL);

  /**
     Crea un socket UNIX que sera asociado al archivo indicado como parametro.

     \param path Ruta del archivo que vamos a usar para transferir datos a traves de este socket.
     \param type Tipo de socket.
     \param transportFactory Factoria de protocolos de transporte a usar por este sockets. Si se indica
     NULL ser usara el protocolo devuelto por anna::comm::Application::getDefaultTransportFactory.
  */
  Socket(const std::string& path, const Type::_v type, TransportFactory* transportFactory = NULL);

  /**
     Abre el socket.
  */
  void open() throw(RuntimeException);

  /**
     Cierra este socket. Si el socket no ha sido creado no tendra ningn efecto.
  */
  virtual void do_close() throw() { ::close(a_fd); }

  /**
     Asocia este Socket a la direccion recibida como parametro.
     \warning Exclusivamente uso interno.
  */
  virtual int do_bind(const struct sockaddr*, const int) throw(RuntimeException);

  /**
     Devuelve la cadena correspondiente a la notificacion recibida como parametro.
     \param v Codigo de notificacion.
     \return La cadena correspondiente a la notificacion recibida como parametro.
  */
  static const char* asText(const Notify::_v v) throw();

private:
  bool a_reuseMode;
};

#define anna_socket_assert(a,b) \
   if ((a)) { \
      std::string msg (asString ()); \
      msg += " | "; \
      msg += b; \
      throw RuntimeException (msg, __FILE__, __LINE__); \
   }

#define anna_comm_socket_check(a,b) \
   if ((a) < 0) { \
      const int xerrno = errno; \
      std::string msg (asString ()); \
      msg += " | "; \
      msg += b; \
      throw RuntimeException (msg, errno, ANNA_FILE_LOCATION); \
   }
}
}


#endif



