// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Transport_hpp
#define anna_comm_Transport_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/DataBlock.hpp>

namespace anna {

namespace comm {

class Communicator;
class Message;

/**
   Clase generica para definir la capa de transporte de cualquier protocolo de comunicaciones.

   Estructura basica que nos permite ordenar el proceso de analizar un mensaje recibido desde
   cualquier medio. El protocolo conoce los detalles semanticos del mensaje que ha recibido,
   es decir, conoce como interpretar cada uno de los bytes que componen el mensaje, cuando un
   mensaje esta completo.

   El principal problema de cualquier protocolo externo a la hora de recibir es conocer cual
   es el tamao de un determinado mensaje.

   Todos los metodos que se deberian reescribir en las clases heredadas se invocan desde un
   metodo MT-safe que se encarga de evitar accesos simultaneos desde varios threads, lo cual,
   evita que tengamos que preocuparnos por establecer secciones criticas en cada uno de los
   metodos reescritos.

   \warning Los supuestos bajo los que se dise ste protocolo facilitan el desarrollo de
   clases que ofrecen un gran rendimiento, pero imposibilitan el desarrollo del sistema de
   re-sincronizacin en caso de que alguno de los mensajes no cumpla los supuestos.
   Es decir, si nos llega un mensaje errneo nuestro proceso no sera capaz de volver a
   sincronizarse nunca mas.
*/
class Transport {
public:
  /**
    Maximum number of bytes kept by each ClientSocket without identifying
    a message for the own protocol.
  */
  WHEN_SINGLETHREAD(
    static const int DefaultOverQuotaSize = 2048;
  )
  WHEN_MULTITHREAD(
    static const int DefaultOverQuotaSize = 8192;
  )

  /**
    Returns true if the transport layer has a timming control system activated.
  */
  bool enableTimeout() const throw() { return a_enableTimeout; }

  /**
    Activates the timming control system for the ClientSocket which were created
    through this transport layer. They will be automatically closed if no activity
    is detected in a time interval.
    \see Communicator::setTimeout.
  */
  void activateTimeout() throw() { a_enableTimeout = true; }

  /**
    Deactivates the timming control system for the ClientSocket which were created
    through this transport layer.
  */
  void deactivateTimeout() throw() { a_enableTimeout = false; }


  // Internal use: returns associated input message
  Message* getInputMessage() throw(RuntimeException) {
    return (a_inputMessage == NULL) ? nullInputMessage() : a_inputMessage;
  }

  /**
    Returns the number of bytes reserved by this protocol for the intermediate buffer.
    @return number of bytes reserved by this protocol for the intermediate buffer.
  */
  int getOverQuotaSize() const throw() { return a_overQuotaSize; }

  /**
    Establece el numero de bytes que puede mantener este procotolo para cada uno de los
    ClientSocket sin que se halla identificado el mensaje como propio del protocolo.
    Si el numero de bytes guardados en la memoria intermedia sobrepasa este numero de
    bytes se cerrara la conexion con el ClientSocket.

    \param overQuotaSize Numero de maximo de bytes que podemos mantener en la memoria intermedia.
  */
  void setOverQuotaSize(const int overQuotaSize) throw() { a_overQuotaSize = (overQuotaSize >= MinOverQuotaSize) ? overQuotaSize : MinOverQuotaSize; }

  /**
     Debe calcular el tamao previsto del mensaje actual.

     Si se detecta una anomalia irrecuperable en el mensaje debe devolver una excepcion
     para indicar el error.

     @param dataBlock Bloque con la parte del mensaje disponible hasta el momento.

     @return Si con la informacion disponible no puede establecer la longitud del
     mensaje devolvera -1 en otro caso devolvera la longitud prevista del mensaje.

     \warning
        Si el protocolo de transporte implementado detecta problemas al calcular la
        longitud del mensaje recibido y lanza una excepcion en este metodo el ClientSocket
        activara los sistemas de recuperacion, si es posible.
  */
  virtual int calculeSize(const DataBlock& dataBlock) throw(RuntimeException) = 0;

  /**
     Debe establecer el modo en que el protocolo va a verificar que el mensaje obtenido
     coincide con el patrn esperado e interpretar el contenido del mensaje.
     Este metodo slo se invoca cuando se considera que el mensaje actual esta completo.

     Si se detecta una anomalia irrecuperable en el mensaje debe devolver una excepcion
     para indicar el error.

     @param message Bloque con lo que hasta el momento se considera el ltimo mensaje recibido
     por completo.

     \return Un bloque de memoria que contiene el mensaje recibido codificado segn las reglas del
     protocolo este protocolo de transporte
  */
  virtual const Message* decode(const DataBlock& message) throw(RuntimeException) = 0;

  /**
     Debe establecer la forma en el protocolo va a preparar el envio a la capa de transporte.

     @param message Bloque de datos con la codificacin obtenida mediante cualquiera de los
     codec disponibles (Ver @ref Codec).

     @return El bloque de memoria con el mensaje que sera enviado a la capa de transporte.

     \warning De no indicarse ninguna otra implementacin devolvera el mensaje tal y como
     sea recibido.
  */
  virtual const DataBlock& code(Message& message) throw(RuntimeException) = 0;

  /**
     Metodo que inicializa el estado de esta capa de transporte. Sera invocado automaticamente por el
     nucleo anna.comm.
  */
  virtual void clear() throw() { a_forCode.clear(); }

protected:
  DataBlock a_forCode; /**< Bloque de memoria usado para guardar el contenido de la codificacion */

  /**
     Constructor.
     \param autoSynchronize Indica si el el protocolo instancia permite la sincronizacion automatica.
     \param overQuotaSize Longitud maxima que puede contener el buffer intermedio antes de cerrar el socket
      por considerar que no puede sincronizarlo.
  */
  Transport() :
    a_inputMessage(NULL),
    a_forCode(true),
    a_enableTimeout(false) {
    a_overQuotaSize = DefaultOverQuotaSize;
  }

  /**
     Establece la instancia del mensaje asociada a este transporte.
     \param inputMessage Instancia del mensaje a asociar.
  */
  void setInputMessage(Message* inputMessage) throw() { a_inputMessage = inputMessage; }

private:
  static const int MinOverQuotaSize = 512;

  int a_overQuotaSize;
  Message* a_inputMessage;
  bool a_enableTimeout;

  static Message* nullInputMessage() throw(RuntimeException);
};

}
}

#endif
