// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_Semaphore_hpp
#define anna_core_mt_Semaphore_hpp

#include <semaphore.h>

#include <anna/core/functions.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Safe.hpp>

namespace anna {

//class RuntimeException;

/**
   Clase para facilitar el uso de semaforos.

   Esta facilidad que nos ofrece el sistema operativo se suele usar para limitar el numero de
   accesos a un determinado recurso.

   La funcionalidad de esta clase solo estara disponible en aplicaciones multithread.
*/
class Semaphore : public Safe {
public:
  // Constructor
  /**
     Constructor.

     @param count Numero de accesos simultaneos que permitimos sobre el recurso que queremos
     compartir.
     \param mode Modo de actuacion de esta instancia en modo ST cuando montamos una
     seccion critca sobre este objeto. En modo MT siempre sera Mode::Normal, es decir
     abre una seccion critica sobre este objeto, que bloqueara a cualquier otro thread
     que intente acceder a el.
  */
  Semaphore(unsigned int count);

  /**
     Destructor.
  */
  ~Semaphore();

  /**
     Notifica que desea acceder al recurso. Si se ha alcanzado la cuenta maxima de accesos
     simultaneos el thread que invoca a este metodo quedara bloqueado mientras que algun
     otro threads no invoque al metodo #signal.

     Cada llamada a este metodo debe tener su correspondiente llamada a #signal.
  */
  void wait() throw(RuntimeException);

  /**
     Comprueba si se ha alcanzado la cuenta maxima de accesos simultaneos y devolvera \em
     true si todavia es posible acceder sin quedar bloqueado o \em false en otro caso.
  */
  bool tryWait() throw(RuntimeException);

  /**
     Decrementa la cuenta de utilizacion del recurso, con lo que algunos de los threads
     que puede haber esperando se desbloqueara.
  */
  void signal() throw(RuntimeException);

  /**
     Devuelve una cadena con informacion relevante de esta instancia.
     \return Una cadena con informacion relevante de esta instancia.
  */
  virtual std::string asString() const throw();

protected:
  /**
    Marca el inicio del acceso seguro a este objeto.
    \warning Cada invocacion a este metodo debe tener su correspondiente llamada al metodo #unlock.
    Es muy aconsejable delegar las llamadas a estos metodos en una instancia Guard.
    \return Siempre devuelve \em true;
  */
  void lock() throw(RuntimeException) { wait(); }

  /**
    Indica el final del acceso seguro iniciado al invocar #lock.
    \warning Indica el final del acceso seguro iniciado al invocar #lock.
    Es muy aconsejable delegar las llamadas a estos metodos en una instancia Guard.
    \return Siempre devuelve \em true;
  */
  void unlock() throw() { signal(); }

private:
  sem_t a_id;
};

} //namespace anna

#endif
