// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_MultiMap_hpp
#define anna_core_util_MultiMap_hpp

// Solaris std::multimap is missing !!

#include <algorithm>
#include <functional>
#include <vector>

#include <anna/core/RuntimeException.hpp>
#include <anna/config/defines.hpp>

namespace anna {

/**
   Patron para ordenar instancias de objetos en base de una clave. El uso de este patrn est aconsejado cuando
   tenemos que gestionar cientos o miles de objetos. La eficiencia de los borrado y las inserciones es O(log N).

   \param T Clase del patron.
   \param SortBy Clase que ofrece el valor por el que ordenar. Debe implementar un metodo constante con la
    signatura: TKey value (const T*)
   \param TKey Tipo de clave usado para calcular la ordenacion. Debe implementar los operadores '=', '<' y '==' y el
   contructor copia.
*/
template < typename T, typename SortBy, typename TKey = int > class MultiMap : public std::vector <T*> {
  struct LessT : public std::binary_function <T*, T*, bool> {
    bool operator()(const T* first, const T* second) const throw() {
      return SortBy::value(first) < SortBy::value(second);
    }
  };

  // El orden de los operandos est impuesto por cmo se invoca al operator() desde stl_algo.h (2685).
  struct LessKey : public std::binary_function <T*, TKey, bool> {
    bool operator()(const T* _vv, const TKey& searched) const throw() {
      return SortBy::value(_vv) < searched;
    }
  };

public:
  typedef typename std::vector <T*> container;
  typedef typename container::iterator iterator;
  typedef typename container::const_iterator const_iterator;
  typedef typename container::value_type value_type;

  /**
     Constructor.
  */
  MultiMap() {;}

  /**
     Constructor copia.
     \param other Instancia de la que copiar.
  */
  explicit MultiMap(const MultiMap& other) : container(other) {}

  /**
     Destructor.
  */
  virtual ~MultiMap() { this->clear(); }

  /**
     Incorpora la instancia recibida como parametro en la lista ordenada de objetos.
     \param _vv Instancia a guardar en el map. Si es NULL la operacion no tendra ningun efecto.
     \return \em true si ha registrado la nueva instancia o \em false en otro caso.
  */
  bool add(T* _vv)
  throw(RuntimeException) {
    if(_vv == NULL)
      return false;

    iterator maxii = this->end();
    iterator ii = std::upper_bound(this->begin(), maxii, _vv, a_lessT);

    if(ii == maxii)
      this->push_back(_vv);
    else
      this->insert(ii, _vv);

    return true;
  }

  /**
     Devolvera \em true si la instancia recibida como parametro esta contenido en el
     map o \em en otro caso. Si la instancia recibida es NULL siempre devolvera \em false.
     \param _vv Instancia a comprobar.
     \return \em true si la instancia recibida como parametro esta contenido en el
     map o \em en otro caso.
  */
  bool contains(const T* _vv) const throw() { return (_vv == NULL) ? false : (find(SortBy::value(_vv)) != NULL); }

  /**
   * Borra las entradas contenidas en el vector ordenado y libera la memoria asociada a las mismas.
   */
  void clearAndDestroy() throw() {
    for(iterator ii = this->begin(), maxii = this->end(); ii != maxii; ii ++)
      delete data(ii);

    this->clear();
  }

  /**
     Elimina la instancia recibida como parametro de la lista ordenada de objetos, pero no se libera su memoria.
     \param _vv Instancia a guardar en el map. Si es NULL la operacion no tendra ningun efecto.
     \return La instancia correspondiente a la entrada borrada o NULL si no se encontr ninguna entrada vlida.
  */
  T* erase(const T* _vv)
  throw(RuntimeException) {
    if(_vv == NULL)
      return NULL;

    iterator maxii = this->end();
    iterator ii = std::lower_bound(this->begin(), this->end(), _vv, a_lessT);

    if(ii == maxii)
      return NULL;

    T* dd = data(ii);

    if(!(SortBy::value(_vv) == SortBy::value(dd)))
      return NULL;

    container::erase(ii);
    return dd;
  }

  /**
   * Borra la entrada asociada al iterador recibido. El iterador recibido deja de ser vlido.
   * \param ii Posicin a borrar
   * \return La nueva posicin a la que debera pasar el iterador una vez borrada la entrada.
   *
   */
  iterator erase_iterator(iterator ii) throw() { return container::erase(ii); }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el map.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  T* find(const TKey& key)
  throw() {
    iterator maxii = this->end();
    iterator ii = lower_bound(this->begin(), maxii, key, a_lessKey);

    if(ii == maxii)
      return NULL;

    T* pos = data(ii);
    return (SortBy::value(pos) == key) ? pos : NULL;
  }

  /**
   * Devuelve el iterador a la instancia asociada a la clave recibida.
   * \return el iterador a la instancia asociada a la clave recibida.
   */
  iterator find_iterator(const TKey& key)
  throw() {
    return lower_bound(this->begin(), this->end(), key, a_lessKey);
  }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el map.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  const T* find(const TKey& key) const throw() {
    return const_cast <MultiMap <T, SortBy, TKey>*>(this)->find(key);
  }

  /**
     Devuelve el objeto referenciado por el iterador.
     \return El objeto referenciado por el iterador.
  */
  static T* data(iterator ii) throw() { return *ii; }

  /**
     Devuelve el objeto referenciado por el iterador.
     \return El objeto referenciado por el iterador.
  */
  static const T* data(const_iterator ii) throw() { return *ii; }

private:
  LessT a_lessT;
  LessKey a_lessKey;
};

}

#endif

