// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_Engine_hpp
#define anna_diameter_comm_Engine_hpp


// STL
#include <map>
#include <vector>
#include <string>
#include <algorithm>

#include <anna/app/Component.hpp>
#include <anna/core/util/Recycler.hpp>

#include <anna/diameter/codec/Engine.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/ServerSession.hpp>
#include <anna/config/defines.hpp>
#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/ServerSession.hpp>

// Standard
#include <time.h>


//------------------------------------------------------------------------------
//---------------------------------------------------------------------- #define
//------------------------------------------------------------------------------

namespace anna {
class DataBlock;
class Millisecond;
}



namespace anna {

namespace diameter {

namespace codec {
class Engine;
}

namespace stack {
class Dictionary;
}

namespace comm {

class Response;
class Entity;
class Server;
class LocalServer;

/**
 *  General manager for connections to several diameter servers and from diameter clients.
 *
 *  Optimizes creation, finding and releasing of established client-sessions to a certain number of
 *  diameter servers through entities.
 *  Optimizes creation, finding and releasing of established server-sessions from a certain number of
 *  diameter clients through local servers.
 *
 *  Implementation example:
 *
 *  \code
 *
 *     class MyEngine : public diameter::comm::Engine {
 *     public:
 *        MyEngine () {;}
 *
 *     private:
 *        anna::Recycler<MyEntity> a_entities;
 *
 *        anna::diameter::comm::Entity* allocateEntity () throw () { return a_entities.create (); }
 *
 *        void releaseEntity (anna::diameter::comm::Entity* entity) throw () {
 *           MyEntity* aux = static_cast <MyEntity*> (entity);
 *           a_entities.release (aux);
 *        }
 *
 *
 *        anna::diameter::comm::LocalServer* allocateLocalServer () throw () { return a_localServers.create (); }
 *
 *        void releaseLocalServer (anna::diameter::comm::LocalServer* localServer) throw () {
 *           MyLocalServer* aux = static_cast <MyLocalServer*> (localServer);
 *           a_localServers.release (aux);
 *        }
 *     };
 *
 *  \endcode
 */
class Engine : public anna::app::Component {
public:

  /**
     Diameter application node origin realm

     @param originRealmName Used to configure the Origin-Realm for outgoing messages.
     If not configured or empty string provided, host domainname will be set.
  */
  void setOriginRealmName(const std::string & originRealmName) throw();

  /**
     Diameter application origin host

     @param originHostName Used to configure the Origin-Host for outgoing messages.
     If not configured or empty string provided, hostname (system name) will be set.
  */
  void setOriginHostName(const std::string & originHostName) throw();

  /**
     Gets the configured diameter application node origin realm

     @return Diameter application node origin realm
  */
  const std::string & getOriginRealmName() const throw() { return a_originRealm; }

  /**
     Gets the configured diameter application origin host

     @return Diameter application node origin host
  */
  const std::string & getOriginHostName() const throw() { return a_originHost; }


  /**
   * Propagate auto recovery configuration to entities within engine. Recovery period is configured at
   * #anna::comm::Communicator::setRecoveryTime. All the client client-sessions created throught #createEntity,
   * will be created based on the engine auto-recovery value (enable by default). But you could access entities,
   * servers or client-sessions independently to change this behaviour.
   *
   * @param autoRecovery Auto recovery indicator. True by default.
   */
  void raiseAutoRecovery(bool autoRecovery = true) throw(anna::RuntimeException);

  /**
   * Returns automatic bind indicator for client-sessions. By default \em true will be used.
   * \return Value for automatic connection bind.
   */
  bool getAutoBind() const throw() { return a_autoBind; }

  /**
   * Sets automatic connection bind indicator for client-sessions. If not asigned, it will be \em true.
   * \param autoBind Value for automatic connection bind.
   *
   * In order to change bind timer, first client-session must be created without autobind, modify time
   * parameter and then invoking bind.
   */
  void setAutoBind(const bool autoBind) throw() { a_autoBind = autoBind; }

  /**
     Sets the milliseconds wait to achieve a client connection to server by mean connect primitive.
     This is a general value for born client-sessions over engine. Particular configuration could be done
     through #ClientSession::setMaxConnectionDelay.

     \param maxConnectionDelay Milliseconds wait to get connection
  */
  void setMaxConnectionDelay(const anna::Millisecond & maxConnectionDelay) throw() { a_maxConnectionDelay = maxConnectionDelay; }

  /**
     Gets the milliseconds wait to achieve a client connection to server by mean connect primitive.
     Returns the global engine value, but it could be overwritten through each client session (#ClientSession::setMaxConnectionDelay).
     Default value is 'anna::comm::ClientSocket::DefaultMaxConnectionDelay'.

     \return Milliseconds wait to get connection
  */
  const anna::Millisecond & getMaxConnectionDelay() throw() { return a_maxConnectionDelay; }

  /**
   * Binds engine entities.
   *
   * @return Returns true if all client-session were successfully bound
   */
  bool bind() throw(anna::RuntimeException);

  /**
   * Sets CER and DWR diameter messages to be used over created client-sessions.
   * Its recommended to set this global configuration although it is possible to configure each client-session separately.
   *
   * @param cer Capabilities-Exchange-Request message (encoded) for the client-sessions bind.
   * @param dwr Device-Watchdog-Request message (encoded) for the client-sessions keep-alive.
   */
  void setClientCERandDWR(const anna::DataBlock & cer, const anna::DataBlock & dwr) throw(anna::RuntimeException);

  /**
   * Sets CER and DWR diameter messages to be used over created client-sessions. If empty string is provided for CER and/or DWR, default version will be configured.
   * Its recommended to set this global configuration although it is possible to configure each client-session separately.
   *
   * @param cer Capabilities-Exchange-Request xml message path file for the client-sessions bind. If empty string is provided (default), a default version for CER will be encoded.
   * @param dwr Device-Watchdog-Request xml message path file for the client-sessions keep-alive. If empty string is provided (default), a default version for DWR will be encoded.
   */
  void setClientCERandDWR(const std::string & cer = "", const std::string & dwr = "") throw(anna::RuntimeException);

  /**
   * Sets the watchdog period (DWR) for client-sessions.
   * Its recommended to set this global configuration although it is possible to configure each client-session separately.
   *
   * @param wp Watchdog period.
   */
  void setWatchdogPeriod(const anna::Millisecond & wp) throw(anna::RuntimeException);

  /**
   * Gets the number of client-sessions per server.
   * \return numberOfClientSessionsPerServer Number of client-sessions per server.
   */
  int getNumberOfClientSessionsPerServer() const throw() { return a_numberOfClientSessionsPerServer; }

  /**
   * Sets the number of client-sessions per server.
   * Its recommended to set this global configuration although it is possible to configure each client-session separately.
   * \param numberOfClientSessionsPerServer Number of client-sessions per server.
   */
  void setNumberOfClientSessionsPerServer(int numberOfClientSessionsPerServer) throw() { a_numberOfClientSessionsPerServer = numberOfClientSessionsPerServer; }


  /**
   * Returns client-session instance identified by (address, port, socketId) provided.
   *
   * \param addr Diameter server address (ip or hostname).
   * \param port Diameter server port.
   * @param socketId Diameter server socket id.
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The client-session instance identified by (address, port, socketId) provided.
   *
   * \warning If no client-session found, an exception is launched by default.
   */
  ClientSession* findClientSession(const std::string & addr, int port, int socketId, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);


  /**
   * Same as #findClientSession, but providing client session key (<address>:<port>|<socket id>)
   */
  ClientSession* findClientSession(const std::string & key, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);


  /**
   * Returns server instance identified by pair (address, port) provided.
   *
   * \param addr Diameter server address (ip or hostname).
   * \param port Diameter server port.
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The server instance identified by pair (address, port) provided.
   *
   * \warning If no server found, an exception is launched by default.
   */
  Server* findServer(const std::string & addr, int port, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Returns entity instance identified by internal index.
   *
   * \param socketList Diameter entity servers list.
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The entity instance identified by id provided.
   *
   * \warning If no entity found, an exception is launched by default.
   */
  Entity* findEntity(const socket_v & socketList, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Returns entity instance identified by internal index.
   *
   * \param addr1 Diameter primary server address (ip or hostname).
   * \param port1 Diameter primary server port.
   * \param addr2 Diameter secondary server address (ip or hostname).
   * \param port2 Diameter secondary server port.
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The entity instance identified by id provided.
   *
   * \warning If no entity found, an exception is launched by default.
   */
  Entity* findEntity(const std::string & addr1, int port1, const std::string & addr2, int port2, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Creates a diameter entity with provided parameters.
   *
   * Depending on auto-bind configuration, capabilities exchange request will be or won't be performed over the entity client-sessions.
   *
   * \param socketList Diameter server priority list (priority-ordered) in order to define whole entity.
   * @param description Optional entity description (empty by default)
   *
   * \return The entity created or exception when any server (address/port) already exists for another entity.
   *
   * \warning The entity won't be almost operative until a notification by mean 'ClientSession::eventResponse'
   * indicates that 'ClassCode::Bind' has been correctly performed for any included client-session.
   */
  Entity* createEntity(const socket_v & socketList, const std::string & description = "")
  throw(anna::RuntimeException);

  /**
   * Creates a standard (dual) diameter entity with provided parameters.
   *
   * Depending on auto-bind configuration, capabilities exchange request will be or won't be performed over the entity client-sessions.
   *
   * \param addr1 Diameter primary server address (ip or hostname).
   * \param port1 Diameter primary server port.
   * \param addr2 Diameter secondary server address (ip or hostname).
   * \param port2 Diameter secondary server port.
   * @param description Optional entity description (empty by default)
   *
   * \return The standard entity created or exception when any server (address/port) already exists for another entity.
   *
   * \warning The entity won't be almost operative until a notification by mean 'ClientSession::eventResponse'
   * indicates that 'ClassCode::Bind' has been correctly performed for any included client-session.
   */
  Entity* createEntity(const std::string & addr1, int port1, const std::string & addr2, int port2, const std::string & description = "")
  throw(anna::RuntimeException);


  /**
   * Returns local server instance identified by pair (address, port) provided.
   *
   * @param addr Diameter server socket address (ip or hostname).
   * @param port Diameter server socket port.
   * \param emode Action when no local server is found with provided parameters (Throw/Ignore).
   *
   * \return The local server instance identified by pair (address, port) provided.
   *
   * \warning If no local server found, an exception is launched by default.
   */
  LocalServer* findLocalServer(const std::string & addr, int port, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Returns server-session instance identified by INetAddress serialization provided.
   *
   * @param socketId Hash for Client Socket INetAddress serialization
   * \param emode Action when no server-session is found with provided parameters (Throw/Ignore).
   *
   * \return The server-session instance identified by global unique socketId provided.
   *
   * \warning If no server-session found, an exception is launched by default.
   */
  ServerSession* findServerSession(int socketId, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Creates a diameter local server with provided parameters.
   *
   * Server socket address could be an IPv4 or hostname. Default port will be standard 3868 for diameter agents,
   * but any other could be configured. Socket server could be created with any max accepted connections: zero
   * value means temporarily disabled and negative values assume no limit (shared bind) for incomming connections.
   *
   * @param addr Diameter server socket address (ip or hostname).
   * @param port Diameter server socket port (standard 3868 by default).
   * @param maxConnections Diameter server max sessions allowed (no limit by default).
   * @param allowedInactivityTime Max inactivity time for server sessions over the local server before being reset.
   * @param category Optional socket server category (1 by default).
   * @param description Optional socket server description (empty by default).
   *
   * \return The local server created or exception when is already created.
   */
  LocalServer *createLocalServer(const std::string & addr, int port = Session::DefaultPort, int maxConnections = -1, const anna::Millisecond & allowedInactivityTime = ServerSession::DefaultAllowedInactivityTime, int category = 1, const std::string & description = "")
  throw(anna::RuntimeException);


  /**
     Close all the engine resources (entities and local servers)
     Optionally all resources may be freed passing true

     @param destroy Free all engine entity resources
  */
  void close(bool destroy = false) throw(anna::RuntimeException) { closeEntities(destroy); closeLocalServers(destroy); }


  /**
     Close all the engine entities (close servers, then close client-sessions within them). Depending on client-session configuration
     ('OnDisconnect' behaviour), pending answers will be wait (graceful) or ignored (immediate-abrupt close).
     Optionally all entities resources may be freed passing true; in this case, close is immediately performed:
     @param destroy Free all engine entity resources
  */
  void closeEntities(bool destroy = false) throw(anna::RuntimeException);


  /**
   * Close entity servers (then, client-sessions included) and optionally free resources including entity itself.
   * If entity is null, this operation has no effect.
   *
   * \param entity Diameter entity to be closed.
   * \param destroy Deletes entity over the engine and all its resources.
   */
  void closeEntity(Entity* entity, bool destroy = false) throw(anna::RuntimeException);


  /**
     Close all the engine local server sockets including their children server sessions.
     Optionally all local server resources may be freed passing true.

     @param destroy Free all engine local servers resources and server sessions within them.
  */
  void closeLocalServers(bool destroy = false) throw(anna::RuntimeException);

  /**
   * Close local server socket and its children server sessions.
   * This is useful when detecting service lost. When service is ready to handle traffic, a new server socket would
   * be created by mean #LocalServer::enable() and new connections could be accepted.
   * Optionally local server resources may be freed passing true.
   *
   * \param localServer Local server to be closed.
   * \param destroy Deletes local server over engine and all its resources (server sessions within it).
   */
  void closeLocalServer(LocalServer * localServer, bool destroy = false) throw(anna::RuntimeException);

  /**
     Gets the number of requests messages over-the-air for entities.

     @return OTA messages.
  */
  int getOTARequestsForEntities() const throw();

  /**
     Gets the number of requests messages over-the-air for local servers.

     @return OTA messages.
  */
  int getOTARequestsForLocalServers() const throw();

  /**
     Gets the number of requests messages over-the-air for entities plus local servers.

     @return OTA messages.
  */
  int getOTARequests() const throw() { return (getOTARequestsForEntities() + getOTARequestsForLocalServers()); }

  /**
     Returns idle state (no pending answers) for entities.

     @return Idle state.
  */
  bool idleForEntities() const throw() { return (getOTARequestsForEntities() == 0); }

  /**
     Returns idle state (no pending answers).

     @return Idle state.
  */
  bool idleForLocalServers() const throw() { return (getOTARequestsForLocalServers() == 0); }

  /**
     Returns idle state (no pending answers for entities or local servers).

     @return Idle state.
  */
  bool idle() const throw() { return (getOTARequests() == 0); }

  /**
     Sent a message to all the engine entities.
     It is used, i.e., in Disconnect-Peer-Request procedure over the engine.

     \param message Message which is being sent.

     @return Returns true (success) only when broadcast is success over all the engine entities. If any entity fails,
     then false is returned. Broadcast try to send all over the resources in spite of any fail.
  */
  bool broadcastEntities(const Message*message) throw(anna::RuntimeException);
  bool broadcastEntities(const Message& message) throw(anna::RuntimeException) { return broadcastEntities(&message); }

  /**
     Sent a message through all the engine local servers.
     It is used, i.e., in Disconnect-Peer-Request procedure over the engine.

     \param message Message which is being sent.

     @return Returns true (success) only when broadcast is success over all the engine local servers. If any local server fails,
     then false is returned. Broadcast try to send all over the resources in spite of any fail.
  */
  bool broadcastLocalServers(const Message*message) throw(anna::RuntimeException);
  bool broadcastLocalServers(const Message& message) throw(anna::RuntimeException) { return broadcastLocalServers(&message); }

  /**
  * Class string representation
  *
  * @return String with class content
  */
  virtual std::string asString(void) const throw();

  /**
     Class XML representation.
     \param parent XML node over which we will put instance information.
     \return XML documentcon with class content.
  */
  virtual anna::xml::Node* asXML(anna::xml::Node* parent) const throw();


  /**
     When there is not bound server session over the engine, this virtual method will be invoked.
     Applications must decide to do any other tasks at this idle/isolated situation.
     Default implementation do nothing.
  */
  virtual void availabilityLostForLocalServers(Engine *) const throw() {;}

  /**
     When there is any bound server session over the engine, this virtual method will be invoked.
     Applications must decide to do be ready for incoming traffic.
     Default implementation do nothing.
  */
  virtual void availabilityRecoveredForLocalServers(Engine *) const throw() {;}

  /**
     When there is not bound entity over the engine, this virtual method will be invoked.
     Many applications must change communicator status to Unavailable when no engines are available.
     Default implementation do nothing.
  */
  virtual void availabilityLostForEntities(Engine *) const throw() {;}

  /**
     When there is any bound entity over the engine, this virtual method will be invoked.
     Many applications must recover communicator status to Available when any engine are available.
     Default implementation do nothing.
  */
  virtual void availabilityRecoveredForEntities(Engine *) const throw() {;}

  /**
     When there is not bound server-session over the local server, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityLost(LocalServer *) const throw() {;}

  /**
     When there is any bound server-session over the local server, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityRecovered(LocalServer *) const throw() {;}

  /**
     When there is not bound server over the entity, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityLost(Entity *) const throw() {;}

  /**
     When there is any bound server over the entity, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityRecovered(Entity *) const throw() {;}

  /**
     When there is not bound client-session over the server, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityLost(Server *) const throw() {;}

  /**
     When there is any bound client-session over the server, this virtual method will be invoked.
     Default implementation do nothing.
  */
  virtual void availabilityRecovered(Server *) const throw() {;}

  /**
     When a subyacent client session is going to be bound, this method is invoked before.
     Default implementation do nothing.
  */
  virtual void bindingClientSession(const ClientSession *) const throw() {;}

  /**
   * Class user should implement this method in order to define Disconnect-Peer-Answer for last received DPR.
   * Origin-Host and Origin-Realm are configured at comm::Engine with hostname and FQDN (Fully Qualified Domain Name).
   * Default implementation imply DPA with DIAMETER_SUCCESS Result-Code, allowing remote disconnection.
   * Any other implementation is responsible to build a valid DPA diameter message.
   * DPR/DPA procedure is disabled with empty definition of this method: no DPA will be sent when DPR is received.
   *
   * @param dpa DPA datablock passed as reference
   * @param dpr Corresponding DPR received (sequence values must be taken into account in order to build DPA)
  */
  virtual void readDPA(anna::DataBlock &dpa, const anna::DataBlock & dpr) throw();

  /**
   * Sets optional CEA from file, when default is not enough
   *
   * @param &ceaPathfile Path file for the CEA xml message provided
   */
  void setCEA(const std::string &ceaPathfile) throw() { a_ceaPathfile = ceaPathfile; }

  /**
   * Class user should implement this method in order to define Capabilities-Exchange-Answer for received CER over server socket.
   * Origin-Host and Origin-Realm are configured at comm::Engine with hostname and FQDN (Fully Qualified Domain Name).
   * Default implementation imply CEA with DIAMETER_SUCCESS Result-Code, and own domain node parameters, but application should
   * analyze the CER message in order to accept it or not (with apropiate non-success Result-Code).
   * If @setCEA was invoked, a message from file is used instead of default implementation.
   * Any other implementation is responsible to build a valid CEA diameter message, even ignoring a possible cea from file when @setCEA is used:
   *
   * If one peer sends a CER message to another Peer and receiver does not have support for
   *
   *  1) any common application then it must return the CEA with Result-Code Avp set to DIAMETER_NO_COMMON_APPLICATION
   *     and should disconnect the transport layer connection (automatically done by diameter::comm module).
   *  2) no common security mechanism then it must return the CEA with Result-Code Avp set to DIAMETER_NO_COMMON_SECURITY
   *     and should disconnect the transport layer connection (automatically done by diameter::comm module).
   *  3) if CER is received from any unknown peer then receiver should discard the message, or send the CEA with the
   *     Result-Code Avp set to DIAMETER_UNKNOWN_PEER.
   *
   *  If the local implementation policy permits to receive CER from unknown hosts, a successful CEA MAY be returned,
   *  and the life time of the peer entry in PEER-Table is equal to the lifetime of the transport connection.
   *  If in any case transport connection fails then all the pending transactions destined to the unknown peer can be discarded.
   *
   *  The CER and CEA messages MUST NOT be proxied, redirected or relayed. Since CER/CEA messages can not be proxied, but still
   *  it is possible that proxy will receive a CER message and proxy does not have any peer to handle the application requested
   *  in CER, in this case proxy set the E bit in CEA and set the Result-Code Avp to DIAMETER_UNABLE_TO_DELIVER, sends back to
   *  CER generator peer.
   *
   * @param cea CEA datablock passed as reference. Empty cea implies to discard CER received.
   * @param cer Corresponding CER received (sequence values must be taken into account in order to build CEA)
  */
  virtual void readCEA(anna::DataBlock &cea, const anna::DataBlock & cer) throw();

  /**
   * Class user should implement this method in order to define Device-Watchdog-Answer for received DWR over server socket.
   * Origin-Host and Origin-Realm are configured at comm::Engine with hostname and FQDN (Fully Qualified Domain Name).
   * Default implementation imply DWA with DIAMETER_SUCCESS Result-Code, and own domain node parameters.
   * Any other implementation is responsible to build a valid DWA diameter message.
   *
   * @param dwa DWA datablock passed as reference
   * @param dwr Corresponding DWR received (sequence values must be taken into account in order to build DWA)
  */
  virtual void readDWA(anna::DataBlock &dwa, const anna::DataBlock & dwr) throw();

  /**
   * DRA basics: CER information is gathered on every server session managed by the diameter comm engine. You could send the message to a
   * specific realm, and optionally you could restrict a host inside it. This is common for requests (answers are normally sent through
   * the same source server session where the request was received). Exception will be thrown if not found an available server session
   * for the Destination-Realm and/or Destination-Host provided
   *
   * @param destinationRealm If empty, NULL is returned, because is nonsense to specify a host out of realm context
   * @param destinationHost If empty, no restriction is applied within the target realm node. Random delivery is applied for the available server sessions
   *
   * @return transactional response reference, or NULL if answer is sent
   */
  const Response* sendRealmHost(const Message* message, const std::string &destinationRealm, const std::string &destinationHost = "") throw(anna::RuntimeException);

  /**
     Reset engine statistics.
     At the moment, only diameter servers processing time is observed.
  */
  void resetStatistics() throw();


  /**
   * Engine lazy initialization. Used if the engine is created when application is already running; for example
   * on dynamic realms registration. At the moment is not actually needed (nothing is done at initialization),
   * but it is recommended to start the component and set its state as 'running' from the point of view of the
   * application.
   */
  void lazyInitialize() throw(RuntimeException);


protected:
  /**
     Constructor.

     @param className Component class name
     @param baseProtocolDictionary This will be used internally when calling \@readCEA, \@readDPA and \@readDWA on
     servers, and also used during base protocol messages tracing (if debug traces are enabled). You could provide
     NULL, but you must be sure that neither of the former situations are going to happen or an exception will be
     thrown (using setClientCERandDWR with DataBlock arguments, expects externally encoded messages and could help).
     It is recommended to set a base protocol dictionary loading 'source/diameter/stack/setups' dictionaries (for
     example 'avps_ietf.xml' plus 'commands_baseProtocol.xml'), or using the dictionary creation API. The dictionary
     could also be an application stack, the only condition is containing the resources to build base protocol messages.
  */
  Engine(const char *className, const stack::Dictionary *baseProtocolDictionary);

  // INTERNAL CREATORS AND CLOSE METHODS
  Server *createServer(Entity*, const socket_t&) throw(anna::RuntimeException);
  void closeServer(Server*, bool) throw(anna::RuntimeException);
  ClientSession *createClientSession(Server*, int) throw(anna::RuntimeException);
  void closeClientSession(ClientSession*, bool) throw(anna::RuntimeException);

  // INTERNAL ALLOCATORS
  Server* allocateServer() throw();
  void releaseServer(Server*) throw();
  ClientSession* allocateClientSession() throw();
  void releaseClientSession(ClientSession*) throw();


  /**
     Entity allocator method.

     It is recommended to use anna::Recycler for entities creation/releasing.

     \see anna::Recycler
  */
  virtual Entity* allocateEntity() throw() { return NULL; }


  /**
     Invoked to free entities.
     \see anna::Recycler
  */
  virtual void releaseEntity(Entity*) throw() {;}


  /**
     Local server allocator method.

     It is recommended to use anna::Recycler for entities creation/releasing.

     \see anna::Recycler
  */
  virtual LocalServer* allocateLocalServer() throw() { return NULL; }


  /**
     Invoked to free local servers.
     \see anna::Recycler
  */
  virtual void releaseLocalServer(LocalServer*) throw() {;}

  //  Gets the base protocol codec engine used internally.
  //  This engine is initializaed on constructor with the base protocol dictionary.
  //  The reason to not reuse any other codec engine from the application is to have this one isolated with no interference
  //  regarding configuration changes (validation depth/mode, fix mode, etc.).
  //
  //  @return Pointer to the internal base protocol codec engine
  codec::Engine *getBaseProtocolCodecEngine() const throw() { return const_cast<codec::Engine *>(&a_baseProtocolCodecEngine); }

private:

  // Internal use: tracing and readCEA/DPA/DWA
  codec::Engine a_baseProtocolCodecEngine;

  std::string a_originRealm;
  std::string a_originHost;
  bool a_autoBind;
  int a_numberOfClientSessionsPerServer;


  // ClientSessions messages:
  anna::DataBlock a_cer;
  anna::DataBlock a_dwr;
  anna::Millisecond a_watchdogPeriod;

//   // ServerSessions messages:
  std::string a_ceaPathfile; // path file to optional CEA (diameter local server configuration)
//   anna::DataBlock a_cea;
//   anna::DataBlock a_dwa;

  // Client connectivity
  anna::Millisecond a_maxConnectionDelay;


  // Availability
  bool a_availableForEntities; // any of the entities must be bound
  void availabilityLostForEntities() throw();
  void availabilityRecoveredForEntities() throw();
  bool refreshAvailabilityForEntities() throw(); // return true if change

  bool a_availableForLocalServers; // any of the local servers must be bound
  void availabilityLostForLocalServers() throw();
  void availabilityRecoveredForLocalServers() throw();
  bool refreshAvailabilityForLocalServers() throw(); // return true if change

  void eraseDeprecatedIdleEntities() throw();

  // Component:
  void do_initialize() throw(anna::RuntimeException);
  void do_stop() throw();

  // Integrity:
  void checkEntityCollision(const socket_v &) throw(anna::RuntimeException);
  void assertBaseProtocolHealth() throw(anna::RuntimeException); // checks the dictionary


  //////////////////////////
  // CLIENT FUNCTIONALITY //
  //////////////////////////

  typedef std::string clientSession_key; // 'ADDR:PORT|id'
  typedef std::map <clientSession_key, ClientSession*> clientSession_container;
  typedef clientSession_container::value_type clientSession_value_type;
  typedef clientSession_container::iterator clientSession_iterator;
  typedef clientSession_container::const_iterator const_clientSession_iterator;
  clientSession_container a_clientSessions;
  anna::Recycler<ClientSession> a_clientSessionsRecycler;
  clientSession_iterator clientSession_find(const clientSession_key&) throw();
  clientSession_iterator clientSession_begin() throw() { return a_clientSessions.begin(); }
  clientSession_iterator clientSession_end() throw() { return a_clientSessions.end(); }
  static ClientSession* clientSession(clientSession_iterator ii) throw() { return ii->second; }
  const_clientSession_iterator clientSession_begin() const throw() { return a_clientSessions.begin(); }
  const_clientSession_iterator clientSession_end() const throw() { return a_clientSessions.end(); }
  static const ClientSession* clientSession(const_clientSession_iterator ii) throw() { return ii->second; }

  typedef socket_t server_key;
  server_key getServerKey(const std::string & addr, int port) const throw();
  typedef std::map <server_key, Server*> server_container;
  typedef server_container::value_type server_value_type;
  typedef server_container::iterator server_iterator;
  typedef server_container::const_iterator const_server_iterator;
  server_container a_servers;
  anna::Recycler<Server> a_serversRecycler;
  server_iterator server_find(const server_key&) throw();
  server_iterator server_begin() throw() { return a_servers.begin(); }
  server_iterator server_end() throw() { return a_servers.end(); }
  static Server* server(server_iterator ii) throw() { return ii->second; }
  const_server_iterator server_begin() const throw() { return a_servers.begin(); }
  const_server_iterator server_end() const throw() { return a_servers.end(); }
  static const Server* server(const_server_iterator ii) throw() { return ii->second; }

  typedef std::string entity_key; // 'ADDR1:PORT1 ADDR2:PORT2 ADDR3:PORT3 ...'
  entity_key getEntityKey(const socket_v &) const throw();
  entity_key getEntityKey(const std::string & addr1, int port1, const std::string & addr2, int port2) const throw();
  typedef std::map <entity_key, Entity*> entity_container;
  typedef entity_container::value_type entity_value_type;
  typedef entity_container::iterator entity_iterator;
  typedef entity_container::const_iterator const_entity_iterator;
  entity_container a_entities;
  entity_iterator entity_find(const entity_key&) throw();
  entity_iterator entity_begin() throw() { return a_entities.begin(); }
  entity_iterator entity_end() throw() { return a_entities.end(); }
  static Entity* entity(entity_iterator ii) throw() { return ii->second; }
  const_entity_iterator entity_begin() const throw() { return a_entities.begin(); }
  const_entity_iterator entity_end() const throw() { return a_entities.end(); }
  static const Entity* entity(const_entity_iterator ii) throw() { return ii->second; }


  //////////////////////////
  // SERVER FUNCTIONALITY //
  //////////////////////////

  // Local servers
  typedef std::map <socket_t, LocalServer*> localServer_container;
  typedef localServer_container::value_type localServer_value_type;
  typedef localServer_container::iterator localServer_iterator;
  typedef localServer_container::const_iterator const_localServer_iterator;
  localServer_container a_localServers;
  localServer_iterator localServer_find(const socket_t&) throw();
  localServer_iterator localServer_begin() throw() { return a_localServers.begin(); }
  localServer_iterator localServer_end() throw() { return a_localServers.end(); }
  static LocalServer* localServer(localServer_iterator ii) throw() { return ii->second; }
  const_localServer_iterator localServer_begin() const throw() { return a_localServers.begin(); }
  const_localServer_iterator localServer_end() const throw() { return a_localServers.end(); }
  static const LocalServer* localServer(const_localServer_iterator ii) throw() { return ii->second; }

  // Server sessions are managed within LocalServer (not at engine) due to dynamic creation nature
  // Here we maintain the Destination-Realm / Destination-Host maps for DRA basics:
  typedef std::vector<ServerSession*> server_sessions_vector_t;
  typedef server_sessions_vector_t::const_iterator server_sessions_it_t;
  typedef server_sessions_vector_t::iterator server_sessions_nc_it_t;
  typedef std::map <std::string /* Destination-Host */, server_sessions_vector_t> dh_server_sessions_map_t;
  typedef dh_server_sessions_map_t::const_iterator dh_server_sessions_it_t;
  typedef dh_server_sessions_map_t::iterator dh_server_sessions_nc_it_t;
  typedef std::map <std::string /* Destination-Realm */, dh_server_sessions_map_t> dr_dh_server_sessions_map_t;
  typedef dr_dh_server_sessions_map_t::const_iterator dr_dh_server_sessions_it_t;
  typedef dr_dh_server_sessions_map_t::iterator dr_dh_server_sessions_nc_it_t;
  dr_dh_server_sessions_map_t a_dr_dh_server_sessions;
  void manageDrDhServerSession(ServerSession *ss, bool register_or_desregister) throw();

  friend class Session;
  friend class ClientSession;
  friend class ServerSession;
  friend class ServerSocket;
  friend class Server;
  friend class Entity;
  friend class LocalServer;
  //friend class Message;
};

}
}
}

#endif

