// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_Server_hpp
#define anna_diameter_comm_Server_hpp


// STL
#include <string>
#include <vector>
#include <map>

// Project
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/config/defines.hpp>
#include <anna/diameter.comm/ClassCode.hpp>
#include <anna/diameter.comm/MessageStatistics.hpp>




namespace anna {
class DataBlock;
namespace timex {
class Engine;
}
}


namespace anna {

namespace diameter {

namespace comm {

class Engine;
class Entity;
class ClientSession;
class Response;
class Message;


/**
   Diameter server with 1..N connections.
*/
class Server {

  // Parent information
  Entity *a_parent;

  // Main server attributes
  socket_t a_socket;

  // ClientSessions
  std::vector<ClientSession*> a_clientSessions;
  int a_maxClientSessions; // -1 means "no limit to add client-sessions" (sockets per server)
  std::vector<ClientSession*>::iterator a_deliveryIterator;
  ClientSession *a_lastUsedResource;

  // Activity
  anna::Millisecond a_lastIncomingActivityTime;   // last unix timestamp (in milliseconds) when message reception was managed over this server
  anna::Millisecond a_lastOutgoingActivityTime;   // last unix timestamp (in milliseconds) when message sending was managed over this server
  void updateIncomingActivityTime() throw();
  void updateOutgoingActivityTime() throw();

  // Engine
  Engine *a_engine;

  // Statistics
  MessageStatistics a_messageStatistics;
  void initializeStatisticResources() throw();
  void resetStatistics() throw();

  // Availability
  bool a_available; // any of the client-sessions must be bound
  void availabilityLost() throw();
  void availabilityRecovered() throw();
  bool refreshAvailability() throw(); // return true if change
  void assertReady() throw(anna::RuntimeException);
  void initialize() throw();
  void childIdle() const throw();

  // Private close/destroy method
  void close(bool destroy) throw(anna::RuntimeException);


public:


  /**
   * Default constructor.
   * @param maxClientSessions Maximum number of client-sessions managed by the server. Default is 1 (monoconnection server).
   * The value -1, means no limit to add client-sessions.
   */
  Server(int maxClientSessions = 1) : a_maxClientSessions(maxClientSessions) { initialize(); }


  /**
  * Add a server to the entity and create all the client-sessions configured at #setSocketsPerDiameterServer within that server.
  *
  * \param socketId Diameter socket identifier within the server.
  */
  void addClientSession(int socketId) throw(anna::RuntimeException);

  /**
     Set timeout to consider failed a request.
     \param v Requests class code.
     \param millisecond Milliseconds wait before considering the requests failed.

     Timers are internally managed and automatically activated.
  */
  void setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw();

  /**
   * Binds server client-sessions.
   *
   * @return Returns true if all client-session were successfully bound
   */
  bool bind() throw(anna::RuntimeException);

  /**
   * Propagate auto recovery configuration to client-sessions within server
   *
   * @param autoRecovery Auto recovery indicator. True by default.
   */
  void raiseAutoRecovery(bool autoRecovery = true) throw(anna::RuntimeException);


// Sent a message to the server using a certain client-session by mean round-robin between socketId's for
//  multiple client client-sessions functionality. If a specific socketId is provided, then uses such specific client-session.
// Last used delivery resource could be known through #getLastUsedResource().
  bool send(const Message*, int socketId = -1 /* client-sessions round-robin */) throw(anna::RuntimeException);
  bool send(const Message& message, int socketId = -1 /* client-sessions round-robin */) throw(anna::RuntimeException) { return send(&message, socketId); }

  /**
     Gets the last used resource (client session) during sending.
     Broadcast doesn't updates this information.
  */
  ClientSession *getLastUsedResource() const throw() { return (a_lastUsedResource); }

  /**
     Sent a message to all the server client-sessions (socketId's) for multiple client client-sessions functionality.
     It is used, i.e., in Disconnect-Peer-Request procedure over a certain server.
     Returns true (success) only when broadcast is success over all the server client-sessions. If any client-session fails,
     then false is returned. Broadcast try to send all over the resources in spite of any fail.
  */
  bool broadcast(const Message*) throw(anna::RuntimeException);
  bool broadcast(const Message& message) throw(anna::RuntimeException) { return broadcast(&message); }

  /**
     Close all the server client-sessions. Depending on client-session configuration ('OnDisconnect' behaviour),
     pending answers will be wait (graceful) or ignored (immediate-abrupt close).
     Resources are not destroyed.
  */
  void close() throw(anna::RuntimeException) { close(false /* no destroy */); }


  /**
     Diameter parent entity.
     \return Diameter parent entity.
  */
  const Entity *getParent() const throw() { return a_parent; }

  /**
     Returns true when any of the server client-sessions is Bound. False when all not-bound.
  */
  bool isAvailable() const throw() { return a_available; }


  std::vector<ClientSession*>::iterator begin() throw() { return a_clientSessions.begin(); }
  std::vector<ClientSession*>::iterator end() throw() { return a_clientSessions.end(); }
  std::vector<ClientSession*>::const_iterator begin() const throw() { return a_clientSessions.begin(); }
  std::vector<ClientSession*>::const_iterator end() const throw() { return a_clientSessions.end(); }

  int getNumberOfClientSessions() const throw() { return a_clientSessions.size(); }
  int getMaxClientSessions() const throw() { return a_maxClientSessions; }
  void setMaxClientSessions(int maxClientSessions) throw() { a_maxClientSessions = maxClientSessions; }

  /**
     Diameter server  address (IP or hostname)
     \return Diameter server address.
  */
  const std::string& getAddress() const throw() { return a_socket.first; }

  /**
     Diameter server port.
     \return Diameter server port.
  */
  int getPort() const throw() { return a_socket.second; }


  /** Server presentation as 'ADDRESS:PORT' */
  std::string socketAsString() const throw();


  /**
     Gets the timestamp for last incoming activity over the server.

     @return Last incoming activity timestamp.
  */
  const anna::Millisecond & getLastIncomingActivityTime() const throw() { return a_lastIncomingActivityTime; }

  /**
     Gets the timestamp for last outgoing activity over the server.

     @return Last outgoing activity timestamp.
  */
  const anna::Millisecond & getLastOutgoingActivityTime() const throw() { return a_lastOutgoingActivityTime; }


  /**
     Gets the number of requests messages over-the-air.

     @return OTA messages.
  */
  int getOTARequests() const throw();

  /**
     Returns idle state (no pending answers).

     @return Idle state.
  */
  bool idle() const throw() { return (getOTARequests() == 0); }


  /**
     Deny resources for delivery restriction.
     Deny all its client sessions
  */
  void hide() throw();

  /**
     Allow resource for delivery permission.
     Allow all its client sessions
  */
  void show() throw();

  /**
     Returns true when all its client session resources are hidden for application messages delivery
  */
  bool hidden() const throw();

  /**
     Returns true when all its client session resources are shown for application messages delivery
  */
  bool shown() const throw();


  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  std::string asString() const throw();

  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();


  // Statistics
  void updateProcessingTimeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw();
  void updateReceivedMessageSizeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw();
//  int getProcessingTimeStatisticConcept() const throw() { return a_processing_time__StatisticConceptId; }
//  int getReceivedMessageSizeStatisticConcept() const throw() { return a_received_message_size__StatisticConceptId; }

protected:

  /**
     Handler about event break connection from diameter server (client-session) over this entity.

     When notified, ANNA.diameter.comm generates an diameter::comm::Entity::eventResponse for every request with pending answers.
  */
  virtual void eventPeerShutdown(const ClientSession*) throw();

  /**
     Handler about a request retransmission over the session.

     \param request Message retransmitted
  */
  virtual void eventRequestRetransmission(const ClientSession*, Message *request) throw();

  /**
     Handler for diameter server (client-session) responses

     \param response Answer container object for corresponding diameter request
  */
  virtual void eventResponse(const Response & response) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) requests

     \param clientSession ClientSession from which request has been received
     \param request Diameter request message received
  */
  virtual void eventRequest(ClientSession *clientSession, const anna::DataBlock &request) throw(anna::RuntimeException);
  //virtual void eventRequest(ClientSession *clientSession, const Message & request) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) responses out of context

     \param clientSession ClientSession from which request has been received
     \param response Answer data block object without context match
  */
  virtual void eventUnknownResponse(ClientSession *clientSession, const anna::DataBlock& response) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) Disconnect-Peer-Answer messages

     \param clientSession ClientSession from which request has been received
     \param response Answer data block object without context match
  */
  virtual void eventDPA(ClientSession *clientSession, const anna::DataBlock& response) throw(anna::RuntimeException);



  friend class Engine;
  friend class ClientSession;
};

}
}
}

#endif
