// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_codec_Avp_hpp
#define anna_diameter_codec_Avp_hpp


// Local
#include <anna/config/defines.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter/codec/basetypes/basetypes.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/stack/Avp.hpp>

#include <anna/core/RuntimeException.hpp>

// STL
#include <string>
#include <map>

//------------------------------------------------------------------------------
//---------------------------------------------------------------------- #define
//------------------------------------------------------------------------------

namespace anna {

class DataBlock;

namespace xml {
class Node;
class Attribute;
}
}

namespace anna {

namespace diameter {


namespace stack {
class Dictionary;
class Format;
class Avp;
}

namespace codec {

namespace basetypes {
class OctetString;
class Integer32;
class Integer64;
class Unsigned32;
class Unsigned64;
class Float32;
class Float64;
class Address;
class Time;
class UTF8String;
class DiameterIdentity;
class DiameterURI;
class Enumerated;
class IPFilterRule;
class QoSFilterRule;
class Unknown;
}

class Avp;
class Message;
class Engine;


typedef std::map < int /* key: insertion pos */, Avp* > avp_container;
typedef avp_container::iterator avp_iterator;
typedef avp_container::const_iterator const_avp_iterator;

// Cache avp-find system
typedef std::pair < AvpId, unsigned int /* position */ > find_key;
typedef std::map<find_key, Avp*> find_container;
typedef std::map<find_key, Avp*>::iterator find_iterator;


using namespace basetypes;

/**
* Diameter avp generic container
* <pre>
*    RFC 3588                Diameter Based Protocol           September 2003
*    4.1.  AVP Header
*
*       The fields in the AVP header MUST be sent in network byte order.  The
*       format of the header is:
*
*        0                   1                   2                   3
*        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
*       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*       |                           AVP Code                            |
*       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*       |   AVP Flags   |                  AVP Length                   |
*       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*       |                        Vendor-ID (opt)                        |
*       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*       |    Data ...
*       +-+-+-+-+-+-+-+-+
*
*
*        AVP Flags: V (vendor-specific), M (mandatory), P (end to end encryption)
*        +-+-+-+-+-+-+-+-+
*        |V M P r r r r r|
*        +-+-+-+-+-+-+-+-+
*
*        AVP Length
*        The AVP Length field is three octets, and indicates the number of
*        octets in this AVP including the AVP Code, AVP Length, AVP Flags,
*        Vendor-ID field (if present) and the AVP data.
*
*        Vendor-ID: IANA "SMI Network Management Private Enterprise Codes" [ASSIGNNO]
*                   http://www.iana.org/assignments/enterprise-numbers
* </pre>
*/
class Avp {

  AvpId a_id;          // code and vendor-code
  U8 a_flags;

  // auxiliary
  int a_insertionPositionForChilds; // used at grouped type


  // --- Developer notes ---
  // 'AVP Length' does not include posible data padding. Thanks to this, 'Data Length'
  // is the difference between 'AVP Length' and sum of code, length, flags and
  // optionally the vendor-ID (all of them are 32-bit boundary), that is to say:
  // 8 or 12 (vendor-specific avps).
  //
  // Grouped avps 'AVP Length' includes own headers plus the total length of all
  //  underlying AVPs, including their headers and padding, then 'AVP Length' is
  //  always multiple of 4 (library will check this), and smae for 'Data Length'
  //  which is an 'whole avp Length with padding' itself.

  // Data containers
  OctetString *a_OctetString;
  Integer32 *a_Integer32;
  Integer64 *a_Integer64;
  Unsigned32 *a_Unsigned32;
  Unsigned64 *a_Unsigned64;
  Float32 *a_Float32;
  Float64 *a_Float64;
  avp_container a_avps; // Grouped
  Address *a_Address;
  Time *a_Time;
  UTF8String *a_UTF8String;
  DiameterIdentity *a_DiameterIdentity;
  DiameterURI *a_DiameterURI;
  Enumerated *a_Enumerated;
  IPFilterRule *a_IPFilterRule;
  QoSFilterRule *a_QoSFilterRule;
  Unknown *a_Unknown;

  // Grouped helpers
  find_container a_finds; // fast access for grouped and message first-level avps


  // Static functions are used when you want a function that is the same for every instance of a class. Such functions do not have access
  //  to "this" pointer and thus cannot access any non static fields. They are used often when you want a function that can be used without
  //  instantiating the class. Friend functions are functions which are not in the class and you want to give them access to private members
  //  of your class.

  // Common (also for Message class)
  static avp_iterator avp_find(avp_container &avps, AvpId id, unsigned int position) throw();
  static const_avp_iterator avp_find(const avp_container &avps, AvpId id, unsigned int position) throw() {
    return (const_avp_iterator)avp_find((avp_container &)avps, id, position);
  }
  static Avp * addAvp(avp_container &avps, int &insertionPositionForChilds, AvpId id, Engine *engine) throw();
  static bool removeAvp(avp_container &avps, find_container &finds, AvpId id, int ocurrence, Engine *engine) throw();
  static void fix(avp_container &avps, find_container &finds, int &insertionPositionForChilds, anna::diameter::stack::const_avprule_iterator ruleBegin, anna::diameter::stack::const_avprule_iterator ruleEnd) throw();
  static bool validLevel(const avp_container &avps, anna::diameter::stack::const_avprule_iterator ruleBegin, anna::diameter::stack::const_avprule_iterator ruleEnd, Engine * engine, const anna::diameter::codec::parent_t & parent, Message *answer) throw(anna::RuntimeException); // validates mandatory/fixed and cardinality
  static const Avp* getAvp(const avp_container &avps, find_container &finds, AvpId id, int ocurrence, Engine *engine, anna::Exception::Mode::_v emode) throw(anna::RuntimeException);
  static int countAvp(const avp_container &avps, AvpId id) throw();
  static const Avp* firstAvp(const avp_container &avps, AvpId id) throw();
  static int countChilds(const avp_container &avps) throw();
  static int addChild(avp_container &avps, int &insertionPositionForChilds, Avp *avp) throw() {
    if(!avp) return -1;

    avps[insertionPositionForChilds++] = avp;
    return insertionPositionForChilds;
  }
  static const anna::diameter::stack::Avp *getStackAvp(AvpId id, Engine *engine) throw();
  const Avp* _getAvp(const char *name, int ocurrence, anna::Exception::Mode::_v emode) const throw(anna::RuntimeException);
  const Avp* _getAvp(AvpId id, int ocurrence, anna::Exception::Mode::_v emode) const throw(anna::RuntimeException);

  // Own
  avp_iterator avp_begin() throw() { return a_avps.begin(); }
  avp_iterator avp_end() throw() { return a_avps.end(); }
  static Avp* avp(avp_iterator ii) throw() { return ii->second; }
  const_avp_iterator avp_begin() const throw() { return a_avps.begin(); }
  const_avp_iterator avp_end() const throw() { return a_avps.end(); }
  static const Avp* avp(const_avp_iterator ii) throw() { return ii->second; }

  // Internal
  bool flagsOK() const throw(); // flags coherence regarding dictionary. Only must be called when AVP is identified at the dictionary.
  int addChild(Avp *avp) throw(anna::RuntimeException) { assertFormat("Grouped"); return addChild(a_avps, a_insertionPositionForChilds, avp); }
  bool hasChildren() throw() { return a_avps.size() != 0; }

  static bool contain(const_avp_iterator begin, const_avp_iterator end, const Avp *parent) throw() { return true; }


  /**
     Fix grouped content regarding dictionary avp positions.
     Avp could remain invalid because of possible fixed/mandatory avps.
     This is useful to give flexibility to the application during message construction before encoding or representing the data.
     Is not recommended to fix a recently decoded message because possible validation problems will be hidden.
  */
  void fix() throw();

  /**
     Validates an Avp regarding dictionary rules like enumerated range, flags coherence, mandatory and fixed types, cardinality qualifiers, etc.

     @param parent Parent description. Internally used for alarms, tracing and Failed-AVP construction
     @param answer Answer could be modified with any validation problem during requests validation

     @return Boolean indicating validation result
  */
  bool valid(const anna::diameter::codec::parent_t & parent, Message *answer) const throw(anna::RuntimeException);

  /**
     Decodes buffer provided over class content. If an error ocurred, decoding will stop launching exception (fatal error) or a warning trace (perhaps the achieved
     message is valid against all odds then validation will go on). In case that validation is enabled (codec::Engine::ValidationMode) an exception will be launched
     depending on validation depth (codec::Engine::ValidationDepth).

     @param db Buffer data block processed
     @param parent Parent description. Internally used for alarms, tracing and Failed-AVP construction
     @param answer Answer built for request decoding/validation
  */
  void decode(const anna::DataBlock &db, const anna::diameter::codec::parent_t & parent, Message *answer) throw(anna::RuntimeException);


  /////////////////////////////////////////////
  // Inherit format-specific virtual methods //
  /////////////////////////////////////////////

  /**
  * Initializes Avp class information.
  * Default implementation supports all anna::diameter formats (including derived ones).
  * Diameter basic formats are managed at #initialize, which will invoke this method at the end.
  */
  virtual void initializeByFormat() throw() {};

  /**
  * Gets avp data-part length.
  * Default implementation supports all anna::diameter formats (including derived ones).
  * Diameter basic formats are managed at #initialize, which will invoke this method at the end.
  *
  * @param stackFormat Stack avp format in which data extraction is based.
  *
  * @return Avp data-part size.
  */
  virtual U24 getLengthByFormat(const anna::diameter::stack::Format *stackFormat) const throw() { return 0; };

  /**
     Gets data or hexadecimal data depending on avp format, for xml creating
     Default implementation supports all anna::diameter formats (including derived ones).
     Diameter basic formats are managed at #initialize, which will invoke this method at the end.

     \param isHex Hexadecimal/Natural data when apply.
     \param stackFormat Stack avp format in which data extraction is based.
     \return xml data representation
  */
  virtual std::string getXMLdataByFormat(bool & isHex, const anna::diameter::stack::Format *stackFormat) const throw() { return ""; };

  /**
     Interpret xml data in order to dump over the class content.
     Default implementation supports all anna::diameter formats (including derived ones).
     Diameter basic formats are managed at #initialize, which will invoke this method at the end.

     \param data Avp data attribute
     \param hexData Avp hex-data attribute
     \param stackFormat Stack avp format in which data extraction is based.
  */
  virtual void fromXMLByFormat(const anna::xml::Attribute* data, const anna::xml::Attribute* hexData, const anna::diameter::stack::Format *stackFormat) throw(anna::RuntimeException) {};


  /**
     Encodes buffer with the class content.
     Default implementation supports all anna::diameter formats (including derived ones).
     Diameter basic formats are managed at #initialize, which will invoke this method at the end.

     @param dataPart Data-part begin pointer
     @param stackFormat Stack avp format in which data extraction is based.
  */
  virtual void codeByFormat(char* dataPart, const anna::diameter::stack::Format *stackFormat) const throw(anna::RuntimeException) {};


  /**
     Decodes Avp data part.
     Default implementation supports all anna::diameter formats (including derived ones).
     Diameter basic formats are managed at #initialize, which will invoke this method at the end.

     @param buffer Avp data part start pointer
     @param size Avp data part size
     @param stackFormat Stack avp format in which data extraction is based.
  */
  virtual void decodeDataPartByFormat(const char * buffer, int size, const anna::diameter::stack::Format *stackFormat) throw(anna::RuntimeException) {};

  /**
     Reserves memory for data part depending on avp format for the identifier provided.
     Default implementation supports all anna::diameter formats (including derived ones).
     Diameter basic formats are managed at #initialize, which will invoke this method at the end.

     @param stackFormat Stack avp format in which data extraction is based.
  */
  virtual void allocationByFormat(const anna::diameter::stack::Format *stackFormat) throw() {};

  /**
  * Clears Avp data-part format containers.
  */
  virtual void clearByFormat() throw() {};



protected:

  /** Codec Engine */
  mutable Engine *a_engine;

  /** Codec Engine getter: avoids have to create base engine when using its child */
  virtual Engine * getEngine() const throw(anna::RuntimeException);

  /**
  * Initializes Avp class information.
  */
  void initialize() throw();

  /**
  * Assert format regarding dictionary
  */
  void assertFormat(const std::string &name) const throw(anna::RuntimeException);

  /**
  * Gets avp total length based on internal data part and header configuration.
  * Padding octets are not included, only header and data part length.
  * The only format which always have total length equal to sum of all its parts is Grouped,
  * because of the 4-multiple nature of its data part length.
  */
  U24 getLength() const throw();

  /**
     Gets data or hexadecimal data depending on avp format, for xml creating

     \param isHex Hexadecimal/Natural data when apply.
     \param stackFormat Stack avp format in which data extraction is based.
     \return xml data representation
  */
  std::string getXMLdata(bool & isHex, const anna::diameter::stack::Format *stackFormat) const throw();


  /**
     Decodes Avp data part.

     @param buffer Avp data part start pointer
     @param size Avp data part size
     @param parent Parent description. Internally used for alarms, tracing and Failed-AVP construction
     @param answer Answer built for request decoding/validation
  */
  void decodeDataPart(const char * buffer, int size, const anna::diameter::codec::parent_t & parent, Message *answer) throw(anna::RuntimeException);


public:

  /**
  * Default constructor
  * @param engine Codec engine used
  */
  Avp(Engine *engine = NULL);

  /**
  * Identified constructor
  * @param id Avp identifier as pair (code,vendor-id).
  * @param engine Codec engine used
  */
  Avp(AvpId id, Engine *engine = NULL);


  /**
  * Sets the codec engine
  *
  * Once assigned (here or at constructor), this method SHALL NOT be used anymore.
  * Also, the associated dictionary SHOULD NOT BE CHANGED through the engine,
  * unless you know what are you doing. If you want to reconfigure the engine,
  * first #clear the avp and then you could reuse the same object with
  * different configurations (execution contexts).
  *
  * Setting a new different engine with different stack, even same engine where the
  * stack has been dynamically changed, could cause a bad behaviour depending on the
  * changes: in general, if the dictionary grows, nothing bad will happen, but if
  * you remove or modified some elements which were processed with a certain format,
  * will be interpreted as 'unknown' with the new dictionary, and then some problems
  * may occur. If you add elements (vendors, avps, messages) is not a problem.
  */
  void setEngine(Engine *engine) throw();


  // Length references
  static const int HeaderLengthVactive;
  static const int HeaderLengthVinactive;

  //    AVP Flags
  //    +-+-+-+-+-+-+-+-+
  //    |V M P r r r r r|
  //    +-+-+-+-+-+-+-+-+
  //
  //      V(endor-specific)
  //      M(andatory)
  //      (encry)P(tion) (end to end encryption)
  //      r(eserved)  - these flag bits are reserved for future use, and
  //                    MUST be set to zero, and ignored by the receiver.
  static const U8 VBitMask;
  static const U8 MBitMask;
  static const U8 PBitMask;

  /**
  * Destructor
  */
  ~Avp();


  // setters

  /**
  * Clears and initializes Avp class information.
  * Application should clear auxiliary avp objects before setting data in a new context.
  */
  void clear() throw(anna::RuntimeException);


  /**
     Sets the avp identifier and clear the former content.
     Internally reserves memory for data part depending on avp format for the identifier provided.
     This must be called at first place because Avp class content is cleared when identifier is configured.
     Generic AVP assignment have no sense and will be ignored.

     @param id Avp identifier as pair (code,vendor-id).
  */
  void setId(AvpId id) throw(anna::RuntimeException);

  /**
     Same as #setId but providing dictionary logical name for Avp searched
  */
  void setId(const char *name) throw(anna::RuntimeException);

  /**
     Sets/unsets M bit activation.
     Application should not have to use this because dictionary information is used in order to configure flags when Avp identifier is stored.
     Anyway, could be useful when build unknown-type avps.

     The 'M' Bit, known as the Mandatory bit, indicates whether support of the AVP is required. If an AVP with the 'M' bit set is received by
     a Diameter client, server, proxy, or translation agent and either the AVP or its value is unrecognized, the message MUST be rejected.
     Diameter Relay and redirect agents MUST NOT reject messages with unrecognized AVPs.

     @param activate Activates/deactivates the bit. True by default.
  */
  void setMandatoryBit(bool activate = true) throw() { if(activate) a_flags |= MBitMask; else a_flags &= (~MBitMask); }

  /**
     Sets/unsets P bit activation.
     Application should not have to use this because dictionary information is used in order to configure flags when Avp identifier is stored.
     Anyway, could be useful when build unknown-type avps.

     @param activate Activates/deactivates the bit. True by default.
  */
  void setEncryptionBit(bool activate = true) throw() { if(activate) a_flags |= PBitMask; else a_flags &= (~PBitMask); }


  /**
     Adds an avp child providing its identifier and reserve internal memory it.
     An exception is launched is the Avp over which we add the new avp, is not a grouped avp.

     @param id Avp identifier as pair (code,vendor-id).

     @return Pointer to the new created avp.
  */
  Avp * addAvp(AvpId id) throw(anna::RuntimeException) { assertFormat("Grouped"); return addAvp(a_avps, a_insertionPositionForChilds, id, getEngine()); }


  /**
     Same as #addAvp but providing dictionary logical name for Avp searched
  */
  Avp * addAvp(const char *name) throw(anna::RuntimeException);

  /**
     Adds an avp child providing a persistent pointer (must be maintained by application).
     An exception is launched is the Avp over which we add the new avp, is not a grouped avp.
     It is not allowed to add an avp with no codec engine configured, neither if the engine
     is not the same.

     @param avp Avp external pointer. If NULL provided, nothing is done and NULL returned.
     Also NULL returned for bad engine configuration.

     @return Pointer to the added avp (again).
  */
  Avp * addAvp(Avp * avp) throw(anna::RuntimeException);


  // Data part access
  /** Access content for OctetString Avp in order to set data part */
  OctetString *        getOctetString() throw(anna::RuntimeException) { assertFormat("OctetString"); return a_OctetString; }
  /** Access content for Integer32 Avp in order to set data part */
  Integer32 *          getInteger32() throw(anna::RuntimeException) { assertFormat("Integer32"); return a_Integer32; }
  /** Access content for Integer64 Avp in order to set data part */
  Integer64 *          getInteger64() throw(anna::RuntimeException) { assertFormat("Integer64"); return a_Integer64; }
  /** Access content for Unsigned32 Avp in order to set data part */
  Unsigned32 *         getUnsigned32() throw(anna::RuntimeException) { assertFormat("Unsigned32"); return a_Unsigned32; }
  /** Access content for Unsigned64 Avp in order to set data part */
  Unsigned64 *         getUnsigned64() throw(anna::RuntimeException) { assertFormat("Unsigned64"); return a_Unsigned64; }
  /** Access content for Float32 Avp in order to set data part */
  Float32 *            getFloat32() throw(anna::RuntimeException) { assertFormat("Float32"); return a_Float32; }
  /** Access content for Float64 Avp in order to set data part */
  Float64 *            getFloat64() throw(anna::RuntimeException) { assertFormat("Float64"); return a_Float64; }
  /** Access content for Address Avp in order to set data part */
  Address *            getAddress() throw(anna::RuntimeException) { assertFormat("Address"); return a_Address; }
  /** Access content for Time Avp in order to set data part */
  Time *               getTime() throw(anna::RuntimeException) { assertFormat("Time"); return a_Time; }
  /** Access content for UTF8String Avp in order to set data part */
  UTF8String *         getUTF8String() throw(anna::RuntimeException) { assertFormat("UTF8String"); return a_UTF8String; }
  /** Access content for DiameterIdentity Avp in order to set data part */
  DiameterIdentity *   getDiameterIdentity() throw(anna::RuntimeException) { assertFormat("DiameterIdentity"); return a_DiameterIdentity; }
  /** Access content for DiameterURI Avp in order to set data part */
  DiameterURI *        getDiameterURI() throw(anna::RuntimeException) { assertFormat("DiameterURI"); return a_DiameterURI; }
  /** Access content for Enumerated Avp in order to set data part */
  Enumerated *         getEnumerated() throw(anna::RuntimeException) { assertFormat("Enumerated"); return a_Enumerated; }
  /** Access content for IPFilterRule Avp in order to set data part */
  IPFilterRule *       getIPFilterRule() throw(anna::RuntimeException) { assertFormat("IPFilterRule"); return a_IPFilterRule; }
  /** Access content for QoSFilterRule Avp in order to set data part */
  QoSFilterRule *      getQoSFilterRule() throw(anna::RuntimeException) { assertFormat("QoSFilterRule"); return a_QoSFilterRule; }
  /** Access content for Unknown Avp in order to set data part */
  Unknown *            getUnknown() throw(anna::RuntimeException) { assertFormat("Unknown"); return a_Unknown; }


  /**
     Removes an Avp within grouped type (first level) and free resources.

     @param id Avp identifier (pair code + vendor-id).
     @param ocurrence Order of appearance for the searched avp. Zero value means remove all avps with provided identifier at first level (no recursiveness would be allowed in the API in order to avoid unexpected behaviour).
     Negative values could be used to reverse access positions: i.e. -1 is the last ocurrence, -2 is the second to last (penultimate), etc.

     @return Returns true if something was removed. False in other cases (including i.e. when this Avp is empty or is not a grouped avp).
  */
  bool removeAvp(AvpId id, int ocurrence = 1) throw(anna::RuntimeException) { assertFormat("Grouped"); return removeAvp(a_avps, (find_container&)a_finds, id, ocurrence, getEngine()); }


  /**
     Same as #removeAvp but providing dictionary logical name for Avp searched
  */
  bool removeAvp(const char *name, int ocurrence = 1) throw(anna::RuntimeException);

  // getters

  /**
     Gets Avp identifier as pair (code, vendor-id).
  */
  const AvpId & getId() const throw() { return a_id; }

  /**
     Gets Avp vendor-id.
  */
  int getVendorId() const throw() { return a_id.second; }

  /**
     Gets stack avp (dictionary avp reference).
  */
  const anna::diameter::stack::Avp *getStackAvp() const throw(anna::RuntimeException) { return getStackAvp(a_id, getEngine()); }

  /** Returns V bit activation state */
  bool vendorBit() const throw() { return ((a_flags & VBitMask) != 0x00); }

  /** Returns M bit activation state */
  bool mandatoryBit() const throw() { return ((a_flags & MBitMask) != 0x00); }

  /** Returns P bit activation state */
  bool encryptionBit() const throw() { return ((a_flags & PBitMask) != 0x00); }

  // Data part access
  /** Access content for OctetString Avp */
  const OctetString *        getOctetString() const throw(anna::RuntimeException) { assertFormat("OctetString"); return a_OctetString; }
  /** Access content for Integer32 Avp */
  const Integer32 *          getInteger32() const throw(anna::RuntimeException) { assertFormat("Integer32"); return a_Integer32; }
  /** Access content for Integer64 Avp */
  const Integer64 *          getInteger64() const throw(anna::RuntimeException) { assertFormat("Integer64"); return a_Integer64; }
  /** Access content for Unsigned32 Avp */
  const Unsigned32 *         getUnsigned32() const throw(anna::RuntimeException) { assertFormat("Unsigned32"); return a_Unsigned32; }
  /** Access content for Unsigned64 Avp */
  const Unsigned64 *         getUnsigned64() const throw(anna::RuntimeException) { assertFormat("Unsigned64"); return a_Unsigned64; }
  /** Access content for Float32 Avp */
  const Float32 *            getFloat32() const throw(anna::RuntimeException) { assertFormat("Float32"); return a_Float32; }
  /** Access content for Float64 Avp */
  const Float64 *            getFloat64() const throw(anna::RuntimeException) { assertFormat("Float64"); return a_Float64; }

  /**
     Access content for Grouped Avp. Exception mode allows different combinations like cascade access:
     <pre>

        try {
           message->getAvp(anna::diameter::helpers::base::AVP__Multiple_Services_Credit_Control, anna::Exception::Mode::Throw)
                  ->getAvp(anna::diameter::helpers::base::AVP__Rating_Group, anna::Exception::Mode::Throw);
        }
        catch(anna::RuntimeException) {;}
     </pre>

     Or step access:

     <pre>
        const Avp *mscc = message->getAvp(anna::diameter::helpers::base::AVP__Multiple_Services_Credit_Control);
        const Avp *rg;
        if (mscc) rg = mscc->getAvp(anna::diameter::helpers::base::AVP__Rating_Group);
     </pre>

     Replacing procedures becomes easy because an Avp can be searched and its pointer reconfigured by mean #setId and data part setters.
     Deleting procedures must use #removeAvp.
     Access is internally cached to speed up the search operations. This cache is reset after calling #fix or #removeAvp methods.

     @param id Avp identifier (pair code + vendor-id).
     @param ocurrence Order of appearance for the searched avp. Zero position is rejected, but negative values could be used to reverse
     access positions: i.e. -1 is the last ocurrence, -2 is the second to last (penultimate), etc.
     @param emode Excepcion mode handling when avp is not found: Ignore (no action is taken but debug trace), Throw (excepcion launched, by default), Trace (trace warning).
     If avp format is not grouped, always exception will be launched and no matter what mode is provided. It would be a development
     error and must be solved.
  */
  const Avp* getAvp(AvpId id, int ocurrence = 1, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) const throw(anna::RuntimeException) {
    return _getAvp(id, ocurrence, emode);
  }

  Avp* getAvp(AvpId id, int ocurrence = 1, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException) {
    return const_cast<Avp*>(_getAvp(id, ocurrence, emode));
  }

  /**
     Same as #getAvp but providing dictionary logical name for Avp searched
  */
  const Avp* getAvp(const char *name, int ocurrence = 1, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) const throw(anna::RuntimeException) {
    return _getAvp(name, ocurrence, emode);
  }

  Avp* getAvp(const char *name, int ocurrence = 1, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException) {
    return const_cast<Avp*>(_getAvp(name, ocurrence, emode));
  }



  /** Access content for Address Avp */
  const Address *            getAddress() const throw(anna::RuntimeException) { assertFormat("Address"); return a_Address; }
  /** Access content for Time Avp */
  const Time *               getTime() const throw(anna::RuntimeException) { assertFormat("Time"); return a_Time; }
  /** Access content for UTF8String Avp */
  const UTF8String *         getUTF8String() const throw(anna::RuntimeException) { assertFormat("UTF8String"); return a_UTF8String; }
  /** Access content for DiameterIdentity Avp */
  const DiameterIdentity *   getDiameterIdentity() const throw(anna::RuntimeException) { assertFormat("DiameterIdentity"); return a_DiameterIdentity; }
  /** Access content for DiameterURI Avp */
  const DiameterURI *        getDiameterURI() const throw(anna::RuntimeException) { assertFormat("DiameterURI"); return a_DiameterURI; }
  /** Access content for Enumerated Avp */
  const Enumerated *         getEnumerated() const throw(anna::RuntimeException) { assertFormat("Enumerated"); return a_Enumerated; }
  /** Access content for IPFilterRule Avp */
  const IPFilterRule *       getIPFilterRule() const throw(anna::RuntimeException) { assertFormat("IPFilterRule"); return a_IPFilterRule; }
  /** Access content for QoSFilterRule Avp */
  const QoSFilterRule *      getQoSFilterRule() const throw(anna::RuntimeException) { assertFormat("QoSFilterRule"); return a_QoSFilterRule; }
  /** Access content for Unknown Avp */
  const Unknown *            getUnknown() const throw(anna::RuntimeException) { assertFormat("Unknown"); return a_Unknown; }


  /**
     Decodes buffer provided over class content. If an error ocurred, decoding will stop launching exception (fatal error) or a warning trace (perhaps the achieved
     avp is valid against all odds then validation will go on). In case that validation is enabled (codec::Engine::ValidationMode) an exception will be launched
     depending on validation depth (codec::Engine::ValidationDepth).

     Useful as serialization procedure with #code

     @param db Buffer data block processed
  */
  void decode(const anna::DataBlock &db) throw(anna::RuntimeException);


  /**
     Interpret xml data in order to dump over the class content.

     \param avpNode Avp root node
  */
  void fromXML(const anna::xml::Node* avpNode) throw(anna::RuntimeException);


  /**
    Encodes buffer with the class content. This method is internally used to encode diameter messages, but is declared as public, to allow
    its use as serialization procedure. Then, it's assumed that this Avp is valid (validation shall be applied as part of a whole diameter
    message but nothing will be verified now).

  * @param buffer Raw data to be encoded (shall be externally allocated)
  * @param size Size of raw data to be encoded
  */
  void code(char* buffer, int &size) const throw(anna::RuntimeException);


  // Helpers

  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

  /**
     Class xml string representation
     \return XML string representation with relevant information for this instance.
  */
  std::string asXMLString() const throw();

  /**
     Comparison operator by mean serialization

     @param a1 Instance 1 for Avp class
     @param a2 Instance 2 for Avp class

     @return Comparison result
  */
  friend bool operator == (const Avp & a1, const Avp & a2) throw() { return (a1.asXMLString() == a2.asXMLString()); }

  /**
     Match a regular expression (string pattern) regarding xml string serialization for this avp.
     This works same as #Message::isLike

     @param pattern Pattern to match

     \return Returns the match result
  */
  bool isLike(const std::string &pattern) const throw();

  /**
     Counts the number of ocurrences of Avps (first level) with the identifier provided

     @param id Avp identifier (pair code + vendor-id).
  */
  int countAvp(AvpId id) const throw(anna::RuntimeException) { assertFormat("Grouped"); return countAvp(a_avps, id); }

  /**
     Same as #countAvp but providing dictionary logical name for Avp searched
  */
  int countAvp(const char *name) const throw(anna::RuntimeException);

  /**
     Counts the number of children within a grouped avp

     @param id Avp identifier (pair code + vendor-id).
  */
  int countChilds() const throw(anna::RuntimeException) { assertFormat("Grouped"); return countChilds(a_avps); }

  /**
     The 'M' Bit, known as the Mandatory bit, indicates whether support of the AVP is required. If an AVP with the 'M' bit set is received by
     a Diameter client, server, proxy, or translation agent and either the AVP or its value is unrecognized, the message MUST be rejected.
     Diameter Relay and redirect agents MUST NOT reject messages with unrecognized AVPs.

     Default implementation launch alarm and counter indicating the anomaly but don't launch exception (traces at warning level).
     Relay and Redirect agents could reimplement this method to avoid oam management (another way is avoid alarm/counter registration on
     these applications). Result-Code DIAMETER_AVP_UNSUPPORTED will be stored for possible answer message.
  */
  virtual void unknownAvpWithMandatoryBit() const throw(anna::RuntimeException);


  friend class Message;
  friend class Engine;
};

}
}
}


#endif
