// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_ldap_Search_hpp
#define anna_ldap_Search_hpp

#include <vector>

#include <anna/core/util/Recycler.hpp>

#include <anna/ldap/defines.hpp>
#include <anna/ldap/Request.hpp>
#include <anna/ldap/Scope.hpp>

namespace anna {

namespace ldap {

class Session;

/**
   Modela las peticiones de busqueda realizadas contra un servidor LDAP.
*/
class Search : public Request {
  typedef Recycler<std::string> attribute_pool;
  typedef std::vector<std::string*> attribute_container;
  typedef attribute_container::iterator attribute_iterator;

public:
  typedef attribute_container::const_iterator const_attribute_iterator;

  /**
     Constructor.
     \param onExpiry Indica la accin a realizar si el temporizador de esta transacin expira.
  */
  Search(const Request::OnExpiry::_v onExpiry = Request::OnExpiry::Abandon) : Request(ClassCode::Search, onExpiry) { clear(); }

  /**
     Devuelve el DN de la entrada en la que se comienza la busqueda.
     \return El DN de la entrada en la que se comienza la busqueda.
  */
  const std::string& getBase() const throw() { return a_base; }

  /**
     Devuelve el ambito de la busqueda.
     \return El ambito de la busqueda.
  */
  Scope::_v getScope() const throw() { return a_scope; }

  /**
     Devuelve el filtro aplicado a la busqueda.
     \return El filtro aplicado a la busqueda.
  */
  const std::string& getFilter() const throw() { return a_filter; }

  /**
     Devuelve el indicador de tratamiento de busquedas.
     \return El indicador de tratamiento de busquedas.
  */
  bool getOnlyType() const throw() { return a_onlyType; }

  /**
     Devuelve el numero maximo de elementos devueltos como resultado de la consulta.
     \return El numero maximo de elementos devueltos como resultado de la consulta.
     \see clearSizeLimit.
  */
  int getSizeLimit() const throw() { return a_sizeLimit; }

  /**
     Establece el DN de la entrada en la que comenzar la busqueda.
     \param base DN de la entrada en la que comenzar la busqueda.
  */
  void setBase(const std::string& base) throw() { a_base = base; }

  /**
     Establece el ambito de la busqueda.
     \param scope Ambito de la busqueda.
  */
  void setScope(const Scope::_v scope) throw() { a_scope = scope; }

  /**
     Establece el ambito de la busqueda.
     \param scope Texto que identifica el ambito de la busqueda. Si no se trata de un texto
     reconocido se devolvera una excepcion.
  */
  void setScope(const char* scope) throw(RuntimeException) { a_scope = Scope::asEnumEx(scope); }

  /**
     Establece el filtro aplicado a la busqueda.
     \param filter Filtro aplicado a la busqueda.
     \see clearFilter
  */
  void setFilter(const std::string& filter) throw() { a_filter = filter; }

  /**
     Determina como se van a tratar los atributos.
     \param onlyType Un valor \em true indica que solo se requieren los tipos de atributos, un valor
     \em false indica que se requieren tipos y valores asociados a los atributos.
  */
  void setOnlyType(const bool onlyType) throw() { a_onlyType = onlyType; }

  /**
     Establece el numero maximo de elementos devueltos como resultado de la consulta.
     \param sizeLimit Numero de entradas retornadas por la busqueda.
     \see clearSizeLimit.
  */
  void setSizeLimit(const int sizeLimit) throw() { a_sizeLimit = sizeLimit; }

  /**
     Elimina el filtro asociado a esta consulta.
  */
  void clearFilter() throw() { a_filter.clear(); }

  /**
     Elimina el limite en cuanto al numero de elementos indicados en la consulta.
  */
  void clearSizeLimit() throw() { a_sizeLimit = 0; }

  /**
   * Elimina los atributos asociados a esta consulta.
   */
  void clearAttributes() throw();

  /**
     Incorpora un elemento a la lista de atributos que deseamos obtener de cada una de las entradas
     que cumplen el filtro establecido.
     Si esta lista esta vacia se devolvera la informacion de todos los atributos implicados.
  */
  void addAttribute(const std::string& attribute) throw() {
    std::string* newString = st_attributes.create();
    *newString = attribute;
    a_attributes.push_back(newString);
  }

  /**
     Inicializa el contenido de esta clase.
  */
  void clear() throw();

  /**
     Devuelve un iterador que apunta el primer atributo contenido en la consulta.
     \return un iterador que apunta el primer atributo contenido en la consulta.
  */
  const_attribute_iterator attribute_begin() const throw() { return a_attributes.begin(); }

  /**
     Devuelve un iterador que apunta el ultimo atributo contenido en la consulta.
     \return un iterador que apunta el ultimo atributo contenido en la consulta.
  */
  const_attribute_iterator attribute_end() const throw() { return a_attributes.end(); }

  /**
     Devuelve el numero de atributos contenidos en la consulta.
     \return El numero de atributos contenidos en la consulta.
  */
  int attribute_size() const throw() { return a_attributes.size(); }

  /**
     Devuelve el atributo apuntado por el iterador recibido.
     \param ii Iterador con el que estamos recorriendo los atributos.
     \return El atributo apuntado por el iterador recibido.
  */
  static const std::string& attribute(const_attribute_iterator ii) throw() { return **ii; }

  /**
     Devuelve una cadena con la informacion referente a este objeto.
     \return Una cadena con la informacion referente a este objeto.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacion referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacion.
     \return Un nodo XML con la informacion referente a este objeto.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

private:
  std::string a_base;
  Scope::_v a_scope;
  std::string a_filter;
  bool a_onlyType;
  int a_sizeLimit;
  attribute_container a_attributes;

  static attribute_pool st_attributes;

  attribute_iterator attribute_begin() throw() { return a_attributes.begin(); }
  attribute_iterator attribute_end() throw() { return a_attributes.end(); }
  int attribute_size() throw() { return a_attributes.size(); }
  static std::string& attribute(attribute_iterator ii) throw() { return **ii; }

  IdMessage send(Session&) const throw();
};

}
}

#endif

