// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_testing_TestCase_hpp
#define anna_testing_TestCase_hpp

// Standard
#include <string>
#include <vector>
#include <map>

// Project
#include <anna/diameter/defines.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/testing/TestCondition.hpp>



namespace anna {
  class Millisecond;

  namespace xml {
    class Node;
  }
  namespace diameter {
    namespace comm {
      class OriginHost;
    }
  }

namespace testing {

  class TestStep;
  class TestStepWait;
class TestCase {

  void assertInitialized() const throw(anna::RuntimeException);
  void assertMessage(const anna::DataBlock &db, bool toEntity) throw(anna::RuntimeException);

public:

  // Debug summary:
  class DebugSummary {

    typedef struct {
      anna::Millisecond Timestamp;
      std::string Hint;
    } event_t;

    std::vector<event_t> a_events;
  public:
    void addHint(const std::string &hint) throw();
    void clear() throw();
    int events() const throw() { return a_events.size(); }
    anna::xml::Node* asXML(anna::xml::Node* parent) const throw();
  };

  TestCase(unsigned int id);
  ~TestCase();

  struct State { enum _v { Initialized, InProgress, Failed, Success }; };
  static const char* asText(const State::_v state) throw();
  const State::_v &getState() const throw() { return a_state; }
  const anna::Millisecond &getStartTimestamp() const throw() { return a_startTime; }
  void addDebugSummaryHint(const std::string &hint) throw() { a_debugSummary.addHint(hint); }
  void setState(const State::_v &state) throw();
  bool isFinished() const throw() { return (getState() == State::Failed || getState() == State::Success); }
  bool inProgress() const throw() { return (getState() == State::InProgress); }
  bool hasSameCondition(const TestCondition &condition) const throw();

  // Interactivity:
  void makeInteractive(bool yes = true) throw() { a_interactiveAmount = (yes ? 0:-1); }
  void addInteractiveAmount(unsigned int amount) throw() {
    if (a_interactiveAmount == -1) makeInteractive();
    if (amount == 0) return;
    a_interactiveAmount += amount;
    process();
  }
  int interactiveAmount() const throw() { return a_interactiveAmount; }
  void interactiveExecution() throw() { a_interactiveAmount --; }

  // Step type & information
  void addTimeout(const anna::Millisecond &timeout) throw(anna::RuntimeException);
  void addSendxml2e(const anna::DataBlock &db, anna::diameter::comm::OriginHost *host, int stepNumber) throw(anna::RuntimeException);
  void addSendxml2c(const anna::DataBlock &db, anna::diameter::comm::OriginHost *host, int stepNumber) throw(anna::RuntimeException);
  void addDelay(const anna::Millisecond &delay) throw(anna::RuntimeException);
  void addWait(bool fromEntity,
                const std::string &code, const std::string &bitR, const std::string &hopByHop, const std::string &applicationId,
                const std::string &sessionId, const std::string &resultCode,
                const std::string &msisdn, const std::string &imsi, const std::string &serviceContextId) throw(anna::RuntimeException);
  void addWaitAnswer(bool fromEntity, int stepNumber) throw(anna::RuntimeException);
  void addWaitRegexpHex(bool fromEntity, const std::string &regexp) throw(anna::RuntimeException);
  void addWaitRegexpXml(bool fromEntity, const std::string &regexp) throw(anna::RuntimeException);
  void addCommand(const std::string &cmd) throw(anna::RuntimeException);


  // Process:
  void nextStep() throw() { a_stepsIt++; }
  bool done() throw();
  bool process() throw(); // false to stop

  // Reset test case and underlaying information (steps context)
  bool reset(bool hard /* hard reset includes in-progress test cases */) throw();

  // getters
  const anna::Millisecond &getStartTime() const throw() { return a_startTime; }
  const unsigned int &getId() const throw() { return a_id; }

  //helpers
  int steps() const throw() { return a_steps.size(); }
  void addStep(TestStep *step) throw() { a_steps.push_back(step); }

  TestStepWait *searchNextWaitConditionFulfilled(const anna::DataBlock &message, bool waitFromEntity) throw();
      // When a message arrives, we identify the test case by mean the Session-Id. Then, from the current step iterator (included),
      //  we search for a fulfilling condition for that message. The first found, is 'completed' and then breaks the search.
  const TestStep *getStep(int stepNumber) const throw();

  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();
  std::string asXMLString() const throw();


private:
  // private members:
  unsigned int a_id;
  std::vector<TestStep*> a_steps;
  std::vector<TestStep*>::const_iterator a_stepsIt;
  std::map<anna::diameter::HopByHop, TestStep*> a_hopByHops; // for wait-answer
  State::_v a_state;
  anna::Millisecond a_startTime;
  DebugSummary a_debugSummary; // used when a test case has failed, uncovered message conditions, and any other hint.
  int a_interactiveAmount;

  friend class TestStep;
};

}
}

#endif
