// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_timex_MicroMeter_hpp
#define anna_timex_MicroMeter_hpp

#include <anna/core/functions.hpp>

namespace anna {

namespace timex {

/**
   Facilita la medicion de los tiempos empleados las distintas partes de nuestra aplicacion.

   Permite calcular tiempos acumulados como tiempos individuales. Por ejemplo:

   \code

   #include <anna.timex.MicroMeter.h>

   void foo () {
      timex::MicroMeter meter;

      goo ();
      Microsecond gooTime = meter.getDelay ();

      hoo ();
      Microsecond goohooTime = meter.setControlPoint ();

      joo ();
      Microsecond jooTime = meter.getDelay ();
   }
   \endcode
*/
class MicroMeter {
public:
  /**
     Constructor
     Inicializa la cuenta de temporizacion.
  */
  MicroMeter() { a_timestamp = functions::microsecond(); a_topReference = (Microsecond)0; }

  /**
   * Constructor copia.
   * Copia la cuenta de utilizacion de la instancia recibida como parametro.
   * \param other Instancia de la copiar los parametros para calcular el tiempo transcurrido.
   */
  MicroMeter(const MicroMeter& other) : a_timestamp(other.a_timestamp) { a_topReference = (Microsecond)0; ;}

  /**
   * Inicializa la cuenta de temporizacion. Este metodo es invocado automaticamente desde el contructor la clase
   * por lo que si vamos usar esta instancia para tomar un unica medida no es necesario invocarlo.
   * \warning Elimina el punto de referencia temporal que puediera haberse establecido con #setTopReference.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   */
  Microsecond setControlPoint() throw() {
    Microsecond now = functions::microsecond();
    Microsecond result = (now > a_timestamp) ? (now - a_timestamp) : (Microsecond)0;
    a_timestamp = now;
    clearTopReference();
    return result;
  }

  /**
   * Se da la posiblidad de establecer un punto de referencia temporal de forma
   * que cuando se invoque a MicroMeter::getDelay, el calculo de la diferencia de tiempo
   * no se har entre la marca de tiempo y el tiempo actual, sino la marca de
   * tiempo y sta marca de referencia.
   *
   * Esta funcionalidad ha sido requerida para medir el tiempo de ejecucin "real"
   * de tareas que se ejecutan dentro de un thread. Ya que puede pasar un tiempo
   * indeterminado desde que se termina la tarea MT (momento en el que se establecer
   * la marca de tiempo) y el ncleo y dems partes pueden tener conocimiento de que
   * esa tarea ha sido finalidad.
   */
  void setTopReference(const Microsecond & topReference) throw() { a_topReference = topReference; }

  /**
   * Elimina el punto de referencia temporal.
   */
  void clearTopReference() throw() { a_topReference = (Microsecond)0; }

  /**
   * Devuelve el nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacin.
   * Si se ha establecido un punto de referencia mediante #setTopReference, devolver la diferencia entre el
   * el punto de control y la referencia, en otro caso, devolver la diferencia entre el punto de control y el
   * momento actual.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   * \warning Si detecta algun fallo devolvera 0.
   */
  Microsecond getDelay() const throw() {
    Microsecond now = (a_topReference == 0) ? functions::microsecond() : a_topReference;
    return (now > a_timestamp) ? (now - a_timestamp) : (Microsecond)0;
  }

  /**
   * Operador copia.
   * \param timestamp Milisegundo en el que inicilizar la cuenta de esta instancia.
   * \warning Elimina el punto de referencia temporal que puediera haberse establecido con #setTopReference.
   * \see anna::functions::microsecond.
   */
  MicroMeter& operator= (const Microsecond & timestamp) throw() { a_timestamp = timestamp; a_topReference = (Microsecond)0; return *this; }

  /**
   * Operador copia.
   * \param other Instancia de la que copiar.
   */
  MicroMeter& operator= (const MicroMeter& other) throw() { a_timestamp = other.a_timestamp; a_topReference = other.a_topReference; return *this; }

private:
  Microsecond a_topReference;
  Microsecond a_timestamp;
};

}
}

#endif


