// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter/functions.hpp>
#include <anna/diameter/helpers/helpers.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/time/functions.hpp>

// Local
#include <anna/diameter.comm/ServerSession.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Response.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter.comm/ApplicationMessageOamModule.hpp>
#include <anna/diameter.comm/TimerManager.hpp>
#include <anna/diameter.comm/Timer.hpp>
#include <anna/diameter.comm/LocalServer.hpp>
#include <anna/diameter.comm/ServerSessionReceiver.hpp>
#include <anna/diameter.comm/ReceiverFactoryImpl.hpp>

#include <anna/app/functions.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/Node.hpp>
#include <anna/timex/Engine.hpp>

// Standard
#include <stdlib.h> // rand()
#include <time.h>



using namespace std;
using namespace anna::diameter;
using namespace anna::diameter::comm;

//static
const anna::Millisecond ServerSession::DefaultAllowedInactivityTime(90000); // Inactivity timeout


ServerSession::ServerSession() : Session("diameter::comm::ServerSession", "Diameter Inactivity Detection Timer"),
  a_receiverFactory(this),
  a_cer(ClassCode::Bind),
  a_dwr(ClassCode::ApplicationMessage) // not actually needed; Message is application type by default
{ initialize(); }

void ServerSession::initialize() throw() {
  Session::initialize();
  a_parent = NULL;
  a_clientSocket = NULL;
  a_deprecated = false; // selected for remove (server session was lost due to event break connection)
}

//ServerSession::~ServerSession() {;}

void ServerSession::setClientSocket(anna::comm::ClientSocket *clientSocket) throw() {
  a_clientSocket = clientSocket;
  a_clientSocket->setReceiverFactory(a_receiverFactory);
}


const std::string& ServerSession::getAddress() const throw() {
  return a_parent->getKey().first;
}

int ServerSession::getPort() const throw() {
  return a_parent->getKey().second;
}

const Response* ServerSession::send(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "send", ANNA_FILE_LOCATION));

  if(!message)
    throw anna::RuntimeException("Cannot send a NULL message", ANNA_FILE_LOCATION);

  // Command id:
  bool isRequest;
  diameter::CommandId cid = message->getCommandId(isRequest);
  diameter::ApplicationId aid = message->getApplicationId();

  LOGDEBUG(
    std::string msg = "Sending diameter message: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  // Checkings:
  if(a_deprecated) {
    string msg(asString());
    msg += " | ServerSession is deprecated and will be erased";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if((a_state == State::Closed) && (cid != helpers::base::COMMANDID__Capabilities_Exchange_Answer)) {
    string msg(asString());
    msg += " | ServerSession is not bound";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  // CER/DWR are not sent from server-side:
  if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) {
    string msg(asString());
    msg += " | Trying to send CER: unexpected message from server-side";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) {
    string msg(asString());
    msg += " | Trying to send DWR: unexpected message from server-side";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  // Check states:
  /*if (cid == helpers::base::COMMANDID__Capabilities_Exchange_Answer) {
     if (a_state != State::Closed) {
        string msg(asString());
        msg += " | Discarding CEA on not closed state";
        throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
     }

  } else*/ if(cid ==  helpers::base::COMMANDID__Device_Watchdog_Answer) {
    if(a_state == State::WaitingDPA) {
      string msg(asString());
      msg += " | DWA is not sent on 'WaitingDPA' state";
      //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      LOGDEBUG(anna::Logger::debug(msg, ANNA_FILE_LOCATION));
      return NULL;
    }

    if(a_state == State::Disconnecting) {
      string msg(asString());
      msg += " | DWA is not sent on 'Disconnecting' state";
      //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      LOGDEBUG(anna::Logger::debug(msg, ANNA_FILE_LOCATION));
      return NULL;
    }
  } else if(cid ==  helpers::base::COMMANDID__Disconnect_Peer_Request) {
    if(a_state == State::WaitingDPA) {
      string msg(asString());
      msg += " | Still waiting for DPR ack (DPA)";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(a_state == State::Disconnecting) {
      string msg(asString());
      msg += " | Server disconnection has already been initiated";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  } else {
    if((a_state == State::WaitingDPA) || (a_state == State::Disconnecting)) {
      if(cid != helpers::base::COMMANDID__Disconnect_Peer_Answer) {
        LOGDEBUG(
          string msg("diameter::comm::ServerSession::send | ");
          msg += asString();
          msg += " | Sents (request or answer) blocked to diameter client (disconnection in progress). Discarding ...";
          anna::Logger::debug(msg, ANNA_FILE_LOCATION);
        );
        return NULL;
      }
    }
  }

  // Trace send operation:
  LOGDEBUG(
    string msg("diameter::comm::ServerSession::send | ");
    msg += asString();
    msg += " | ";
    msg += message->asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  bool fixed = false; // answers cannot be fixed
  Message * message_nc = const_cast<Message*>(message);

  if(isRequest) {
    // Fixing indicator:
    fixed = message_nc->fixRequestSequence(a_nextHopByHop, a_nextEndToEnd);
    message_nc->updateRequestTimestampMs(); // statistics purposes (processing time for request type)
  }

  // Send message
  try {
    message->send(*this);

    // Next hop by hop & end to end identifiers:
    if(isRequest) generateNextSequences();

    //   Transaction state
    //         The Diameter protocol requires that agents maintain transaction
    //         state, which is used for failover purposes.  Transaction state
    //         implies that upon forwarding a request, the Hop-by-Hop identifier
    //         is saved; the field is replaced with a locally unique identifier,
    //         which is restored to its original value when the corresponding
    //         answer is received.  The request's state is released upon receipt
    //         of the answer.  A stateless agent is one that only maintains
    //         transaction state.
    //
    updateOutgoingActivityTime();
    // OAM
    countSendings(cid, aid, true /* send ok */);
    // Trace non-application messages:
    LOGDEBUG(

      if( (cid == helpers::base::COMMANDID__Device_Watchdog_Request) ||
          (cid == helpers::base::COMMANDID__Disconnect_Peer_Request)) {
        anna::Logger::debug("Sent DataBlock to XML representation:", ANNA_FILE_LOCATION);
        try {
          anna::diameter::codec::Message msg(a_engine->getBaseProtocolCodecEngine()); msg.decode(message->getBody()); /* decode to be traced */
        }
        catch(anna::RuntimeException &ex) {
          std::string msg = ex.getText();
          msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages full tracing";
          anna::Logger::debug(msg, ANNA_FILE_LOCATION);
        }
      }
    );

    // Restore sequences:
    if(fixed) message_nc->restoreSequencesAfterFix();  // restore to application sequences after fix
  } catch(anna::RuntimeException&) {
    if(fixed) message_nc->restoreSequencesAfterFix();  // restore to application sequences after fix

    // OAM
    countSendings(cid, aid, false /* send no ok */);
    throw;
  }

  // Renew states:
  /*if (cid ==  helpers::base::COMMANDID__Capabilities_Exchange_Answer) {
     setState(State::Bound); // Done at sendCEA if proceed

  } else */if(cid ==  helpers::base::COMMANDID__Disconnect_Peer_Request) {
    LOGWARNING(anna::Logger::warning("DPR has been sent to the peer (diameter client)", ANNA_FILE_LOCATION));
    setState(State::WaitingDPA);
  }

  // Answers are not temporized:
  if(!isRequest) return NULL;

  // Request will have context responses:
  Response* result(NULL);
  result = Response::instance(message->getClassCode(), a_nextHopByHop - 1 /* current request sent to client */);
  result->setRequest(message);
  response_add(result);
  return result;
}



bool ServerSession::unbind(bool forceDisconnect) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "unbind", ANNA_FILE_LOCATION));

  if(a_state == State::Closed)
    return false;

  // Client socket:
  anna::comm::ClientSocket * cs = a_clientSocket;
  //anna::comm::ClientSocket * cs = const_cast<anna::comm::ClientSocket*>(a_clientSocket);

  if(forceDisconnect) {
    LOGDEBUG(anna::Logger::debug("Immediate disconnection (forceDisconnect)", ANNA_FILE_LOCATION));

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

//   if (a_state == State::Disconnecting) {
//      LOGDEBUG(
//         string msg("diameter::comm::ServerSession::unbind | ");
//         msg += asString();
//         msg += " | Disconnection already in progress !";
//         anna::Logger::debug(msg, ANNA_FILE_LOCATION);
//      );
//      return false;
//   }
//
//
//   if (a_onDisconnect == OnDisconnect::IgnorePendings) {
//      LOGDEBUG(anna::Logger::debug("Immediate disconnection (IgnorePendings)", ANNA_FILE_LOCATION));
//
//      if (cs) cs->requestClose(); // this will invoke finalize()
//
//      return true;
//   }
//
//   if (getOTARequests() == 0) { // No pendings
//      LOGDEBUG(anna::Logger::debug("No pending answers. Perform client-session close.", ANNA_FILE_LOCATION));
//
//      if (cs) cs->requestClose(); // this will invoke finalize()
//
//      return true;
//   }
  return false;
}

void ServerSession::eventPeerShutdown() throw() {
  // Inform father server:
  a_parent->eventPeerShutdown(this);
}

void ServerSession::eventRequestRetransmission(Message *request) throw() {

  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::RequestRetransmitted);
  oamModule.count(OamModule::Counter::RequestRetransmittedOnServerSession);

  // Inform father server:
  a_parent->eventRequestRetransmission(this, request);
}

void ServerSession::eventResponse(const Response& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventResponse(response);
}

void ServerSession::eventRequest(const anna::DataBlock &request) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventRequest(this, request);
}

void ServerSession::eventUnknownResponse(const anna::DataBlock& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventUnknownResponse(this, response);
}

void ServerSession::eventDPA(const anna::DataBlock& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventDPA(this, response);
}

//------------------------------------------------------------------------------------------
// Se invoca desde el diameter::comm::Receiver
//------------------------------------------------------------------------------------------
void ServerSession::receive(const anna::comm::Message& message)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "receive", ANNA_FILE_LOCATION));
  // Activity:
  updateIncomingActivityTime();
  activateTimer();
  // Command id:
  const anna::DataBlock & db = message.getBody();
  diameter::CommandId cid = codec::functions::getCommandId(db);
  bool isRequest = cid.second;
  LOGDEBUG(
    std::string msg = "Received diameter message: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);

    if( (cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) ||
        (cid.first == helpers::base::COMMANDID__Device_Watchdog_Request.first)) {
      try {
        anna::diameter::codec::Message dmsg(a_engine->getBaseProtocolCodecEngine()); dmsg.decode(db); /* decode to be traced */
      }
      catch(anna::RuntimeException &ex) {
        std::string msg = ex.getText();
        msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages full tracing";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      }
    }
  );
  // Main counters:
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(isRequest ? OamModule::Counter::RequestReceived : OamModule::Counter::AnswerReceived);
  oamModule.count(isRequest ? OamModule::Counter::RequestReceivedOnServerSession : OamModule::Counter::AnswerReceivedOnServerSession);
  // Statistic (size)
  a_parent->updateReceivedMessageSizeStatisticConcept(message.getSize(), cid); // only on reception (application could manage sent sizes)

  if(isRequest) {
    // Si recibo un request, el message solo tiene fiable el DataBlock. Como por defecto se construye como ApplicationMessage,
    // el unico caso que no cuadraria seria la recepcion de un CER. Lo que hacemos es NO PROGRESAR NUNCA un CER (*).
    // El DWR sin embargo, si podriamos progresarlo al ser de aplicacion, pero no les sirve para nada (**).
    /////////////////////////////
    // Here received a request //
    /////////////////////////////

    // Received CER
    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) {
      oamModule.count(OamModule::Counter::CERReceived);

      if(a_state == State::Bound) {
        LOGWARNING(anna::Logger::warning("Received another CER over already bound connection. Anyway, will be replied with CEA", ANNA_FILE_LOCATION));
      }

      a_cer.setBody(db);
      // Basic DRA:
      getParent()->getEngine()->manageDrDhServerSession(this, true /* register */);

      sendCEA();
      //activateTimer(); // Ya se invoca al inicio de este metodo ::receive
      //bool changes = a_parent->refreshAvailability();
      return; // (*)
    }
    // Received DWR
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) {
      oamModule.count(OamModule::Counter::DWRReceived);
      a_dwr.setBody(db);
      sendDWA();
      return; // (**)
    }
    // Received DPR
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) {
      oamModule.count(OamModule::Counter::DPRReceived);

      if(a_state == State::Bound) {
        a_dpr.setBody(db);
        setState(State::Disconnecting);
        LOGWARNING(anna::Logger::warning("DPR has been received from peer (diameter client)", ANNA_FILE_LOCATION));
        // Ignore pending on server sessions:
        /*if (getOTARequests() == 0) */sendDPA();
        return; // DPR won't be informed because virtual readDPA is available for this
      }
    }

    try {
      // application message counters
      ApplicationMessageOamModule::instantiate().count(cid.first, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Request_Received_AsServer);

      eventRequest(db);
    } catch(anna::RuntimeException& ex) {
      ex.trace();
    }

    return;
  }

  /////////////////////////////
  // Here received an answer //
  /////////////////////////////
  bool doUnbind = false;
  int resultCode = 0;

  try {
    resultCode = helpers::base::functions::getResultCode(db);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  // Received DPA
  if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) {
    oamModule.count(OamModule::Counter::DPAReceived);

    if(a_state == State::WaitingDPA) {
      if(resultCode != helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS) {
        LOGWARNING(anna::Logger::warning("Received DPA with non-success Result-Code. Ignoring and recovering Bound state", ANNA_FILE_LOCATION));
        setState(State::Bound);
      } else {
        LOGWARNING(anna::Logger::warning("Received DPA With Result-Code = DIAMETER_SUCCESS. Disconnect now.", ANNA_FILE_LOCATION));
        doUnbind = true;
      }
    }

    eventDPA(db);

  } else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) {  // non usual (server should not send DWR's)
    oamModule.count(OamModule::Counter::DWAReceived);
  }

  HopByHop hopByHop = codec::functions::getHopByHop(db); // context identification
  Response* response = response_find(hopByHop);

  // Out-of-context responses:
  if(!response) {
    // OAM
    oamModule.count(OamModule::Counter::AnswerReceivedUnknown);
    oamModule.count(OamModule::Counter::AnswerReceivedOnServerSessionUnknown);
    oamModule.activateAlarm(OamModule::Alarm::AnswerReceivedOnServerSessionUnknown);

    // application message counters
    ApplicationMessageOamModule::instantiate().count(cid.first, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Answer_UnknownReceived_AsServer);

    eventUnknownResponse(db);

    string msg(asString());
    msg += anna::functions::asString(" | Response received from client, for non registered context (HopByHop: %u)", hopByHop);
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  response->setResultCode(Response::ResultCode::Success);
  response->cancelTimer();
  LOGDEBUG(
    string msg("diameter::comm::ServerSession::receive | ");
    msg += asString();
    msg += " | Received answer";
    msg += response->asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // Statistics
  anna::Millisecond current = (anna::Millisecond)anna::functions::millisecond();
  anna::Millisecond request = response->getRequest()->getRequestTimestampMs();
  anna::Millisecond timeToAnswerMs = current - request;
  a_parent->updateProcessingTimeStatisticConcept(timeToAnswerMs, cid);
  //LOGDEBUG
  //(
  //  std::string msg = "This diameter request context lasted ";
  //  msg += anna::functions::asString(timeToAnswerMs);
  //  msg += " milliseconds at diameter client (included network time)";
  //  anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  //);
  // Progress origin for tracking purposes on asyncronous boxes with both diameter interfaces (entities and clients)
  Message * requestMessage = const_cast<Message*>(response->getRequest());
  requestMessage->setRequestClientSessionKey(response->getRequest()->getRequestClientSessionKey()); // "" means unkown/unset

  if(cid != helpers::base::COMMANDID__Disconnect_Peer_Answer) {
    // don't progress DPA: unbind is automatically performed and not open to any application decision
    try {
      response->setMessage(&db);
      // Restore received datablock
      LOGDEBUG(
        string msg("diameter::comm::ClientSession::receive | Restore answer to original request sequences (hop-by-hop = ");
        msg += anna::functions::asString(response->getRequest()->getRequestHopByHop());
        msg += ", end-to-end = ";
        msg += anna::functions::asString(response->getRequest()->getRequestEndToEnd());
        msg += ")";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      diameter::codec::functions::setHopByHop((anna::DataBlock&)db, response->getRequest()->getRequestHopByHop());
      diameter::codec::functions::setEndToEnd((anna::DataBlock&)db, response->getRequest()->getRequestEndToEnd());

      // application message counters
      ApplicationMessageOamModule::instantiate().count(cid.first, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Answer_Received_AsServer);

      eventResponse(*response);

    } catch(anna::RuntimeException& ex) {
      ex.trace();
    }
  }

  response_erase(response);

  // Unbind trigger
  if(doUnbind)
    unbind(true /* always immediate */);
}

void ServerSession::finalize() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "finalize", ANNA_FILE_LOCATION));
  // Configuration overiddings
  setOnDisconnect(OnDisconnect::IgnorePendings);
  notifyOrphansOnExpiration(false);
  // session will be idle after 'Session::finalize'
  // (anyway at 99% of the cases, server session would be idle because is not usual to send request from diameter servers)
  // Closing state won't be used on server-sessions
  Session::finalize(); // sets closed state (unbind at closeServerSession won't repeat client socket close, because returns at the beginning)
  // stops timers

  // "delete this" indirecto (seria mas elegante borrar las server sessions deprecated mediante un temporizador de purgado)
  try {
    if(idle()) getParent()->closeServerSession(this);  // http://www.parashift.com/c++-faq-lite/delete-this.html
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  // Inform father local server (availability changes):
  bool changes = getParent()->refreshAvailability();
  // OAM
  bool multipleConnections = (getParent()->getMaxConnections() > 1);
  std::string socket = anna::functions::socketLiteralAsString(getAddress(), getPort());
  OamModule &oamModule = OamModule::instantiate();

  if(multipleConnections) {
    oamModule.activateAlarm(OamModule::Alarm::LostConnectionForServerSessionAtLocalServer__s__ServerSessionId__d__, socket.c_str(), getSocketId());
    oamModule.count(OamModule::Counter::LostConnectionForServerSession);
  } else {
    oamModule.activateAlarm(OamModule::Alarm::LostConnectionForServerSessionAtLocalServer__s__, socket.c_str());
    oamModule.count(OamModule::Counter::LostConnectionForServerSession);
  }
}



void ServerSession::sendCEA()
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "sendCEA", ANNA_FILE_LOCATION));
  anna::DataBlock cea(true);
  a_engine->readCEA(cea, a_cer.getBody()); // Asume that CEA is valid ...
  // If one peer sends a CER message to another Peer and receiver does not have support for
  //
  //  1) any common application then it must return the CEA with Result-Code Avp set to DIAMETER_NO_COMMON_APPLICATION
  //     and should disconnect the transport layer connection (automatically done by diameter::comm module).
  //  2) no common security mechanism then it must return the CEA with Result-Code Avp set to DIAMETER_NO_COMMON_SECURITY
  //     and should disconnect the transport layer connection (automatically done by diameter::comm module).
  //  3) if CER is received from any unknown peer then receiver should discard the message, or send the CEA with the
  //     Result-Code Avp set to DIAMETER_UNKNOWN_PEER.

  if(cea.isEmpty()) {
    LOGDEBUG(anna::Logger::debug("Empty CEA message. Remote client never will bound this connection at application level", ANNA_FILE_LOCATION));
    LOGWARNING(anna::Logger::warning("Discarding received CER: cannot send empty CEA (consider to send CEA with Result-Code DIAMETER_UNKNOWN_PEER)", ANNA_FILE_LOCATION));
    return;
  }

  Message msgCea;
  msgCea.setBody(cea);
  send(&msgCea);
  // Here, CEA has been sent (no exception). Analize CEA Result-Code chosen:
  int resultCode = 0;

  try {
    resultCode = helpers::base::functions::getResultCode(cea);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  switch(resultCode) {
  case helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS:
    setState(State::Bound);
    break;
  case helpers::base::AVPVALUES__Result_Code::DIAMETER_NO_COMMON_APPLICATION:
    LOGWARNING(anna::Logger::warning("DIAMETER_NO_COMMON_APPLICATION CEA Result-Code implies unbinding connection", ANNA_FILE_LOCATION));
    unbind(true /* always immediate */); // no delegamos en un planning o similar
    // Realmente el cliente diameter tambien deberia cerrar la conexion al recibir este Result-Code
    break;
  case helpers::base::AVPVALUES__Result_Code::DIAMETER_NO_COMMON_SECURITY:
    LOGWARNING(anna::Logger::warning("DIAMETER_NO_COMMON_SECURITY CEA Result-Code implies unbinding connection", ANNA_FILE_LOCATION));
    unbind(true /* always immediate */); // no delegamos en un planning o similar
    // Realmente el cliente diameter tambien deberia cerrar la conexion al recibir este Result-Code
    break;
    // ...
    // ...
  }
}

void ServerSession::sendDWA()
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "sendDWA", ANNA_FILE_LOCATION));
  anna::DataBlock dwa(true);
  a_engine->readDWA(dwa, a_dwr.getBody()); // Asume that DWA is valid ...

  if(dwa.isEmpty())
    throw anna::RuntimeException("This diameter agent defines an empty DWA message. Remote client never will validate this connection health", ANNA_FILE_LOCATION);

  Message msgDwa;
  msgDwa.setBody(dwa);
  send(&msgDwa);
}


//-------------------------------------------------------------------------
// Se invoca desde diameter::comm::Timer
//-------------------------------------------------------------------------
void ServerSession::expireResponse(diameter::comm::Response* response)
throw() {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "expireResponse", ANNA_FILE_LOCATION));
  Session::expireResponse(response);
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::RequestSentExpired);
  oamModule.count(OamModule::Counter::RequestSentOnServerSessionExpired);
  oamModule.activateAlarm(OamModule::Alarm::RequestSentOnServerSessionExpired);
//   // "delete this" indirecto (seria mas elegante borrar las server sessions deprecated mediante un temporizador de purgado)
//   if (idle()) a_parent->eraseServerSession(*a_clientSocket); // http://www.parashift.com/c++-faq-lite/delete-this.html
}

std::string ServerSession::asString() const
throw() {
  string result = Session::asString();
  result += " | Parent Local Server: ";
  result += anna::functions::socketLiteralAsString(getAddress(), getPort());
  result += " | Client Socket: ";
  result += a_clientSocket->asString();
  // Diferente del timeout de ApplicationMessage:
  result += " | Allowed inactivity time: ";
  result += getTimeout().asString();
  result += " | Deprecated: ";
  result += (a_deprecated ? "yes" : "no");
  return result += " }";
}

anna::xml::Node* ServerSession::asXML(anna::xml::Node* parent) const
throw() {
  anna::xml::Node* result = Session::asXML(parent);
  parent->createChild("diameter.comm.ServerSession");
  result->createAttribute("ParentLocalServer", anna::functions::socketLiteralAsString(getAddress(), getPort()));
  result->createAttribute("ClientSocket", a_clientSocket->asString());
  // Diferente del timeout de ApplicationMessage:
  result->createAttribute("AllowedInactivityTime", getTimeout().asString());
  result->createAttribute("Deprecated", a_deprecated ? "yes" : "no");
  return result;
}


//------------------------------------------------------------------------------
//------------------------------------------------------ ServerSession::expire()
//------------------------------------------------------------------------------
void ServerSession::expire(anna::timex::Engine *timeController) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "expire (inactivity check timer)", ANNA_FILE_LOCATION));
  LOGWARNING(anna::Logger::warning("Detecting anomaly (too inactivity time) over server session. Resetting", ANNA_FILE_LOCATION));
  // OAM
  bool multipleConnections = (getParent()->getMaxConnections() > 1);
  std::string socket = anna::functions::socketLiteralAsString(getAddress(), getPort());
  OamModule &oamModule = OamModule::instantiate();

  if(multipleConnections) {
    oamModule.activateAlarm(OamModule::Alarm::UnbindConnectionForServerSessionAtLocalServer__s__ServerSessionId__d__DueToInactivityTimeAnomaly, socket.c_str(), getSocketId());
    oamModule.count(OamModule::Counter::UnbindConnectionForServerSessionDueToInactivityTimeAnomaly);
  } else {
    oamModule.activateAlarm(OamModule::Alarm::UnbindConnectionForServerSessionAtLocalServer__s__DueToInactivityTimeAnomaly, socket.c_str());
    oamModule.count(OamModule::Counter::UnbindConnectionForServerSessionDueToInactivityTimeAnomaly);
  }

  unbind(true /* always immediate */); // no delegamos en un planning o similar
}

void ServerSession::setAllowedInactivityTime(const anna::Millisecond & allowedInactivityTime) throw() {
  setTimeout(allowedInactivityTime);
}

//------------------------------------------------------------------------------
//---------------------------------- ServerSession::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void ServerSession::updateIncomingActivityTime() throw() {
  Session::updateIncomingActivityTime();
  a_parent->updateIncomingActivityTime();
}


//------------------------------------------------------------------------------
//---------------------------------- ServerSession::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void ServerSession::updateOutgoingActivityTime(void) throw() {
  Session::updateOutgoingActivityTime();
  a_parent->updateOutgoingActivityTime();
}


//------------------------------------------------------------------------------
//----------------------------------------------- ServerSession::countSendings()
//------------------------------------------------------------------------------
void ServerSession::countSendings(const diameter::CommandId & cid, unsigned int aid, bool ok)throw() {
  OamModule &oamModule = OamModule::instantiate();
  ApplicationMessageOamModule &appMsgOamModule = ApplicationMessageOamModule::instantiate();

  bool isRequest = cid.second;

  if(ok) {
    // Main counters:
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOK : OamModule::Counter::AnswerSentOK);
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOnServerSessionOK : OamModule::Counter::AnswerSentOnServerSessionOK);

    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Answer) oamModule.count(OamModule::Counter::CEASentOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) oamModule.count(OamModule::Counter::DWASentOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) oamModule.count(OamModule::Counter::DWRSentOK);  // not usual
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) oamModule.count(OamModule::Counter::DPASentOK);
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) oamModule.count(OamModule::Counter::DPRSentOK);
    // Application messages:
    else {
      appMsgOamModule.count(cid.first, aid, isRequest ? ApplicationMessageOamModule::Counter::Request_SentOK_AsServer : ApplicationMessageOamModule::Counter::Answer_SentOK_AsServer);
    }
  } else {
    // Main counters:
    oamModule.count(isRequest ? OamModule::Counter::RequestSentNOK : OamModule::Counter::AnswerSentNOK);
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOnServerSessionNOK : OamModule::Counter::RequestSentOnServerSessionNOK);

    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Answer) oamModule.count(OamModule::Counter::CEASentNOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) oamModule.count(OamModule::Counter::DWASentNOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) oamModule.count(OamModule::Counter::DWRSentNOK);  // not usual
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) oamModule.count(OamModule::Counter::DPASentNOK);
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) oamModule.count(OamModule::Counter::DPRSentNOK);
    // Application messages:
    else {
      appMsgOamModule.count(cid.first, aid, isRequest ? ApplicationMessageOamModule::Counter::Request_SentNOK_AsServer : ApplicationMessageOamModule::Counter::Answer_SentNOK_AsServer);
    }
  }
}


