// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/codec/EngineImpl.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/Avp.hpp>
#include <anna/diameter/stack/Engine.hpp>
#include <anna/diameter/stack/Dictionary.hpp>
#include <anna/diameter/stack/Command.hpp>
#include <anna/diameter/internal/sccs.hpp>
#include <anna/diameter/defines.hpp>

#include <anna/xml/xml.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/mt/Guard.hpp>


using namespace anna::diameter::codec;


//------------------------------------------------------------------------------
//----------------------------------------------------- EngineImpl::EngineImpl()
//------------------------------------------------------------------------------
EngineImpl::EngineImpl(const char* className, const stack::Dictionary * dictionary) :
  anna::Component(className),
  a_dictionary(dictionary),
  a_validationDepth(ValidationDepth::FirstError),
  a_validationMode(ValidationMode::AfterDecoding),
  a_singleFailedAVP(true),
  a_ignoreFlags(false),
  a_fixMode(FixMode::BeforeEncoding) {
  anna::diameter::sccs::activate();
  anna::xml::functions::initialize();
}


//------------------------------------------------------------------------------
//------------------------------------------------------ EngineImpl::createAvp()
//------------------------------------------------------------------------------
Avp* EngineImpl::createAvp(const AvpId *id) throw(anna::RuntimeException) {
  Avp *result(NULL);
  anna::Guard guard(this, "diameter::codec::EngineImpl::createAvp");

  if((result = allocateAvp()) == NULL)
    throw anna::RuntimeException("diameter::codec::EngineImpl::allocateAvp returns NULL", ANNA_FILE_LOCATION);

  // Sets engine
  result->setEngine((Engine*)this);

  //result->clear(); better clear this at releaseAvp(), see class-help implementation example
  if(id) result->setId(*id);

  return result;
}


//------------------------------------------------------------------------------
//-------------------------------------------------- EngineImpl::createMessage()
//------------------------------------------------------------------------------
Message* EngineImpl::createMessage(const CommandId *id) throw(anna::RuntimeException) {
  Message *result(NULL);
  anna::Guard guard(this, "diameter::codec::EngineImpl::createMessage");

  if((result = allocateMessage()) == NULL)
    throw anna::RuntimeException("diameter::codec::EngineImpl::allocateMessage returns NULL", ANNA_FILE_LOCATION);

  // Sets engine
  result->setEngine((Engine*)this);

  //result->clear(); better clear this at releaseMessage(), see class-help implementation example
  if(id) result->setId(*id);

  return result;
}


//------------------------------------------------------------------------------
//-------------------------------------------------- EngineImpl::createMessage()
//------------------------------------------------------------------------------
Message *EngineImpl::createMessage(const std::string & xmlPathFile) throw(anna::RuntimeException) {
  Message *result = createMessage();
  result->loadXML(xmlPathFile);
  return result;
}



std::string EngineImpl::asString(void) const throw() {
  std::string result = anna::Component::asString();
  result += "\nValidationDepth: ";
  result += asText(a_validationDepth);
  result += "\nValidationMode: ";
  result += asText(a_validationMode);
  result += "\nSingle Failed-AVP: ";
  result += a_singleFailedAVP ? "yes" : "no";
  result += "\nIgnore flags: ";
  result += a_ignoreFlags ? "yes" : "no";
  result += "\nFixMode: ";
  result += asText(a_fixMode);
  result += "\nActivated Dictionary: ";
  result += a_dictionary ? (a_dictionary->getName()) : "<null>";
  return result;
}

//------------------------------------------------------------------------------
//---------------------------------------------------------- EngineImpl::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* EngineImpl::asXML(anna::xml::Node* parent) const
throw() {
  parent = anna::Component::asXML(parent);
  anna::xml::Node* result = parent->createChild("diameter.codec.EngineImpl");
  result->createAttribute("ValidationDepth", asText(a_validationDepth));
  result->createAttribute("ValidationMode", asText(a_validationMode));
  result->createAttribute("SingleFailedAVP", a_singleFailedAVP ? "yes" : "no");
  result->createAttribute("IgnoreFlags", a_ignoreFlags ? "yes" : "no");
  result->createAttribute("FixMode", asText(a_fixMode));
  anna::xml::Node* dictionary = result->createChild("EngineImpl.ActivatedDictionary");

  if(a_dictionary) a_dictionary->asXML(result);

  return result;
}

//------------------------------------------------------------------------------
//--------------------------------------------------------- EngineImpl::asText()
//------------------------------------------------------------------------------
const char* EngineImpl::asText(const ValidationDepth::_v vd)
throw() {
  static const char* text [] = { "Complete", "FirstError" };
  return text [vd];
}

//------------------------------------------------------------------------------
//--------------------------------------------------------- EngineImpl::asText()
//------------------------------------------------------------------------------
const char* EngineImpl::asText(const ValidationMode::_v vm)
throw() {
  static const char* text [] = { "BeforeEncoding", "AfterDecoding", "Always", "Never" };
  return text [vm];
}

//------------------------------------------------------------------------------
//--------------------------------------------------------- EngineImpl::asText()
//------------------------------------------------------------------------------
const char* EngineImpl::asText(const FixMode::_v fm)
throw() {
  static const char* text [] = { "BeforeEncoding", "AfterDecoding", "Always", "Never" };
  return text [fm];
}

//------------------------------------------------------------------------------
//---------------------------------------------- EngineImpl::validationAnomaly()
//------------------------------------------------------------------------------
void EngineImpl::validationAnomaly(const std::string & description) const throw(anna::RuntimeException) {
  std::string msg = "Validation error: ";

  if(a_validationDepth == ValidationDepth::FirstError)
    throw anna::RuntimeException(msg + description, ANNA_FILE_LOCATION);

  LOGWARNING(anna::Logger::warning(msg + description, ANNA_FILE_LOCATION));
}


//------------------------------------------------------------------------------
//--------------------------------------------------- EngineImpl::avpIdForName()
//------------------------------------------------------------------------------
anna::diameter::AvpId EngineImpl::avpIdForName(const char * name) throw(anna::RuntimeException) {
  if(!name)
    throw anna::RuntimeException("Provided NULL Avp Logical Name", ANNA_FILE_LOCATION);

  if(!a_dictionary) {
    std::string msg = "Cannot find identifier '"; msg += name;
    msg += "' for the avp: no dictionary available";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  const stack::Avp * stackAvp = a_dictionary->getAvp(name);

  if(!stackAvp) {
    std::string msg = "Cannot find identifier '"; msg += name;
    msg += "' for the avp at the available dictionary";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return (stackAvp->getId());
}


//------------------------------------------------------------------------------
//----------------------------------------------- EngineImpl::commandIdForName()
//------------------------------------------------------------------------------
anna::diameter::CommandId EngineImpl::commandIdForName(const char * name) throw(anna::RuntimeException) {
  if(!name)
    throw anna::RuntimeException("Provided NULL Command Logical Name", ANNA_FILE_LOCATION);

  if(!a_dictionary) {
    std::string msg = "Cannot find identifier '"; msg += name;
    msg += "' for the command: no dictionary available";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  const stack::Command * stackCommand = a_dictionary->getCommand(name);

  if(!stackCommand) {
    std::string msg = "Cannot find identifier '"; msg += name;
    msg += "' for the command at the available dictionary";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return (stackCommand->getId());
}

