// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <time.h>
#include <signal.h>

#include <ldap.h>
#include <lber.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

#include <anna/xml/Node.hpp>

#include <anna/ldap/internal/sccs.hpp>
#include <anna/ldap/Engine.hpp>
#include <anna/ldap/Session.hpp>

using namespace std;
using namespace anna;
using namespace anna::ldap;

Engine::Engine() :
  app::Component(getClassName()),
  a_autoBind(true) {
  ldap::sccs::activate();
  sigset(SIGALRM, alarmnCatcher);
}

Session* Engine::createSession(const char* url, const char* user, const char* password, const int category)
throw(RuntimeException) {
  ldap::Session* result(NULL);
  Guard guard(this, "ldap::Engine::createSession");
  url = completeURL(url);
  session_iterator ii = session_find(url, (user == NULL) ? "" : user);

  if(ii == session_end()) {
    if((result = allocateSession(category)) == NULL)
      throw RuntimeException("ldap::Engine::allocateSession returns NULL", ANNA_FILE_LOCATION);

    result->a_category = category;
    result->a_url = url;
    result->a_externalID = -1;

    if(user && *user != 0)
      result->a_user = user;
    else
      result->a_user.clear();

    if(password && *password != 0)
      result->a_password = password;
    else
      result->a_password.clear();

    session_key key(result->a_url, result->a_keymap = result->a_user);
    a_sessions.insert(session_value_type(key, result));
    LOGDEBUG(
      string msg("ldap::Engine::createSession | ");
      msg += result->asString();
      msg += functions::asText(" | AutoBind: ", a_autoBind);
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  } else
    result = session(ii);

  if(result->getState() == Session::State::Closed && a_autoBind == true)
    result->bind();

  return result;
}

Session* Engine::createSession(const char* url, const int id, const char* user, const char* password, const int category)
throw(RuntimeException) {
  ldap::Session* result(NULL);
  Guard guard(this, "ldap::Engine::createSession");
  url = completeURL(url);
  session_iterator ii = session_find(url, id);

  if(ii == session_end()) {
    if((result = allocateSession(category)) == NULL)
      throw RuntimeException("ldap::Engine::allocateSession returns NULL", ANNA_FILE_LOCATION);

    result->a_category = category;
    result->a_url = url;
    result->a_externalID = id;

    if(user && *user != 0)
      result->a_user = user;
    else
      result->a_user.clear();

    if(password && *password != 0)
      result->a_password = password;
    else
      result->a_password.clear();

    session_key key(result->a_url, result->a_keymap = anna::functions::asString(id));
    a_sessions.insert(session_value_type(key, result));
    LOGDEBUG(
      string msg("ldap::Engine::createSession | ");
      msg += result->asString();
      msg += functions::asText(" | AutoBind: ", a_autoBind);
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  } else {
    result = session(ii);

    if((result->getUser() != user) || (result->getPassword() != password)) {
      LOGWARNING(
        std::string msg = "Returned session already existed but with different credentiales regarding provided ones. Reuse could be inappropiate.";
        Logger::warning(msg, ANNA_FILE_LOCATION);
      );
    }
  }

  if(result->getState() == Session::State::Closed && a_autoBind == true)
    result->bind();

  return result;
}

Session* Engine::findSession(const char* url, const char* user, Exception::Mode::_v emode)
throw(RuntimeException) {
  Guard guard(this, "ldap::Engine::findSession");
  url = completeURL(url);
  session_iterator ii = session_find(url, user);

  if(ii != session_end())
    return session(ii);

  if(emode != Exception::Mode::Ignore) {
    string msg("ldap::Engine::findSession | URL: ");
    msg += url;
    msg += " | User: ";
    msg += user;
    msg += " | Session not found";
    RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}

Session* Engine::findSession(const char* url, const int id, Exception::Mode::_v emode)
throw(RuntimeException) {
  Guard guard(this, "ldap::Engine::findSession (int)");
  session_iterator ii = session_find(url, id);

  if(ii != session_end())
    return session(ii);

  if(emode != Exception::Mode::Ignore) {
    string msg("ldap::Engine::findSession | URL: ");
    msg += url;
    msg += functions::asText(" | ID: ", id);
    msg += " | Session not found";
    RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}

void Engine::closeSession(Session* session)
throw(RuntimeException) {
  if(session == NULL)
    return;

  LOGDEBUG(
    string msg("ldap::Engine::closeSession | ");
    msg += session->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  Guard guard(this, "ldap::Engine::closeSession");
  session_iterator ii = session_find(session->a_url, session->a_keymap);

  if(ii == session_end())
    return;

  try {
    session->unbind();
    releaseSession(session);
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  a_sessions.erase(ii);
}

void Engine::do_stop()
throw() {
  LOGMETHOD(TraceMethod tttm("anna::ldap::Engine", "do_stop", ANNA_FILE_LOCATION));

  for(session_iterator ii = session_begin(), maxii = session_end(); ii != maxii; ii ++)
    session(ii)->unbind();
}

xml::Node* Engine::asXML(xml::Node* parent) const
throw() {
  parent = app::Component::asXML(parent);
  xml::Node* result = parent->createChild("ldap.Engine");
  const Session* session;
  result->createAttribute("AutoBind", functions::asString(a_autoBind));

  for(const_session_iterator ii = session_begin(), maxii = session_end(); ii != maxii; ii ++) {
    Guard guard(session = Engine::session(ii));
    session->asXML(result);
  }

  return result;
}

Engine::session_iterator Engine::session_find(const char* url, const int id)
throw() {
  return a_sessions.find(session_key(url, anna::functions::asString(id)));
}

/*static*/
int Engine::setDebugLevel(const int level)
throw(RuntimeException) {
  int result(-1);
#ifdef LDAP_OPT_DEBUG_LEVEL
  int _level = htonl(level);

  if(ldap_get_option(NULL, LDAP_OPT_DEBUG_LEVEL, &result) != LDAP_OPT_SUCCESS)
    throw RuntimeException("Can not get LDAP_OPT_DEBUG_LEVEL", ANNA_FILE_LOCATION);

  if(ldap_set_option(NULL, LDAP_OPT_DEBUG_LEVEL, &_level) != LDAP_OPT_SUCCESS)
    throw RuntimeException("Can not set LDAP_OPT_DEBUG_LEVEL", ANNA_FILE_LOCATION);

  ber_set_option(NULL, LBER_OPT_DEBUG_LEVEL, &_level);
#else
  Logger::error("Can not set debug level", ANNA_FILE_LOCATION);
#endif
  return result;
}

const char* Engine::completeURL(const char* url)
throw() {
  static const char* protocol = "ldap://";
  static const int protocolLen = anna_strlen(protocol);

  if(anna_strstr(url, "://") == 0) {
    a_auxURL = protocol;
    a_auxURL += url;
    return a_auxURL.c_str();
  }

  return url;
}

// static
void Engine::alarmnCatcher(int)
throw() {
}
