// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <time.h>
#include <ldap.h>
#include <poll.h>
#include <signal.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/app/functions.hpp>

#include <anna/comm/Communicator.hpp>

#include <anna/ldap/Session.hpp>
#include <anna/ldap/ResultCode.hpp>
#include <anna/ldap/Attribute.hpp>
#include <anna/ldap/Response.hpp>
#include <anna/ldap/Request.hpp>

#include <anna/ldap/internal/Exception.hpp>

using namespace std;
using namespace anna;
using namespace anna::ldap;

//static
const Millisecond Session::DefaultTimeout(1000);

Session::Session() :
  comm::Handler(comm::Handler::Type::Custom, comm::Handler::Support::None),
  a_state(State::Closed),
  a_ldap(NULL),
  a_defer(Option::Defer::Never),
  a_referral(Option::Referral::Off),
  a_externalID(-1) {
  comm::Handler::a_communicator = app::functions::component <comm::Communicator> (ANNA_FILE_LOCATION);

  for(int i = ClassCode::Min; i < ClassCode::Max; i ++)
    a_timeouts [i] = DefaultTimeout;

  a_networkTimeout.tv_sec = -1; a_networkTimeout.tv_usec = 0;
}

//---------------------------------------------------------------------------------
// Se invoca desde el ldap::Engine
//---------------------------------------------------------------------------------
void Session::bind()
throw(RuntimeException) {
  if(a_state != State::Closed)
    return;

  LOGDEBUG(
    string msg("ldap::Session::bind | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  Guard guard(this, "ldap::Session::bind");
  LDAP* handle = NULL;
  int aux;
  ResultCode resultCode;
  Response* response(NULL);

  try {
    resultCode = ldap_initialize(&handle, a_url.c_str());

    if(resultCode != LDAP_SUCCESS)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    aux = LDAP_VERSION3;
    resultCode = ldap_set_option(handle, LDAP_OPT_PROTOCOL_VERSION, &aux);

    if(resultCode.isOk() == false)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    aux = 1;
    resultCode = ldap_set_option(handle, LDAP_OPT_RESULT_CODE, &aux);

    if(resultCode.isOk() == false)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    switch(a_defer) {
    case Session::Option::Defer::Never: aux = LDAP_DEREF_NEVER; break;
    case Session::Option::Defer::Searching: aux = LDAP_DEREF_SEARCHING; break;
    case Session::Option::Defer::Finding: aux = LDAP_DEREF_FINDING; break;
    case Session::Option::Defer::Always: aux = LDAP_DEREF_ALWAYS; break;
    }

    resultCode = ldap_set_option(handle, LDAP_OPT_DEREF, &aux);

    if(resultCode.isOk() == false)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    resultCode = ldap_set_option(handle, LDAP_OPT_REFERRALS, (a_referral == Session::Option::Referral::On) ? LDAP_OPT_ON : LDAP_OPT_OFF);

    if(resultCode.isOk() == false)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    if(a_networkTimeout.tv_sec != -1) {
      resultCode = ldap_set_option(handle, LDAP_OPT_NETWORK_TIMEOUT, &a_networkTimeout);

      if(resultCode.isOk() == false)
        throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));
    }
  } catch(RuntimeException&) {
    if(handle != NULL)
      ldap_memfree(&handle);

    throw;
  }

  try {
    resultCode = ldap_simple_bind(handle, Request::asCString(a_user), Request::asCString(a_password));

    if(resultCode < LDAP_SUCCESS)
      throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

    int fd = -1;
    ldap_get_option(handle, LDAP_OPT_DESC, &fd);
    setfd(fd);
    app::functions::component <comm::Communicator> (ANNA_FILE_LOCATION)->attach(this);
    a_ldap = handle;
    a_state = State::WaitingBind;
    response_add(Response::instance(ClassCode::Bind, resultCode.getValue()));
  } catch(RuntimeException&) {
    ldap_unbind(handle);
    throw;
  }
}

const Response* Session::send(const Request* request)
throw(RuntimeException) {
  if(a_state == State::Closed) {
    string msg(asString());
    msg += " | Session::bind must be called";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_state == State::WaitingBind) {
    string msg(asString());
    msg += " | Waiting for connection ack";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  Response* result(NULL);
  LOGDEBUG(
    string msg("ldap::Session::send | ");
    msg += asString();
    msg += " | ";
    msg += request->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  Guard guard(this, "ldap::Session::send");
  ResultCode resultCode = request->send(*this);

  if(resultCode < 0) {
    resultCode.extractResultCode(this);
    throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));
  }

  result = Response::instance(request->getClassCode(), resultCode.getValue());
  result->setRequest(request);
  response_add(result);
  return result;
}

void Session::unbind()
throw(RuntimeException) {
  if(a_state == State::Closed)
    return;

  /**
   * Esta invocacin terminar invocando al mtodo ldap::Session::finalize
   */
  a_communicator->detach(this);
}

//------------------------------------------------------------------------------------------
// Se invoca desde el comm::Communicator cuando se detecta activada sobre el fd's asociado
// a la ldap::Session
//------------------------------------------------------------------------------------------
void Session::apply()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod("ldap::Session", "apply", ANNA_FILE_LOCATION));
  LDAP* handle = (LDAP*) a_ldap;
  LDAPMessage* hmessage(NULL);
  ResultCode resultCode;
  // La SIGALRM se tratar en Engine::alarmnCatcher => por ahora slo se ignora
  ualarm(100000, 0);
  // El cdigo que recibe la respuesta no est protegido ante la aparecin de una seal.
  // Cuando este mtodo se ejecuta es porque ya sabemos que ha algo => no habr espera
  // Cuando el MSGID de la respuesta recibida no se encuentra en la lista de respuestas esperadas la OpenLDAP
  // invoca a un select en el que supuestamente se pone a esperar una respuesta 'esperada' ... esper 100 ms
  // como mximo ya que la seal que ponemos lo sacar de la espera.
  resultCode = ldap_result(handle, LDAP_RES_ANY, LDAP_MSG_ONE, NULL, &hmessage);
  int xerrno = errno;
  ualarm(0, 0);

  if(resultCode.getValue() == -1) {
    bool disconnect(false);

    if(resultCode.extractResultCode(this) == false)
      disconnect = true;
    else if(resultCode.isServerDown() == true)
      disconnect = true;
    else if(resultCode.isConnectError() == true)
      disconnect = true;

    // Si espera fue interrumpida por una excepcin el cdigo de erro LDAP ser -1, pero errno ser EINTR
    // En ese caso slo ignoramos el mensaje, pero no cerramos la conexin.
    if(disconnect == true && xerrno == EINTR)
      disconnect = false;

    ldap::Exception lex(this, resultCode, ANNA_FILE_LOCATION);

    if(disconnect == true)
      lex.trace();

    eventResponseError(resultCode, disconnect);

    if(disconnect == true)
      a_communicator->detach(this);

    return;
  }

  IdMessage idMessage = ldap_msgid(hmessage);

  try {
    switch(resultCode.getValue()) {
    case LDAP_RES_BIND: receiveBind(idMessage, hmessage); break;
    case LDAP_RES_SEARCH_ENTRY: receiveEntry(idMessage, hmessage); break;
    case LDAP_RES_SEARCH_REFERENCE: receiveReference(idMessage, hmessage); break;
    case LDAP_RES_SEARCH_RESULT: receiveResult(idMessage, hmessage); break;
    default:
      LOGWARNING(
        string msg(asString());
        msg += functions::asText(" | IdMessage: ", idMessage);
        msg += " | ";
        msg += resultCode.asString();
        Logger::warning(msg, ANNA_FILE_LOCATION);
      );
      break;
    }

    ldap_msgfree(hmessage);
  } catch(RuntimeException&) {
    ldap_msgfree(hmessage);
    throw;
  }
}

/**
 * invocado por el comm::Communicator::detach.
 */
void Session::finalize()
throw() {
  LDAP* handle = (LDAP*) a_ldap;
  a_state = State::Closed;

  if(handle != NULL) {
    ldap_unbind(handle);
    a_ldap = NULL;
  }

  eventServerShutdown();
  ResultCode resultCode;
  Response* response;
  /*
   * Notifica la finalizacin de las respuestas pendientes de recibir
   */
  resultCode = LDAP_UNAVAILABLE;

  for(response_iterator ii = response_begin(), maxii = response_end(); ii != maxii; ii ++) {
    response = Session::response(ii);
    response->setResultCode(resultCode);
    response->cancelTimer();

    try {
      eventResponse(*response);
    } catch(RuntimeException& ex) {
      ex.trace();
    }

    Response::release(response);
  }

  a_responses.clear();
}

void Session::receiveBind(const IdMessage idMessage, Session::HandleMessage _hmessage)
throw(RuntimeException) {
  LDAP* handle = (LDAP*) a_ldap;
  LDAPMessage* hmessage = (LDAPMessage*) _hmessage;

  // Si se invoca al unbind despus de hacer el bind ... y llega la respuesta
  if(a_state != State::WaitingBind) {
    string msg("ldap::Session::receiveBind | ");
    msg += asString();
    msg += " | Unexpected Bind-response";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  ResultCode resultCode;
  int error = LDAP_SUCCESS;
  Response* response = response_find(idMessage);
  LOGDEBUG(
    string msg("ldap::Session::receiveBind | ");
    msg += asString();
    msg += " | ";
    msg += response->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  response->cancelTimer();
  const int ldap_result  = ldap_parse_result(handle, hmessage, &error, NULL, NULL, NULL, NULL, 0);
  bool unbindAfterDone = false;
  resultCode.setValue(ldap_result, error);
  LOGDEBUG(
    string msg("ldap::Session::receiveBind | ");
    msg += asString();
    msg += " | ";
    msg += resultCode.asString();
    msg += functions::asText(" | LDAP Result: ", ldap_result);
    msg += functions::asText(" | LDAP Error: ", error);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  )

  if(resultCode.isOk() == false) {
    Exception ldapex(this, resultCode = error, ANNA_FILE_LOCATION);
    ldapex.trace();
    response->setResultCode(resultCode);
    unbindAfterDone = true;
  } else
    a_state = State::Bound;

  try {
    eventResponse(*response);
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  response_erase(response);

  if(unbindAfterDone == true)
    unbind();
}

//------------------------------------------------------------------------------------------------
// (1) Necesario para detectar el final de los atributos
//------------------------------------------------------------------------------------------------
void Session::receiveEntry(const IdMessage idMessage, Session::HandleMessage _hmessage)
throw(RuntimeException) {
  LDAP* handle = (LDAP*) a_ldap;
  LDAPMessage* hmessage = (LDAPMessage*) _hmessage;
  BerElement *ber(NULL);
  BerValue name;
  Response* response = response_find(idMessage);
  LOGDEBUG(
    string msg("ldap::Session::receiveEntry | ");
    msg += response->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  ResultCode resultCode(ldap_get_dn_ber(handle, hmessage, &ber, &name));

  if(resultCode.isOk() == false) {
    Exception ldapex(this, resultCode , ANNA_FILE_LOCATION);
    ldapex.trace();
    response->setResultCode(resultCode);
    eventIntermediateResponseError(*response);
    return;
  }

  BerValue* values;
  string value;
  Attribute* attribute;
  response->setName(value.assign(name.bv_val, name.bv_len));
  LOGDEBUG(
    string msg("ldap::Session::receiveEntry | DN: ");
    msg += value;
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  while(ldap_get_attribute_ber(handle, hmessage, ber, &name, &values) == LDAP_SUCCESS) {
    if(name.bv_val == NULL)                             // (1)
      break;

    attribute = response->createAttribute(value.assign(name.bv_val, name.bv_len));
    LOGDEBUG(
      string msg("ldap::Session::receiveEntry | Attribute: ");
      msg += value;
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );

    if(values) {
      for(int i = 0; values [i].bv_val != NULL; i ++) {
        attribute->add(value.assign(values [i].bv_val, values [i].bv_len));
        LOGDEBUG(
          string msg("ldap::Session::receiveEntry | Value: ");
          msg += value;
          Logger::debug(msg, ANNA_FILE_LOCATION);
        );
      }

      ber_memfree(values);
    }
  }

  ber_free(ber, 0);
}

void Session::receiveReference(const IdMessage idMessage, Session::HandleMessage _hmessage)
throw(RuntimeException) {
  LDAP* handle = (LDAP*) a_ldap;
  LDAPMessage* hmessage = (LDAPMessage*) _hmessage;
  char **values(NULL);
  Response* response = response_find(idMessage);
  LOGDEBUG(
    string msg("ldap::Session::receiveReference | ");
    msg += response->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  ResultCode resultCode(ldap_parse_reference(handle, hmessage, &values, NULL, 0));

  if(resultCode.isOk() == false) {
    Exception ldapex(this, resultCode , ANNA_FILE_LOCATION);
    ldapex.trace();
    response->setResultCode(resultCode);
    eventIntermediateResponseError(*response);
    return;
  }

  if(values != NULL) {
    for(int i = 0; values [i] != NULL; i ++)
      response->createReferral(values [i]);

    ber_memvfree((void**) values);
  }
}

void Session::receiveResult(const IdMessage idMessage, Session::HandleMessage _hmessage)
throw(RuntimeException) {
  LDAP* handle = (LDAP*) a_ldap;
  LDAPMessage* hmessage = (LDAPMessage*) _hmessage;
  Response* response = response_find(idMessage);
  LOGDEBUG(
    string msg("ldap::Session::receiveResult | ");
    msg += asString();
    msg += " | ";
    msg += response->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  int error = LDAP_SUCCESS;
  char **values(NULL);
  ResultCode resultCode;
  bool isOk = true;
  response->cancelTimer();
  const int ldap_result = ldap_parse_result(handle, hmessage, &error, NULL, NULL, &values, NULL, 0);
  resultCode.setValue(ldap_result, error);
  LOGDEBUG(
    string msg("ldap::Session::receiveResult | ");
    msg += asString();
    msg += " | ";
    msg += resultCode.asString();
    msg += functions::asText(" | LDAP Result: ", ldap_result);
    msg += functions::asText(" | LDAP Error: ", error);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  )

  if(resultCode.isOk() == false) {
    Exception ldapex(this, resultCode = error, ANNA_FILE_LOCATION);
    ldapex.trace();
    response->setResultCode(resultCode);
  } else if(values != NULL) {
    for(int i = 0; values [i] != NULL; i ++)
      response->createReferral(values [i]);

    ldap_value_free(values);
  }

  try {
    eventResponse(*response);
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  response_erase(response);
}

//-------------------------------------------------------------------------
// Se invoca desde ldap::timer::Prototype::expire
//-------------------------------------------------------------------------
void Session::expireResponse(ldap::Response* response)
throw() {
  LDAP* handle = (LDAP*) a_ldap;
  ResultCode resultCode;
  /*
   * Si lo que ha caducado es una peticin de Bind hay que cerrar la conexin
   * y liberar los recursos.
   */
  bool unbindAfterDone(true);

  if(response->getClassCode() != ClassCode::Bind) {
    if(response->getRequest()->getOnExpiry() == Request::OnExpiry::Abandon) {
      resultCode = ldap_abandon(handle, response->getIdMessage());

      if(resultCode.isOk() == false) {
        Exception ldapex(this, resultCode, ANNA_FILE_LOCATION);
        ldapex.trace();
      }
    }

    unbindAfterDone = false;
  }

  response->setResultCode(resultCode = LDAP_TIMEOUT);

  try {
    eventResponse(*response);
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  response_erase(response);

  if(unbindAfterDone == true)
    unbind();
}

void Session::response_add(Response* response)
throw() {
  a_responses.add(response);
  response->setSession(this);

  try {
    response->activateTimer();
  } catch(anna::Exception& ex) {
    ex.trace();
  }
}

void Session::response_erase(Response* response)
throw() {
  a_responses.erase(response);
  Response::release(response);
}

Response* Session::response_find(const IdMessage idMessage)
throw(RuntimeException) {
  ldap::Response* result = a_responses.find(idMessage);

  if(result == NULL) {
    string msg(asString());
    msg += functions::asText(" | IdMessage: ", idMessage);
    msg += " | Message not registered at session";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return result;
}

std::string Session::asString() const
throw() {
  string result("ldap::Session { ");
  result += comm::Handler::asString();
  result += " | State: ";
  result += asText(a_state);
  result += " | URL: ";
  result += a_url;
  result += " | User: ";
  result += Request::asText(a_user);

  if(a_externalID != -1)
    result += functions::asText(" | ExternalID: ", a_externalID);

  if(hasNetworkTimeout()) {
    result += functions::asText(" | Network timeout { Sec: ", (int) a_networkTimeout.tv_sec);
    result += functions::asText(" | uSec: ", (int) a_networkTimeout.tv_usec);
    result += " }";
  }

  return result += " }";
}

xml::Node* Session::asXML(xml::Node* parent) const
throw() {
  parent = comm::Handler::asXML(parent);
  xml::Node* result = parent->createChild("ldap.Session");
  result->createAttribute("State", asText(a_state));
  result->createAttribute("User", Request::asText(a_user));
  result->createAttribute("N", a_responses.size());
  result->createChild("URL")->createAttribute("Value", a_url);

  if(a_externalID != -1)
    result->createChild("ExternalID")->createAttribute("Value", a_externalID);

  if(hasNetworkTimeout())
    result->createChild("NetworkTimeout")->createAttribute("Value", getNetworkTimeout());

  xml::Node* requests = result->createChild("ldap.Requests");
  const Response* response;
  const Request* request;

  for(const_response_iterator ii = response_begin(), maxii = response_end(); ii != maxii; ii ++) {
    if((request = Session::response(ii)->getRequest()) != NULL)
      request->asXML(requests);
  }

  return result;
}

int Session::getDangerousFileDescriptor() const
throw(RuntimeException) {
  ResultCode resultCode;
  int result = -1;
  resultCode = ldap_get_option((LDAP*) a_ldap, LDAP_OPT_DESC, &result);

  if(resultCode.isOk() == false)
    throw RuntimeException(ldap::Exception(this, resultCode, ANNA_FILE_LOCATION));

  LOGWARNING(
    string msg("Session::getDangerousFileDescriptor | Result: ");
    msg += functions::asString(result);
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
  return result;
}

const char* Session::asText(const State::_v state)
throw() {
  static const char* states [] = { "Closed", "WaitingBind", "Bound" };
  return states [state];
}

IdMessage Session::SortById::value(const Response* response)
throw() {
  return response->getIdMessage();
}
