// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


//------------------------------------------------------------------------------
//--------------------------------------------------------------------- #include
//------------------------------------------------------------------------------


// Local
#include <anna/statistics/Accumulator.hpp>
#include <anna/statistics/Engine.hpp>

#include <anna/time/Date.hpp>

// Standard
#include <math.h>
#include <sys/time.h>
#include <limits.h>

#include <anna/xml/xml.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>



using namespace anna::statistics;
using namespace anna::time;


// Private functions:


//------------------------------------------------------------------------------
//---------------------------------------------------- Accumulator::initialize()
//------------------------------------------------------------------------------
void Accumulator::initialize(const int & conceptId) throw() {
  _concept_data_t conceptData;
  conceptData.reset();
  a_concept_data_map[conceptId] = conceptData;
}


//------------------------------------------------------------------------------
//---------------------------------------------------- Accumulator::addConcept()
//------------------------------------------------------------------------------
int Accumulator::addConcept(const std::string & description, const std::string & unit, const bool & integerNatureSample, const char *conceptNameFormat) throw() {
  std::string conceptName = anna::functions::asString(conceptNameFormat, description.c_str(), a_name.c_str());
  return anna::statistics::Engine::instantiate().addConcept(conceptName, unit, integerNatureSample);
}


//------------------------------------------------------------------------------
//---------------------------------------------------- Accumulator::getConcept()
//------------------------------------------------------------------------------
_concept_data_t *Accumulator::getConcept(const int & conceptId) const throw(anna::RuntimeException) {
  _concept_data_map_iter it = a_concept_data_map.find(conceptId);

  if(it == a_concept_data_map.end()) {  // not found
    // Check if concept id is registered:
    Engine& engine = Engine::instantiate();
    std::string conceptDescription, conceptUnitDescription;
    bool integerNature;

    if(!engine.getConcept(conceptId, conceptDescription, conceptUnitDescription, integerNature)) {
      std::string msg = "Can't find at engine nothing about concept id = ";
      msg += anna::functions::asString(conceptId);
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    Accumulator * nc_this = const_cast<Accumulator *>(this);
    nc_this->initialize(conceptId);
    it = a_concept_data_map.find(conceptId);
  }

  return ((_concept_data_t *) & ((*it).second));
}


//------------------------------------------------------------------------------
//--------------------------------------------------- Accumulator::floatFormat()
//------------------------------------------------------------------------------
std::string Accumulator::floatFormat(const int & numberOfDecimals) const throw() {
  std::string result = "%.";
  result += anna::functions::asString(numberOfDecimals);
  result += "f";
  return (result);
}


//------------------------------------------------------------------------------
//------------------------------------------ Accumulator::getStandardDeviation()
//------------------------------------------------------------------------------
double Accumulator::getStandardDeviation(const _concept_data_t * conceptData) const throw(anna::RuntimeException) {
  // SD = sqrt (1/N SUM (xi^2) - X^2) = sqrt (SquareSum / N - Average^2)
  if(conceptData->Size == 0)
    throw anna::RuntimeException("Divide by zero: sample size = 0 for Standard Deviation !!", ANNA_FILE_LOCATION);

  return (::sqrt((conceptData->SquareSum / conceptData->Size) - (::pow(conceptData->Average, (double)2))));
}


//------------------------------------------------------------------------------
//------------------------------------ Accumulator::getBesselStandardDeviation()
//------------------------------------------------------------------------------
double Accumulator::getBesselStandardDeviation(const _concept_data_t * conceptData) const throw(anna::RuntimeException) {
  // BSD = sqrt (1/(N-1) SUM (xi^2) - N/(N-1) X^2) = sqrt (SquareSum / (N-1) - N/(N-1) Average^2)
  if(conceptData->Size == 1)
    throw anna::RuntimeException("Divide by zero: sample size = 1 for bessel's Standard Deviation !!", ANNA_FILE_LOCATION);

  double N = conceptData->Size;
  return (::sqrt((conceptData->SquareSum / (N - 1)) - (N / (N - 1)) * (::pow(conceptData->Average, (double)2))));
}



// Public functions:


//------------------------------------------------------------------------------
//------------------------------------------------------- Accumulator::process()
//------------------------------------------------------------------------------
void Accumulator::process(const int & conceptId, const double & value) throw(anna::RuntimeException) {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "process", ANNA_FILE_LOCATION));
  Engine& engine = Engine::instantiate();

  if(!engine.enabled()) return;

  _concept_data_t *ptr_auxConceptData = getConcept(conceptId);
  anna::Millisecond current = anna::Millisecond::getTime();
  // Optional sample file dump:
  Engine::instantiate().logSample(conceptId, current, value); // Accumulator is friend of Engine

  // Statistics calculations
  if(ptr_auxConceptData->Size == ULLONG_MAX)  // Statistics is better during processing until reset
    ptr_auxConceptData->reset();

  ptr_auxConceptData->Size ++;

  // If first:
  if(ptr_auxConceptData->Size == 1) {
    ptr_auxConceptData->MinimumProcessed = value;
    ptr_auxConceptData->MaximumProcessed = value;
    ptr_auxConceptData->MinimumEventTimestampMs = current;
    ptr_auxConceptData->MaximumEventTimestampMs = current;
  } else {
    if(value < ptr_auxConceptData->MinimumProcessed) {
      ptr_auxConceptData->MinimumProcessed = value;
      ptr_auxConceptData->MinimumEventTimestampMs = current;
    }

    if(value > ptr_auxConceptData->MaximumProcessed) {
      ptr_auxConceptData->MaximumProcessed = value;
      ptr_auxConceptData->MaximumEventTimestampMs = current;
    }
  }

  ptr_auxConceptData->Sum += value;
  ptr_auxConceptData->SquareSum += ::pow(value, (double)2); // (*) standard deviation formula simplification
  ptr_auxConceptData->Average = ptr_auxConceptData->Sum / ptr_auxConceptData->Size;
  //ptr_auxConceptData->SquareSumSampleDeviation += ::pow(value - ptr_auxConceptData->Average, (double)2);
  // Don't use former method to get standard deviation because is not correct: in each processing, average
  //  will be change (is not fixed). There is a simplification for standard deviation, based on square sum (*):
  // Take X as average for x1, x2, x3, etc., xN
  // SUM (xi - X)^2    = SUM (xi^2 + X^2 - 2xiX) = SUM (xi^2) + NX^2 - 2X SUM (xi) =
  //                   = SUM (xi`2) + NX^2 - 2X(NX) = SUM (xi^2) - NX^2
  //
  // Then, SD = sqrt (1/N SUM (xi^2) - X^2) = sqrt (SquareSum / N - Average^2)
  // And corrected SD (Bessel's standard deviation) is:
  // BSD = sqrt (1/(N-1) SUM (xi^2) - N/(N-1) X^2) = sqrt (SquareSum / (N-1) - N/(N-1) Average^2)
  // All these values don't need to be calculated now
  // Worked with pointer
  //a_concept_data_map[conceptId] = *ptr_auxConceptData;
}


//------------------------------------------------------------------------------
//------------------------------------------------------- Accumulator::operator=
//------------------------------------------------------------------------------
const Accumulator & Accumulator::operator = (const Accumulator & accumulator) {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "operator=", ANNA_FILE_LOCATION));
  // Avoid self copy: i.e., Accumulator a,b; a=&b; a=b; b=a;
  if(this == &accumulator) return (*this);

  a_concept_data_map = accumulator.a_concept_data_map;
  return (*this);
}



//------------------------------------------------------------------------------
//--------------------------------------------------------- Accumulator::reset()
//------------------------------------------------------------------------------
void Accumulator::reset(void) throw() {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "reset", ANNA_FILE_LOCATION));
  _concept_data_map_iter it;
  _concept_data_map_iter it_min(a_concept_data_map.begin());
  _concept_data_map_iter it_max(a_concept_data_map.end());

  for(it = it_min; it != it_max; it++) {
    //reset ((*it).first);
    _concept_data_t * ptr_auxConceptData = (_concept_data_t*) & ((*it).second);
    ptr_auxConceptData->reset();
  }
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Accumulator::reset()
//------------------------------------------------------------------------------
void Accumulator::reset(const int & conceptId) throw(anna::RuntimeException) {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "reset", ANNA_FILE_LOCATION));
  _concept_data_t *ptr_auxConceptData = getConcept(conceptId); // will initialize if didn't associated to this accumulator,
                                                               // exception if engine knows nothing about such concept id.
  //if (ptr_auxConceptData == NULL) return; // Not possible: getConcept initializes it if not found or exception was launched
                                            //               previously if the engine don't know the concept id.
  ptr_auxConceptData->reset();
}


// Gets:

//------------------------------------------------------------------------------
//---------------------------------------------------- Accumulator::sampleSize()
//------------------------------------------------------------------------------
unsigned long long int Accumulator::sampleSize(const int & conceptId) const throw(anna::RuntimeException) {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "sampleSize", ANNA_FILE_LOCATION));
  _concept_data_t *ptr_auxConceptData = getConcept(conceptId);
  //if (ptr_auxConceptData == NULL) return 0; // Not possible: getConcept initializes it if not found.
  return (ptr_auxConceptData->Size);
}


//------------------------------------------------------------------------------
//------------------------------------------------------ Accumulator::getValue()
//------------------------------------------------------------------------------
double Accumulator::getValue(const int & conceptId, const Operation::Type & operation) const throw(anna::RuntimeException) {
// LOGMETHOD (TraceMethod tttm ("anna::statistics::Accumulator", "getValue", ANNA_FILE_LOCATION));
  const _concept_data_t *ptr_auxConceptData = getConcept(conceptId);

  switch(operation) {
    //case Operation::Sum:
    // return (ptr_auxConceptData->Sum);
  case Operation::Average:
    return (ptr_auxConceptData->Average);
  case Operation::StandardDeviation:
    return (getStandardDeviation(ptr_auxConceptData));
  case Operation::BesselStandardDeviation:
    return (getBesselStandardDeviation(ptr_auxConceptData));
  case Operation::Minimum:
    return (ptr_auxConceptData->MinimumProcessed);
  case Operation::Maximum:
    return (ptr_auxConceptData->MaximumProcessed);
  default:
    throw anna::RuntimeException("Unrecognized operation", ANNA_FILE_LOCATION);
  }

  //if (ptr_auxConceptData == NULL) return ((double)0); // Not possible: getConcept initializes it if not found.
}


//------------------------------------------------------------------------------
//------------------------------------------------------ Accumulator::asString()
//------------------------------------------------------------------------------
std::string Accumulator::asString(const int & numberOfDecimals) const throw() {
  std::string trace;
  _concept_data_map_iter iter;
  _concept_data_map_iter iter_min(a_concept_data_map.begin());
  _concept_data_map_iter iter_max(a_concept_data_map.end());
  time_t unixTimestamp;
  time::Date time_now; // set current time on local TZ
  time::Date time_aux; // for events
  Engine& engine = Engine::instantiate();
  std::string fFormat = floatFormat(numberOfDecimals);
  trace =  "\n=====================";
  trace +=  "\nStatistic Information";
  trace +=  "\n=====================";
  trace += "\nAccumulator name: ";
  trace += a_name;
  trace += "\nCurrent Time: ";
  trace += time_now.asString();

  if(a_concept_data_map.size() == 0) {
    trace += "\nNo concept data found (empty map)";
    return trace;
  }

  trace += "\nNumber of elements (measured concepts) = "; trace += anna::functions::asString(a_concept_data_map.size()); trace += "\n";
  int conceptId;
  std::string conceptDescription;
  std::string conceptUnitDescription;
  bool integerNatureSample;

  for(iter = iter_min; iter != iter_max; iter++) {
    conceptId = (*iter).first;
    const _concept_data_t * ptrConceptData = &((*iter).second);
    engine.getConcept(conceptId, conceptDescription, conceptUnitDescription, integerNatureSample);  // never launch exception
    trace += "\n   Concept:              [";
    trace += anna::functions::asString(conceptId);
    trace += "] ";
    trace += conceptDescription; trace += " ("; trace += conceptUnitDescription; trace += ")";

    if(ptrConceptData->Size != 0) {
      trace += "\n   Sample Size:                 "; trace += anna::functions::asString((anna::U64)(ptrConceptData->Size));
      trace += "\n   Average:                     "; trace += anna::functions::asString(fFormat.c_str(), ptrConceptData->Average);
      trace += "\n   Standard Deviation:          "; trace += anna::functions::asString(fFormat.c_str(), getStandardDeviation(ptrConceptData));

      if(ptrConceptData->Size != 1) {
        trace += "\n   Bessel's Standard Deviation: "; trace += anna::functions::asString(fFormat.c_str(), getBesselStandardDeviation(ptrConceptData));
      }

      trace += "\n   Minimum:                     ";

      if(integerNatureSample)
        trace += anna::functions::asString((int)(ptrConceptData->MinimumProcessed));
      else
        trace += anna::functions::asString(fFormat.c_str(), ptrConceptData->MinimumProcessed);

      // Minimum Timestamp:
      unixTimestamp = (ptrConceptData->MinimumEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      trace += " ("; trace += time_aux.asString(); trace += ")";
      trace += "\n   Maximum:                     ";

      if(integerNatureSample)
        trace += anna::functions::asString((int)(ptrConceptData->MaximumProcessed));
      else
        trace += anna::functions::asString(fFormat.c_str(), ptrConceptData->MaximumProcessed);

      // Maximum Timestamp:
      unixTimestamp = (ptrConceptData->MaximumEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      trace += " ("; trace += time_aux.asString(); trace += ")";
      // Last Reset Event:
      unixTimestamp = (ptrConceptData->LastResetEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      trace += "\n   Last Reset Timestamp:        "; trace += time_aux.asString();

      // Processing Rate: it is not updated at 'process' to avoid 'time_now' use on every call
      if(time_now != time_aux) {
        trace += "\n   Processing rate:             ";
        int lapse = time_now.getUnixTimestamp() - time_aux.getUnixTimestamp();
        trace += anna::functions::asString(fFormat.c_str(), ptrConceptData->Size / ((double)lapse));
        trace += " vps (values per second)";
      }
    } else {
      trace += "\n   Not enough data to get statistics";
    }
  }

  trace += "\n";
  return (trace);
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Accumulator::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* Accumulator::asXML(anna::xml::Node* parent, const int & numberOfDecimals) const throw() {
  anna::xml::Node* result = parent->createChild("anna.statistics.Accumulator");
  result->createAttribute("Name", a_name);
  _concept_data_map_iter iter;
  _concept_data_map_iter iter_min(a_concept_data_map.begin());
  _concept_data_map_iter iter_max(a_concept_data_map.end());
  time_t unixTimestamp;
  time::Date time_now; // set current time on local TZ
  time::Date time_aux; // for events
  Engine& engine = Engine::instantiate();
  std::string fFormat = floatFormat(numberOfDecimals);
  //result->createAttribute("CurrentTime", time_now.asString());
  anna::xml::Node* currentTime = result->createChild("CurrentTime");
  time_now.asXML(currentTime);

  if(a_concept_data_map.size() == 0) {
    result->createAttribute("ConceptDataMapSize", "0: No concept data found (empty map)");
    return result;
  }

  result->createAttribute("ConceptDataMapSize", anna::functions::asString(a_concept_data_map.size()));
  int conceptId;
  std::string conceptDescription;
  std::string conceptUnitDescription;
  bool integerNatureSample;
  std::string cad_aux;

  for(iter = iter_min; iter != iter_max; iter++) {
    conceptId = (*iter).first;
    const _concept_data_t * ptrConceptData = &((*iter).second);
    engine.getConcept(conceptId, conceptDescription, conceptUnitDescription, integerNatureSample);  // never launch exception
    anna::xml::Node* concept = result->createChild("Concept");
    concept->createAttribute("Id", conceptId);
    concept->createAttribute("Description", conceptDescription);
    concept->createAttribute("Unit", conceptUnitDescription);
    anna::xml::Node* data = concept->createChild("Sample");

    if(ptrConceptData->Size != 0) {
      data->createAttribute("Size", anna::functions::asString((anna::U64)ptrConceptData->Size));
      data->createAttribute("Average", anna::functions::asString(fFormat.c_str(), ptrConceptData->Average));
      data->createAttribute("StandardDeviation", anna::functions::asString(fFormat.c_str(), getStandardDeviation(ptrConceptData)));

      if(ptrConceptData->Size != 1)
        data->createAttribute("BesselStandardDeviation", anna::functions::asString(fFormat.c_str(), getBesselStandardDeviation(ptrConceptData)));

      // Minimum:
      anna::xml::Node* minimum = data->createChild("Minimum");

      if(integerNatureSample)
        minimum->createAttribute("Value", anna::functions::asString((int)(ptrConceptData->MinimumProcessed)));
      else
        minimum->createAttribute("Value", anna::functions::asString(fFormat.c_str(), ptrConceptData->MinimumProcessed));

      unixTimestamp = (ptrConceptData->MinimumEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      time_aux.asXML(minimum);
      // Maximum:
      anna::xml::Node* maximum = data->createChild("Maximum");

      if(integerNatureSample)
        maximum->createAttribute("Value", anna::functions::asString((int)(ptrConceptData->MaximumProcessed)));
      else
        maximum->createAttribute("Value", anna::functions::asString(fFormat.c_str(), ptrConceptData->MaximumProcessed));

      unixTimestamp = (ptrConceptData->MaximumEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      time_aux.asXML(maximum);
      // Last Reset Event:
      anna::xml::Node* lastResetEvent = data->createChild("LastResetTimestamp");
      unixTimestamp = (ptrConceptData->LastResetEventTimestampMs / 1000); // not neccessary such precision
      time_aux.storeUnix(unixTimestamp);
      time_aux.asXML(lastResetEvent);

      // Processing Rate: it is not updated at 'process' to avoid 'time_now' use on every call
      if(time_now != time_aux) {
        int lapse = time_now.getUnixTimestamp() - time_aux.getUnixTimestamp();
        cad_aux = anna::functions::asString(fFormat.c_str(), ptrConceptData->Size / ((double)lapse)); cad_aux += " vps (values per second)";
        data->createAttribute("ProcessingRate", cad_aux);
      }
    } else {
      concept->createAttribute("Sample", "Not enough data to get statistics");
    }
  }

  return result;
}
